/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
package com.portlandwebworks.commons.dao;


import com.portlandwebworks.commons.reflection.FieldUtils;
import org.springframework.beans.BeanUtils;

import javax.persistence.*;
import java.beans.PropertyDescriptor;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.*;


/**
 * Utility class for introspecting entity classes for various DAO-related properties or annotations.  These utility
 * methods were created for internal use but may serve some purpose outside of the innards of this library.
 * 
 * @author Ronnie Northup
 */
public class DaoUtils
{
	private static Map<Class, List<SortSpecification>> defaultEntitySortSpecs = new HashMap<Class, List<SortSpecification>>();

	/**
	 * Returns true if the entity class contains a named field, and that field is considered "multi-value".  A field is
	 * multi-value if it an instance of a <code>Collection</code> or <code>Map</code>.
	 *
	 * @param entityClass
	 * @param field
	 * @return              <code>true</code> if the named field is multi-value as defined above.
	 * @throws              IllegalArgumentException if the named field does not exist on the entity class or is inaccessible.
	 */
	public static boolean isMultiValueField(Class entityClass, String field)
	{
		// Look at the getter for this field to see if it returns a subclass of Collection.
		PropertyDescriptor desc = BeanUtils.getPropertyDescriptor(entityClass, field);
		if (desc == null) {
			throw new IllegalArgumentException("Field " + field + " does not exist on entity " + entityClass.getName());
		}
		return (Collection.class.isAssignableFrom(desc.getPropertyType()) || Map.class.isAssignableFrom(desc.getPropertyType()));
	}

	/**
	 * Returns a list of <code>SortSpecification</code> instances representing the combination of all fields on the
	 * entity or its mapped superclasses on which queries should sort/order if no other sort parameters are provided.
	 * Default sort is specified on an entity using the <code>@DefaultOrderBy</code> annotation.  If the default sort
	 * fields have sort aliases indicated by a <code>@SortProperty</code> annotation, that alias will be used instead
	 * of the field name.
	 *
	 * @param entityClass   The entity whose default sort fields are to be evaluated.
	 *
	 * @return  A <code>List</code> or <code>SortSpecification</code> instances.  If no default sort specifications are
	 *          defined for the class, an empty list is returned.
	 */
	public static List<SortSpecification> getDefaultSortSpecifications(Class entityClass)
	{
		// First, check the cache.
		List<SortSpecification> specs = defaultEntitySortSpecs.get(entityClass);
		if (specs == null) {

			List<DefaultOrderBy> orderBys = null;
			Map<DefaultOrderBy, String> orderByProperties = null;

			Field[] defaultOrderedFields = FieldUtils.getFieldsWithAnnotation(entityClass, DefaultOrderBy.class);

			for (Field field : FieldUtils.getFields(entityClass)) {
				DefaultOrderBy dob = field.getAnnotation(DefaultOrderBy.class);
				if (dob != null) {
					if (orderBys == null) {
						orderBys = new ArrayList<DefaultOrderBy>();
						orderByProperties = new HashMap<DefaultOrderBy, String>();
					}
					orderBys.add(dob);
					orderByProperties.put(dob, field.getName());
				}
			}

			// Create a container for the sort specs to be built.
			specs = new ArrayList<SortSpecification>(defaultOrderedFields.length);

			// If more than one, sort them by order of priority on the annotation.
			if (defaultOrderedFields.length > 1) {
				// Sort the list of order by annotations based on their priority
				Arrays.sort(defaultOrderedFields, new Comparator<Field>() {
					public int compare(Field o1, Field o2)
					{
						return o1.getAnnotation(DefaultOrderBy.class).priority() - o2.getAnnotation(DefaultOrderBy.class).priority();
					}
				});
			}

			if (defaultOrderedFields.length > 0) {
				// In order of priority, convert each annotation into a sort specification.
				for (Field defaultOrderedField : defaultOrderedFields) {
					String sortProp = null;
					SortProperty sortProperty = defaultOrderedField.getAnnotation(SortProperty.class);
					if (sortProperty != null) {
						sortProp = sortProperty.value();
					}
					else {
						sortProp = defaultOrderedField.getName();
					}

					specs.add(new SortSpecification(
							sortProp,
							defaultOrderedField.getAnnotation(DefaultOrderBy.class).ascending()? SortDirection.ASC : SortDirection.DESC
					));
				}
			}

			// Cache for later reuse
			defaultEntitySortSpecs.put(entityClass, specs);
		}

		return specs;
	}

	public static boolean isFieldLazyLoaded(Class clazz, String path)
	{
		Field f = FieldUtils.getField(clazz, path);
		for (Annotation annotation : f.getAnnotations()) {
			if (annotation instanceof OneToMany) {
				return ((OneToMany)annotation).fetch() == FetchType.LAZY;
			}
			else if (annotation instanceof ManyToOne) {
				return ((ManyToOne)annotation).fetch() == FetchType.LAZY;
			}
			else if (annotation instanceof ManyToMany) {
				return ((ManyToMany)annotation).fetch() == FetchType.LAZY;
			}
		}
		return false;
	}
}
