/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.dao;

import org.apache.commons.lang.StringUtils;
import org.codehaus.jackson.JsonFactory;
import org.codehaus.jackson.JsonParser;
import org.codehaus.jackson.JsonToken;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * This class works in conjunction with PagedSearchResult to restrict a large result set by returning one page of
 * results at a time.
 *
 * @author Ronnie Northup
 */
public class PagedSearchOptions implements IPagedSearchOptions
{
	private static List<SortSpecification> noSort = Collections.EMPTY_LIST;

	static JsonFactory factory = new JsonFactory();
	static {
		factory.configure(JsonParser.Feature.ALLOW_UNQUOTED_FIELD_NAMES, true);
		factory.configure(JsonParser.Feature.ALLOW_SINGLE_QUOTES, true);
	}

	private List<SortSpecification> sort;
	
	private String sortText;

	private Integer start;
	
	private Integer limit;

	public PagedSearchOptions()
	{
		this(0, Integer.MAX_VALUE);
	}

	public PagedSearchOptions(int start, int limit)
	{
		this.start = start;
		this.limit = limit;
	}

	public List<SortSpecification> getSort()
	{
		if (sort == null && StringUtils.isNotBlank(sortText)) {
			try {
				SortSpecification spec = null;
				JsonParser p = factory.createJsonParser(sortText);
				JsonToken token = p.nextToken();
				while (token != JsonToken.END_ARRAY) {
					if (token == JsonToken.START_ARRAY) {
						sort = new ArrayList<SortSpecification>();
					}
					else if (token == JsonToken.END_OBJECT) {
						sort.add(spec);
						spec = null;
					}
					else if (token == JsonToken.START_OBJECT) {
						spec = new SortSpecification();
					}
					else {
						String fieldName = p.getCurrentName();
						if (fieldName.equals("property")) {
							token = p.nextToken();
							spec.setProperty(p.getText());
						}
						else if (fieldName.equals("direction")) {
							token = p.nextToken();
							spec.setDirection(p.getText().equals("ASC")? SortDirection.ASC : SortDirection.DESC);
						}
					}
					token = p.nextToken();
				}
			}
			catch (IOException e) {
				throw new IllegalArgumentException("Could not parse sortText--invalid JSON: " + sortText);
			}
		}

		if (sort != null) {
			return sort;
		}
		else {
			return noSort;
		}
	}

	public void setSort(List<SortSpecification> sort)
	{
		this.sort = sort;
	}

	/**
	 * 
	 * @param sortText
	 */
	public void setSortText(String sortText)
	{
		this.sortText = sortText;
	}
	
	public String getSortText()
	{
		return sortText;
	}

	public void addSort(String propertyName, SortDirection direction)
	{
		if (propertyName == null)
			throw new IllegalArgumentException("sort property must not be null");

		if (direction == null)
			throw new IllegalArgumentException(("sort direction must not be null"));

		if (this.sort == null) {
			this.sort = new ArrayList<SortSpecification>();
		}

		this.sort.add(new SortSpecification(propertyName, direction));
	}

	public int getStart()
	{
		return start;
	}

	public void setStart(int start)
	{
		this.start = start;
	}

	public int getLimit()
	{
		return limit;
	}

	public void setLimit(int limit)
	{
		this.limit = limit;
	}
}