/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.data;

import com.portlandwebworks.commons.domain.CodeValue;
import com.portlandwebworks.commons.domain.EntityBase;
import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.domain.LookupValue;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;

import java.io.Serializable;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class EntityUtils
{
	/**
	 * Scans a list of entities, return the entity whose ID was specified.
	 *
	 * @param source
	 * @param id
	 * @param <T>
	 * @return      The entity with the matching ID, or <code>null</code> if no entity with the specified ID was found.
	 */
	public static <T extends Identifiable> T findById(List<T> source, Serializable id)
	{
		for (T entity : source) {
			if (entity.getId().equals(id)) {
				return entity;
			}
		}
		return null;
	}

	/**
	 * Finds all entities whose IDs are contained in the collection of IDs provided.
	 *
	 * @param source
	 * @param ids
	 * @param <T>
	 * @return
	 */
	public static <T extends Identifiable> Map<Serializable, T> findAllByIds(List<T> source, Collection<? extends Serializable> ids)
	{
		return findAllByIds(source, (Serializable[])new ArrayList(ids).toArray(new Serializable[ids.size()]));
	}

	/**
	 *
	 *
	 * @param source
	 * @param ids
	 * @param <T>
	 * @return
	 */
	public static <T extends Identifiable> Map<Serializable, T> findAllByIds(List<T> source, Serializable[] ids)
	{
		HashMap<Serializable, T> all = new HashMap<Serializable,T>();
		for (T entity : source) {
			if (ArrayUtils.contains(ids, entity.getId())) {
				all.put(entity.getId(), entity);
			}
		}
		return all;
	}

	/**
	 *
	 *
	 * @param source
	 * @param ids
	 * @param <T>
	 * @return
	 */
	public static <T extends Identifiable> Map<Serializable, T> findComplementByIds(List<T> source, Serializable[] ids)
	{
		HashMap<Serializable, T> complement = new HashMap<Serializable,T>();
		for (T entity : source) {
			if (!ArrayUtils.contains(ids, entity.getId())) {
				complement.put(entity.getId(), entity);
			}
		}
		return complement;
	}

	/**
	 * Returns an array containing the unique identifiers for all entities in the collection provided.
	 *
	 * @param entities
	 * @return          An
	 */
	public static <T> Long[] getEntityIdArray(Collection<? extends EntityBase> entities)
	{
		if (entities == null || entities.size() == 0) {
			return new Long[0];
		}

		Long[] entityIds = new Long[entities.size()];
		int index = 0;
		for (EntityBase entity : entities) {
			entityIds[index++] = entity.getId();
		}
		return entityIds;
	}

	/**
	 * Generates a list of IDs from a collection of entities.
	 *
	 * @param identifiables
	 *
	 * @return
	 */
	public static <T extends Serializable> List<T> getEntityIdList(Collection<? extends Identifiable<T>> identifiables)
	{
		List<T> entityIds = new ArrayList<T>();
		if (! CollectionUtils.isEmpty(identifiables)) {
			for (Identifiable<T> entity : identifiables) {
				entityIds.add(entity.getId());
			}
		}
		return entityIds;
	}

	/**
	 *
	 * @param lookupValues
	 * @return
	 */
	public static Integer[] getLookupValueIdArray(Collection<? extends LookupValue> lookupValues)
	{
		if (lookupValues == null || lookupValues.size() == 0) {
			return new Integer[0];
		}

		Integer[] lookupIds = new Integer[lookupValues.size()];
		int index = 0;
		for (LookupValue lookups : lookupValues) {
			lookupIds[index++] = lookups.getId();
		}
		return lookupIds;
	}

	/**
	 *
	 * @param lookupValues
	 * @return
	 */
	public static List<Integer> getLookupValueIdList(Collection<? extends LookupValue> lookupValues)
	{
		List<Integer> lookupIds = new ArrayList<Integer>();
		if (! CollectionUtils.isEmpty(lookupValues)) {
			for (LookupValue lookups : lookupValues) {
				lookupIds.add(lookups.getId());
			}
		}
		return lookupIds;
	}

	/**
	 *
	 * @param codeValues
	 * @return
	 */
	public static Integer[] getCodeValueIdArray(Collection<? extends CodeValue> codeValues)
	{
		if (codeValues == null || codeValues.size() == 0) {
			return new Integer[0];
		}

		Integer[] codeIds = new Integer[codeValues.size()];
		int index = 0;
		for (CodeValue codes : codeValues) {
			codeIds[index++] = codes.getId();
		}
		return codeIds;
	}

	/**
	 *
	 * @param codeValues
	 * @return
	 */
	public static List<Integer> getCodeValueIdList(Collection<? extends CodeValue> codeValues)
	{
		List<Integer> codeIds = new ArrayList<Integer>();
		if (! CollectionUtils.isEmpty(codeValues)) {
			for (CodeValue code : codeValues) {
				codeIds.add(code.getId());
			}
		}
		return codeIds;
	}
}
