package com.portlandwebworks.commons.hibernate;

import com.portlandwebworks.commons.data.EntityMapping;
import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.reflection.ClassUtils;
import com.portlandwebworks.commons.reflection.FieldUtils;
import com.portlandwebworks.commons.reflection.NodeType;
import org.apache.commons.lang.ArrayUtils;
import org.hibernate.transform.ResultTransformer;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * This transformer is capable of collapsing raw result set data into complex graphs of data.
 *
 * TODO: This class is very inefficient.  It creates nested elements for every row of repeated data before it is collapsed.
 *
 * @author Ronnie Northup
 */
public class AliasToBeanGraphResultTransformer implements ResultTransformer
{
	private Class<?> rootClass;
	private static final Integer SINGLE_OBJECT_KEY = 0;

	/**
	 * Constructor with return type.
	 *
	 * @param rootClass The desired type for return values created by this transformer.
	 */
	public AliasToBeanGraphResultTransformer(Class<?> rootClass)
	{
		this.rootClass = rootClass;
	}
	
	private HashMap<Class<? extends Identifiable>, Map<Serializable, Identifiable>> objectRegistry = new HashMap<Class<? extends Identifiable>, Map<Serializable, Identifiable>>();
	
	public <T extends Identifiable> T getObject(Class<T> clazz, Serializable identifier)
	{
		Map<Serializable, Identifiable> set = objectRegistry.get(clazz);
		if (set == null) {
			return null;
		}
		return (T)set.get(identifier);
	}
	
	public <T extends Identifiable> void putObject(T object)
	{
		Class clazz = object.getClass();
		Map<Serializable, Identifiable> set = objectRegistry.get(clazz);
		if (set == null) {
			set = new HashMap<Serializable, Identifiable>();
			objectRegistry.put(clazz, set);
		}
		set.put(object.getId(), object);
	}

	/**
	 * Called by Hibernate to convert one row of raw result set data into a single object.
	 *
	 * @param tuple     A row of raw result set data
	 * @param aliases   The aliases for the tuple.
	 *
	 * @return          A bean populated with the tuple data.
	 */
	public Object transformTuple(Object[] tuple, String[] aliases)
	{
		final Object entity = ClassUtils.newInstanceOfClass(rootClass);

		for (int i = 0; i < tuple.length; i++) {
			final String alias = aliases[i];
			final Object value = tuple[i];

			FieldUtils.traversePathInClassGraph(rootClass, alias, new FieldUtils.FieldTypeVisitor() {
				/**
				 * This will be a reference to the parent object that has been built.
				 */
				Object targetObject = entity;

				public boolean visit(String path, Class<?> elementType, NodeType nodeType)
				{
					//TODO: shouldn't this type be passed in to this method?
					Field targetField = FieldUtils.getField(targetObject.getClass(), path);

					if (nodeType == NodeType.ValueType) {
							// We've reached the point in the association path where we can apply the value that was provided
							// by the tuple.  Only apply a null value if the target field supports setting a null value.
							if (value != null || ! targetField.getType().isPrimitive()) {
								FieldUtils.setFieldValue(targetObject, targetField, value);
							}
							return false;
					}
					else if (nodeType == NodeType.Object) {
							// field is not multi-value field or a scalar value.  It must have its own entity mapping to
							// continue.
							EntityMapping mapping = targetField.getType().getAnnotation(EntityMapping.class);
							if (mapping != null) {
								// TODO: construct new object from the type specified on the EntityMapping, and if the
								// TODO: target field does not accept that type, throw an exception
								Object nestedElement = ClassUtils.newInstanceOfClass(elementType);
								FieldUtils.setFieldValue(targetObject, targetField, nestedElement);
								targetObject = nestedElement;
							}
							else {
								throw new RuntimeException("Nested element has no entity mapping: " + targetField.getName());
							}

							// Not done traversing path yet.
							return true;
					}
					else {
						// If the current collection array or map is not initialized, initialize it now.
						Object container = FieldUtils.getFieldValue(targetObject, targetField);
						if (container == null) {
							switch (nodeType) {
								case Collection:
									container = ClassUtils.newInstanceOfClass(targetField.getType());
									FieldUtils.setFieldValue(targetObject, targetField, container);
									break;
								case Map:
									container = ClassUtils.newInstanceOfClass(targetField.getType());
									FieldUtils.setFieldValue(targetObject, targetField, container);
									break;
								case Array:
									Object[] array;
									if (value != null) {
										array = new Object[1];
									}
									else {
										array = new Object[0];
									}
									container = array;
									FieldUtils.setFieldValue(targetObject, targetField, container);
									break;
							}
						}

						// Determines how the value to be applied is applied
						boolean isValueType = FieldUtils.isValueType(elementType);

						// If working with a value type, simply add it to the container.  If an object type, check the
						// container for it, add it if it's not there.
						if (value != null) {
							Object nestedElement = null;
							switch (nodeType) {
								case Collection:
									Collection collection = (Collection)container;
									if (isValueType ) {
										collection.add(value);
									}
									else {
										if (collection.size() > 0) {
											nestedElement = collection.iterator().next();
										}
										else {
											nestedElement = ClassUtils.newInstanceOfClass(elementType);
											collection.add(nestedElement);
										}
									}
									break;
								case Map:
									Map map = (Map)container;
									if (isValueType) {
										map.put(SINGLE_OBJECT_KEY, value);
									}
									else {
										if (map.size() > 0) {
											nestedElement = map.get(SINGLE_OBJECT_KEY);
										}
										else {
											nestedElement = ClassUtils.newInstanceOfClass(elementType);
											map.put(SINGLE_OBJECT_KEY, nestedElement);
										}
									}
									break;
								case Array:
									Object[] array = (Object[])container;
									if (isValueType) {
										array[0] = value;
									}
									else {
										nestedElement = ClassUtils.newInstanceOfClass(elementType);
										array[0] = nestedElement;
									}
									break;
							}

							if (isValueType) {
							// Don't continue processing if a value type was just added to a collection
								return false;
							}
							else {
								// Item added to container can be traversed.  Track it.
								targetObject = nestedElement;
								return true;
							}
						}
						else {
							// Value to apply is null.  No need to continue traversal.
							return false;
						}
					}
				}
			});
		}

		return entity;
	}

	/**
	 * Called by Hibernate to collapse possibly repeating objects built by separate callse to transformTuple().
	 *
	 * @param collection
	 * @return
	 */
	public List transformList(List collection)
	{
		return transformCollection(collection);
	}

	/**
	 * Iterates through each result in the result set collection, and collapse all repeating records.  When two  results
	 * are collapsed into one, combines the contents of both
	 *
	 * @param collection
	 * 
	 * @return
	 */
	protected List transformCollection(Collection collection)
	{
		Field[] uniqueFields = null;

		//Field[] fields = null;

		ArrayList<Object> uniqueResults = new ArrayList<Object>();

		//
		for (Object currResult : collection) {

			// Once we know what type of object we're working with, can get the set of fields that uniquely identify one
			// of these objects.  This of course assumes the collection is homogeneous.
			if (uniqueFields == null) {
				uniqueFields = FieldUtils.getIdentifierFieldsForClass(currResult.getClass());
				if (uniqueFields.length == 0) {
					throw new RuntimeException("Target entity for transformation has no identifier field mapping defined");
				}
			}

			// Get a reference to the last root object that was processed, if any were previously processed.
			Object lastResult = null;
			if (uniqueResults.size() > 0) {
				lastResult = uniqueResults.get(uniqueResults.size() - 1);
			}

			// If there is no previous result, we are processing the first row.  Just add it to the return collection
			// and move on.
			if (lastResult == null) {
				uniqueResults.add(currResult);
				continue;
			}

			// Sanity check.  Make sure that the root items are of the same class.
			if (! lastResult.getClass().equals(currResult.getClass())) {
				throw new RuntimeException(String.format(
						"Unable to process root result elements of different types: %s <> %s",
						lastResult.getClass().getName(), currResult.getClass().getName()));
			}

			// Get the currently and previously processed result root items so that they can be compared for equality.
			Object[] lastUnique = FieldUtils.getFieldValues(lastResult, uniqueFields);
			Object[] currUnique = FieldUtils.getFieldValues(currResult, uniqueFields);
			if (! ArrayUtils.isEquals(lastUnique, currUnique)) {
				// Current root object does not match the last one.  Must be a new root object, so add it to the return
				// collection.
				uniqueResults.add(currResult);
			}
			else {
				final Stack targetEntityStack = new Stack();
				targetEntityStack.push(lastResult);
				
				FieldUtils.traverseObjectGraph(currResult, new FieldUtils.NodeVisitor()
				{
					public boolean visit(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
					{
						Object targetEntity = targetEntityStack.peek();
						switch (nodeType) {
							case ValueType:
								// Do nothing.
								break;
							case Object: {
								// If nodeField is non-null, it means we are processing a one-to-one relationship.  Nothing
								// to do here except put the target entity's nested value on the stack to mirror traversal
								// of the object in the source object graph.
								if (nodeField != null) {
									Object newTarget = FieldUtils.getFieldValue(targetEntity, nodeField);
									targetEntityStack.push(newTarget);
								}
								break;
							}
							case Collection: {
								Collection toMerge = (Collection)nodeValue;
								if (toMerge.isEmpty()) {
									// Nothing to do here.  Stop traversing this path.
									return false;
								}

								// Because of the nature of the flattened result set data, there will only ever be one
								// member in this collection.
								Object srcObj = toMerge.iterator().next();

								// The collection in the target entity to which this member is to be merged.
								Collection targetCollection = (Collection)FieldUtils.getFieldValue(targetEntity, nodeField);

								if (FieldUtils.isValueType(srcObj.getClass())) {
									// Data is value type.  Check target collection and add if it's not there.
									if (! targetCollection.contains(srcObj)) {
										targetCollection.add(srcObj);
									}
									return false;
								}
								else {
									if (targetCollection.size() == 0) {
										// Easy decision to make.  This object is definitely not in the target collection.
										// Add it to the target collection and terminate this traversal path
										targetCollection.add(srcObj);
										return false;
									}
									else {
										// Working with entities.  Compare objects by ID(s) in source and target
										// collections.
										boolean found = false;
										Field[] idFields = FieldUtils.getIdentifierFieldsForClass(toMerge.iterator().next().getClass());

										for (Object destObj : targetCollection) {
											Object[] srcIds = FieldUtils.getFieldValues(srcObj, idFields);
											Object[] dstIds = FieldUtils.getFieldValues(destObj, idFields);

											if (ArrayUtils.isEquals(srcIds, dstIds)) {
												// This entity exists in both the source and target collections.
												// Don't replace the target collection's version with this one, but
												// there may be aggregate collections that require merging so add
												// the target collection's member to the stack so that on the next
												// node visit we can compare the source entity's nested entities to
												// the target entity's nested entities, if there are any.
												found = true;
												targetEntityStack.push(destObj);
												break;
											}
										}

										if (! found) {
											// This entity is not a member of the target collection.  Add it and end
											// this node traversal, since there is nothing more that would need to
											// be merged
											targetCollection.add(srcObj);
											return false;
										}
									}
								}
								break;
							}
							case Map:
								// TODO: Check target entity's map for this child entity.  If not found, add it.
								break;
							case Array: {
								Object[] toMerge = (Object[])nodeValue;
								if (toMerge.length > 0) {
									
									Object srcObj = toMerge[0];
									
									Object[] targetArray = (Object[])FieldUtils.getFieldValue(targetEntity, nodeField);
									if (targetArray.length == 0) {
										FieldUtils.setFieldValue(targetEntity, nodeField, toMerge);
										targetEntityStack.push(srcObj);
									}
									else {
										boolean found = false;
										Field[] idFields = FieldUtils.getIdentifierFieldsForClass(srcObj.getClass());

										for (Object destObj : targetArray) {
											Object[] srcIds = FieldUtils.getFieldValues(srcObj, idFields);
											Object[] destIds = FieldUtils.getFieldValues(destObj, idFields);

											if (ArrayUtils.isEquals(srcIds, destIds)) {
												found = true;
												targetEntityStack.push(destObj);
												break;
											}
										}

										if (! found) {
											Object[] newArray = new Object[targetArray.length + 1];
											System.arraycopy(targetArray, 0, newArray, 0, targetArray.length);
											newArray[newArray.length - 1] = srcObj;
											FieldUtils.setFieldValue(targetEntity, nodeField, newArray);
											targetEntityStack.push(srcObj);
										}
									}
								}
								break;
							}
						}
						return true;
					}

					public void leaving(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
					{
						if (nodeType != NodeType.ValueType && nodeType != NodeType.Collection) {
						//if (nodeType != NodeType.ValueType && nodeField != null) {
							targetEntityStack.pop();
						}
					}
				});
/*

				
				// Iterate over the multi-valued fields on this object, and add the "current" item's collection members
				// to the "last" item's collections.
				if (fields == null) {
					fields = FieldUtils.getFields(currResult.getClass());
				}

				// Iterate through each field on the result type, looking for arrays and collections.  These fields must
				// have their members combined into the version of the root item that we are going to preserve for the
				// return list.
				for (Field field : fields) {
					NodeType nodeType = FieldUtils.getNodeTypeForClass(field.getType());
					switch (nodeType) {
						case Array:
							try {
								// Strategy for copying one item's array elements to another.
								Object[] currValues = (Object[])field.get(currResult);
								Object[] lastValues = (Object[])field.get(lastResult);
								Object[] combined = ArrayUtils.addAll(lastValues, currValues);
								field.set(lastResult, combined);
							}
							catch (IllegalAccessException e) {
								throw new RuntimeException("Unhandled exception coalescing arrays", e);
							}
							break;
						case Collection:
							Collection currCollection;
							Collection lastCollection;
							try {
								currCollection = (Collection)field.get(currResult);
								lastCollection = (Collection)field.get(lastResult);
								if (CollectionUtils.isEmpty(lastCollection)) {
									lastCollection = currCollection;
								}
								else if (CollectionUtils.isEmpty(currCollection)) {
								}
								else {
									lastCollection = CollectionUtils.union(lastCollection, currCollection);
								}
								//TODO: This collapsing of repeating data will fail if item doesn't have an equals() method.
							}
							catch (IllegalAccessException e) {
								throw new RuntimeException("Unhandled exception coalescing collections", e);
							}
							break;
						case Map:
							//TODO: do
							break;
					}
				}

*/
			}
		}
		return uniqueResults;
	}
}
