package com.portlandwebworks.commons.hibernate;

import com.portlandwebworks.commons.dao.*;
import com.portlandwebworks.commons.data.EntityMapping;
import com.portlandwebworks.commons.data.FieldMapping;
import com.portlandwebworks.commons.domain.EntityBase;
import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.reflection.FieldUtils;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;
import org.hibernate.*;
import org.hibernate.Query;
import org.hibernate.annotations.Cascade;
import org.hibernate.criterion.*;
import org.hibernate.exception.ConstraintViolationException;
import org.hibernate.exception.GenericJDBCException;
import org.hibernate.impl.CriteriaImpl;
import org.hibernate.impl.SessionFactoryImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.orm.hibernate3.SessionFactoryUtils;

import javax.persistence.*;
import java.io.Serializable;
import java.lang.reflect.*;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @author rnorthup
 */
public class HibernateEntityManager implements IEntityManager
{
	private static Logger logger = LoggerFactory.getLogger(HibernateEntityManager.class);

	private SessionFactory sessionFactory;

/*
	private Map<Class<?>, List<CriteriaImpl.Subcriteria>> projectionSubcriteria =
			new HashMap<Class<?>, List<CriteriaImpl.Subcriteria>>();
	{
		projectionSubcriteria = LazyMap.decorate(projectionSubcriteria, new Factory()
		{
			public Object create()
			{
				return new ArrayList<CriteriaImpl.Subcriteria>();
			}
		});
	}
*/

	private boolean flushWrites;

	/**
	 * This allows returning a subclass of PagedSearchResult.
	 */
	private Class<PagedSearchResult> pagedSearchResultType = PagedSearchResult.class;

	/**
	 *
	 */
	public HibernateEntityManager()
	{
	}

	/**
	 * This session factory is used by the DAO for all database access and should be set by the container at runtime.
	 *
	 * @param sessionFactory
	 */
	@Autowired(required = true)
	public void setSessionFactory(SessionFactory sessionFactory)
	{
		this.sessionFactory = sessionFactory;
	}

	public void setPagedSearchResultType(Class<PagedSearchResult> pagedSearchResultType)
	{
		this.pagedSearchResultType = pagedSearchResultType;
	}

	public void setFlushWrites(boolean flushWrites)
	{
		this.flushWrites = flushWrites;
	}

	/**
	 * Utility method to retrieve the default schema configured for the underlying Hibernate session.
	 *
	 * @return	The string name of the schema, or <code>null</code> if no default schema is configured.
	 */
	protected String getDefaultSchemaName()
	{
		// This interface doesn't have the necessary api.  Need to downcast to the impl.
		SessionFactory factory = getSession().getSessionFactory();
		if (factory instanceof SessionFactoryImpl) {
			SessionFactoryImpl impl = (SessionFactoryImpl)factory;
			return impl.getSettings().getDefaultSchemaName();
		}
		return null;
	}

	/**
	 * Provides DAO subclasses with the ability to access the underlying hibernate session.
	 *
	 * @return
	 */
	protected Session getSession()
	{
		Session session = SessionFactoryUtils.getSession(sessionFactory, true);
		// TODO: This doesn't seem to do anything!!
		if (this.flushWrites) {
			session.setFlushMode(FlushMode.COMMIT);
		}
		return session;
	}


	/**
	 * Obtains the implementation of Criteria from the instance passed to this method.  Useful for
	 * @param query
	 * @return
	 */
	private CriteriaImpl downcastCriteria(Criteria query)
	{
		CriteriaImpl impl;
		if (query instanceof CriteriaImpl) {
			impl = (CriteriaImpl)query;
		}
		else if (query instanceof DetachedCriteria) {
			impl = (CriteriaImpl)((DetachedCriteria)query).getExecutableCriteria(getSession());
		}
		else if (query instanceof DetachedCriteriaAdapter) {
			impl = (CriteriaImpl)((DetachedCriteriaAdapter)query).getExecutableCriteria(getSession());
		}
		else {
			throw new RuntimeException("Could not downcast Criteria to a known subclass");
		}
		return impl;
	}


	/**
	 * Exposes the Hibernate session to the user of the API as a last resort to accomplish tasks not supported by the
	 * interface.
	 *
	 * @return
	 */
	public Object getPersistenceContext()
	{
		return getSession();
	}

	/**
	 * 
	 * @param <T>
	 * @return
	 */
	protected <T> PagedSearchResult<T> buildPagedSearchResult()
	{
		try {
			return pagedSearchResultType.newInstance();
		}
		catch (Exception e) {
			throw new RuntimeException("Unable to create instance of custom PagedSearchResult implementation " + pagedSearchResultType.toString());
		}
	}


	@SuppressWarnings("unchecked")
	public <T> T findById(final Class<T> returnType, final Serializable id, Object... fetchOptions)
			throws DataNotFoundException, IllegalArgumentException
	{
		if (id == null) {
			throw new IllegalArgumentException("id is null");
		}

		// Check for a composite primary key on this return type.  Hibernate does not gracefully handle the situation 
		// where one of the primary key values is null, so validate that now.
		if (HibernateUtils.isIdentifiable(returnType) && id.getClass().getAnnotation(Embeddable.class) != null) {
			Field[] fields = FieldUtils.getFieldsWithAnnotation(id.getClass(), Column.class);
			for (Field field : fields) {
				try {
					if (field.get(id) == null) {
						throw new IllegalArgumentException("Composite key has null value " + id.getClass().getName() +
								"." + field.getName());
					}
				}
				catch (IllegalAccessException e) {
					throw new IllegalArgumentException("Can't access composite key field " + field.getName());
				}
			}
		}

		QueryExecutor<T> addIdRestrictionAndGetUnique = new BasicQueryExecutor<T>(returnType) {
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = super.prepareQuery();

				// Apply the unique identifier restriction. By contract, all entities have a field named "id"
				query.add(Restrictions.idEq(id));
				
				return query;
			}

			@Override
			public T execute(Criteria query)
			{
				try {
					T entity = (T)query.uniqueResult();

					// Don't allow returning a null value when querying by unique identifier.
					if (entity == null) {
						throw new DataNotFoundException(returnType, id);
					}

					return entity;
				}
				catch (StaleStateException e) {
					throw new DataNotFoundException(returnType, id);
				}
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {
			return doWithEagerFetch(returnType, addIdRestrictionAndGetUnique, fetchOptions);
		}
		else {
			return doWithProjection(returnType, addIdRestrictionAndGetUnique);
		}
	}

	@SuppressWarnings({"unchecked"})
	public <T> List<T> findAll(final Class<T> returnType, Object... fetchOptions)
	{
		final ListQueryExecutor<T> applySortAndList = new BasicListQueryExecutor<T>(returnType)
		{
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = super.prepareQuery();
				HibernateUtils.applySorting(returnType, (CriteriaImpl)query);
				return query;
			}

			@Override
			public List<T> executeList(Criteria query)
			{
				query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
				return super.executeList(query);				
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {
			return doAllWithEagerFetch(returnType, applySortAndList, fetchOptions);
		}
		else {
			return doAllWithProjection(returnType, applySortAndList);
		}
	}

	/**
	 * Supports preparing queries and executing them.
	 * 
	 * @param <R>
	 */
	public interface QueryExecutor<R> {

		/**
		 * This method should build a Criteria instance and apply all necessary predicates, eager fetches, sorting and
		 * paging.
		 * 
		 * @return
		 */
		Criteria prepareQuery();

		/**
		 * Allows custom handling of query execution when looking for a list of results.  This method should be used to
		 * set a result transformer for the query, if necessary, because some Criteria APIs will reset a projection that
		 * is set during prepareQuery().
		 * 
		 * @param query
		 * @return
		 */
		R execute(Criteria query);
	}
		
	public interface ListQueryExecutor<R> {

		/**
		 * This method should build a Criteria instance and apply all necessary predicates, eager fetches, sorting and
		 * paging.
		 *
		 * @return
		 */
		Criteria prepareQuery();

		/**
		 * Allows custom handling of query execution when looking for a list of results.  This method should be used to
		 * set a result transformer for the query, if necessary, because some Criteria APIs will reset a projection that
		 * is set during prepareQuery().
		 * 
		 * @param query
		 * @return
		 */
		List<R> executeList(Criteria query);
	}
	
	public class BasicQueryExecutor<R> implements QueryExecutor<R> {
		
		private Class<?> type;
		
		public BasicQueryExecutor(Class<?> type)
		{
			this.type = type;
		}
		
		public Criteria prepareQuery()
		{
			return createEntityQuery(type);
		}
		
		public R execute(Criteria query)
		{
			return (R)query.uniqueResult();
		}
	}

	public class BasicListQueryExecutor<R> implements ListQueryExecutor<R> {

		private Class<?> type;

		public BasicListQueryExecutor(Class<?> type)
		{
			this.type = type;
		}

		public Criteria prepareQuery()
		{
			return createEntityQuery(type);
		}

		public List<R> executeList(Criteria query)
		{
			return (List<R>)query.list();
		}
	}

	/**
	 * This helper method allows a database query that is to be executed in a <code>Closure</code> to be executed with
	 * its return values being projected.
	 *
	 *
	 * @param returnType
	 * @param executor
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected <T> T doWithProjection(Class<T> returnType, QueryExecutor executor) {
		
		CriteriaImpl query = (CriteriaImpl)executor.prepareQuery();

		ProjectionList projectionList = Projections.projectionList();

		// Group all of the projections on the return type by entity.  Root entity projections will be stored under
		// the null-valued key.
		Map<String, List<FieldProjection>> projectionMap = HibernateUtils.groupProjectionsByEntity(returnType);

		// Grab all projections on the root object and add them to the query.
		List<FieldProjection> rootLevelProjections = projectionMap.remove(null);
		for (FieldProjection fieldProjection : rootLevelProjections) {
			String property = HibernateUtils.getPropertyAlias(query, fieldProjection.getEntityPath());
			projectionList.add(new CustomPropertyAliasProjection(property, fieldProjection.getBeanPath()));
		}
		
		if (! projectionMap.isEmpty()) {
	
			// Grab one nested entity collection and add that to the query as well.  Not wise to join more than one 1-M
			// relationship as it would create too many result rows with mostly redundant data.
	        for (String key : projectionMap.keySet()) {
		        List<FieldProjection> projectionsForFirstNestedEntity = projectionMap.remove(key);
	            for (FieldProjection fieldProjection : projectionsForFirstNestedEntity) {
		            String property = HibernateUtils.getPropertyAlias(query, fieldProjection.getEntityPath());
		            projectionList.add(new CustomPropertyAliasProjection(property, fieldProjection.getBeanPath()));
	            }
		        break;
	        }
		}
		
		// Apply the projections selected so far
		query.setProjection(Projections.distinct(projectionList));

		// This will ensure that in the event the query becomes a cartesian product, hibernate collapses the results into
		// one row, unique by primary key
		query.setResultTransformer(new AliasToBeanGraphResultTransformer(returnType));

		T entity = (T)executor.execute(query);

		if (entity != null) {
			
			// Check for any remaining projections on nested entities that need to be retrieved that were deferred 
			// until after the root entity query.  If any are found, execute a new query for each.
			for (String key : projectionMap.keySet()) {
				List<FieldProjection> nestedProjections = projectionMap.get(key);
				ProjectionList nestedList = Projections.projectionList();

				// Have query executor build us a fresh query.
				CriteriaImpl another = (CriteriaImpl)executor.prepareQuery();

				for (FieldProjection fieldProjection : nestedProjections) {
					String property = HibernateUtils.getPropertyAlias(another, fieldProjection.getEntityPath());
					nestedList.add(new CustomPropertyAliasProjection(property, fieldProjection.getBeanPath()));
				}

				another.setProjection(Projections.distinct(nestedList));
				another.setResultTransformer(new AliasToBeanGraphResultTransformer(returnType));
				
				// Execute query with new projections applied.
				T temp = (T)executor.execute(another);

				// Copy the nested field we just queried for into the object generated by the root entity query.
				Object nestedValue = FieldUtils.getFieldValue(temp, key);
				FieldUtils.setFieldValue(entity, key, nestedValue);
			}
		}
		
		return entity;
	}

	/**
	 * Wraps a hibernate query execution with a projection of the values exposed as entity-mapped fields in the return
	 * type specified.
	 *
	 *
	 * @param returnType    asdf.
	 * @param executor      asdf.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected <T> List<T> doAllWithProjection(Class<T> returnType, ListQueryExecutor<T> executor) {

		CriteriaImpl query = (CriteriaImpl)executor.prepareQuery();

		ProjectionList projectionList = Projections.projectionList();

		// Group all of the projections on the return type by entity.  Root entity projections will be stored under
		// the null-valued key.
		Map<String, List<FieldProjection>> projectionMap = HibernateUtils.groupProjectionsByEntity(returnType);

		// Grab all projections on the root object and add them to the query.
		List<FieldProjection> rootLevelProjections = projectionMap.remove(null);
		for (FieldProjection fieldProjection : rootLevelProjections) {
			String property = HibernateUtils.getPropertyAlias(query, fieldProjection.getEntityPath());
			projectionList.add(new CustomPropertyAliasProjection(property, fieldProjection.getBeanPath()));
		}

		if (! projectionMap.isEmpty()) {

			// Grab one nested entity collection and add that to the query as well.  Not wise to join more than one 1-M
			// relationship as it would create too many result rows with mostly redundant data.
			for (String key : projectionMap.keySet()) {
				List<FieldProjection> projectionsForFirstNestedEntity = projectionMap.remove(key);
				for (FieldProjection fieldProjection : projectionsForFirstNestedEntity) {
					String property = HibernateUtils.getPropertyAlias(query, fieldProjection.getEntityPath());
					projectionList.add(new CustomPropertyAliasProjection(property, fieldProjection.getBeanPath()));
				}
				break;
			}
		}

		// Apply the projections selected so far
		query.setProjection(Projections.distinct(projectionList));

		// This will ensure that in the event the query becomes a cartesian product, hibernate collapses the results into
		// one row, unique by primary key
		query.setResultTransformer(new AliasToBeanGraphResultTransformer(returnType));

		List<T> results = executor.executeList(query);

		if (results.size() > 0 && projectionMap.size() > 0) {
			
			Field[] idFields = FieldUtils.getIdentifierFieldsForClass(results.get(0).getClass());
			if (idFields.length > 1) {
				throw new RuntimeException("Can't do this. Sorry!");
			}
			
			Map<Object, T> mappedResults = new HashMap<Object, T>();
			for (T result : results) {
				Object id = FieldUtils.getFieldValue(result, idFields[0]);
				mappedResults.put(id, result);
			}
			
			FieldMapping idMapping = idFields[0].getAnnotation(FieldMapping.class);

			for (String key : projectionMap.keySet()) {
				List<FieldProjection> nestedProjections = projectionMap.get(key);
				ProjectionList nestedList = Projections.projectionList();
				
				CriteriaImpl another = (CriteriaImpl)executor.prepareQuery();

				//todo; add projection for parent entity's unique ID
				nestedList.add(new CustomPropertyAliasProjection(idMapping.value(), idFields[0].getName()));

				for (FieldProjection fieldProjection : nestedProjections) {
					String property = HibernateUtils.getPropertyAlias(another, fieldProjection.getEntityPath());
					nestedList.add(new CustomPropertyAliasProjection(property, fieldProjection.getBeanPath()));
				}

				another.setProjection(Projections.distinct(nestedList));
				another.setResultTransformer(new AliasToBeanGraphResultTransformer(returnType));
				
				List<T> temps = executor.executeList(another);
				for (T temp : temps) {

					// Copy the nested field we just queried for into the object generated by the root entity query.
					Object nestedValue = FieldUtils.getFieldValue(temp, key);
					Object parentId = FieldUtils.getFieldValue(temp, idFields[0]);
					Object entity = mappedResults.get(parentId);
					FieldUtils.setFieldValue(entity, key, nestedValue);
				}
			}
		}

		return results;
	}

	/**
	 *
	 *
	 * @param returnType
	 * @param executor
	 * @param fetchOptions
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected <T> T doWithEagerFetch(Class<T> returnType, QueryExecutor executor, Object... fetchOptions) {

		CriteriaImpl query = (CriteriaImpl)executor.prepareQuery();
		
		List deferred = HibernateUtils.applyEagerFetchOptions(returnType, query, fetchOptions);

		Object entity = executor.execute(query);

		if (! CollectionUtils.isEmpty(deferred)) {

			HibernateUtils.loadDeferredFetchOptions(query, entity, deferred.toArray());
		}

		return (T)entity;
	}

	/**
	 *
	 *
	 * @param returnType
	 * @param executor
	 * @param fetchOptions
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected <T> List<T> doAllWithEagerFetch(Class<T> returnType, ListQueryExecutor executor, Object... fetchOptions) {

		CriteriaImpl query = (CriteriaImpl)executor.prepareQuery();

		List deferred = HibernateUtils.applyEagerFetchOptions(returnType, query, fetchOptions);

		List<T> entities = (List<T>)executor.executeList(query);

		if (! CollectionUtils.isEmpty(deferred)) {

			for (Object entity : entities) {
				HibernateUtils.loadDeferredFetchOptions(query, entity, deferred.toArray());
			}
		}

		return entities;
	}

	@SuppressWarnings("unchecked")
	public <T, K extends Serializable> Map<K, T> mapAll(final Class<T> returnType, final String keyField, Object... fetchOptions)
	{
		List<T> results;
		if (HibernateUtils.isIdentifiable(returnType)) {
			results = doAllWithEagerFetch(returnType, new BasicListQueryExecutor<T>(returnType), fetchOptions);
		}
		else {
			results = doAllWithProjection(returnType, new BasicListQueryExecutor<T>(returnType));
		}

		Map<K, T> resultMap = new HashMap<K, T>();
		try {
			for (T result : results) {
				resultMap.put((K)PropertyUtils.getProperty(result, keyField), result);
			}
		}
		catch (Exception e) {
			throw new RuntimeException("Ack!", e);
		}

		return resultMap;
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> findAllByIds(final Class<T> returnType, final Serializable[] ids, Object... fetchOptions) throws DataNotFoundException
	{
		ListQueryExecutor<T> addIdRestrictionAndList = new BasicListQueryExecutor<T>(returnType) {
			
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = super.prepareQuery();
				query.add(Restrictions.in("id", ids));
				return query;
			}

			@Override
			public List<T> executeList(Criteria query) {
				query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

				List list = query.list();

				if (list.size() < ids.length) {
					// Returned entity list size is less than size of requested ID array.  This means that at least one ID does
					// not represent a persisted entity.  Report an error by identifying the first ID that doesn't match the
					// ID of any returned entity.
					for (Serializable id : ids) {
						boolean match = false;
						for (Object obj : list) {
							Identifiable entity = (Identifiable)obj;
							if (entity.getId().equals(id)) {
								match = true;
								break;
							}
						}
						if (! match) {
							throw new DataNotFoundException(returnType, id);
						}
					}
				}
				return list;
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {
			return doAllWithEagerFetch(returnType, addIdRestrictionAndList, fetchOptions);
		}
		else {
			return doAllWithProjection(returnType, addIdRestrictionAndList);
		}
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> findAllByCriteria(final Class<T> returnType, final SearchCriteria searchCriteria, Object... fetchOptions)
	{
		ListQueryExecutor<T> applySearchCriteriaSortAndList = new BasicListQueryExecutor<T>(returnType)
		{
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = createEntityQuery(returnType);

				HibernateUtils.applySearchCriteria(searchCriteria, (CriteriaImpl)query);

				// Even though no sort options are specified
				if (PagedSearchOptions.class.isAssignableFrom(searchCriteria.getClass())) {
					HibernateUtils.applySorting(returnType, (CriteriaImpl)query, ((PagedSearchOptions)searchCriteria).getSort());
				}
				else {
					HibernateUtils.applySorting(returnType, (CriteriaImpl)query);
				}

				return query;
			}

			@Override
			public List<T> executeList(Criteria query)
			{
				if (HibernateUtils.isIdentifiable(returnType)) {
					// Applying predicates may cause joins to occur if a predicate spans entity relationships.
					query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
				}

				return query.list();
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {

			return doAllWithEagerFetch(returnType, applySearchCriteriaSortAndList, fetchOptions);
		}
		else {
			return doAllWithProjection(returnType, applySearchCriteriaSortAndList);
		}
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> findAllByPredicate(final Class<T> returnType, final Predicate... predicates)
	{
		ListQueryExecutor<T> applyPredicatesSortAndList = new BasicListQueryExecutor<T>(returnType) {
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = super.prepareQuery();

				HibernateUtils.applyPredicates((CriteriaImpl)query, Arrays.asList(predicates));

				HibernateUtils.applySorting(returnType, (CriteriaImpl)query);

				return query;
			}
			
			@Override
			public List<T> executeList(Criteria query) {

				// Applying predicates may cause joins to occur if a predicate spans entity relationships.
				if (HibernateUtils.isIdentifiable(returnType)) {
					query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
				}
				
				return query.list();
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {
			return doAllWithEagerFetch(returnType, applyPredicatesSortAndList);
		}
		else {
			return doAllWithProjection(returnType, applyPredicatesSortAndList);
		}
	}

	public <T> T projectByPredicate(Class<? extends Identifiable> rootEntity, String field, Predicate... predicates)
	{
		CriteriaImpl query = (CriteriaImpl)getSession().createCriteria(rootEntity);

		String property = HibernateUtils.getPropertyAlias(query, field);
		query.setProjection(new CustomPropertyAliasProjection(property, field));

		HibernateUtils.applyPredicates(query, Arrays.asList(predicates));

		HibernateUtils.applySorting(rootEntity, query);

		return (T)query.uniqueResult();
	}

	public <T> List<T> projectAllByPredicate(final Class<? extends Identifiable> rootEntity, String field, Predicate... predicates)
	{
		CriteriaImpl query = (CriteriaImpl)getSession().createCriteria(rootEntity);

		String property = HibernateUtils.getPropertyAlias(query, field);
		query.setProjection(new CustomPropertyAliasProjection(property, field));

		HibernateUtils.applyPredicates(query, Arrays.asList(predicates));

		HibernateUtils.applySorting(rootEntity, query);
		
		return query.list();
	}

	public Map projectByPredicate(final Class<? extends Identifiable> rootEntity, String[] fields, Predicate... predicates)
	{
		QueryExecutor<Map> applyPredicates = new BasicQueryExecutor<Map>(rootEntity) {
			@Override
			public Criteria prepareQuery()
			{
				return getSession().createCriteria(rootEntity);
			}
			
			@Override
			public Map execute(Criteria query)
			{
				query.setResultTransformer(Criteria.ALIAS_TO_ENTITY_MAP);
				return (Map)query.uniqueResult();
			}
		};
		
		return doWithProjection(Map.class, applyPredicates);
	}

	@SuppressWarnings("unchecked")
	public List<Map> projectAllByPredicate(final Class<? extends Identifiable> rootEntity, String[] fields, Predicate... predicates)
	{
		ListQueryExecutor<Map> applyPredicatesSortAndList = new BasicListQueryExecutor<Map>(rootEntity) {
			@Override
			public Criteria prepareQuery()
			{
				return getSession().createCriteria(rootEntity);
			}
			
			@Override
			public List<Map> executeList(Criteria query)
			{
				query.setResultTransformer(Criteria.ALIAS_TO_ENTITY_MAP);
				
				return query.list();
			}
		};
		
		return doAllWithProjection(Map.class, applyPredicatesSortAndList);
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> findAllByPredicate(final Class<T> returnType, final SortSpecification sort, final Predicate... predicates)
	{
		ListQueryExecutor<T> applyPredicatesSortAndList = new BasicListQueryExecutor<T>(returnType) {
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = super.prepareQuery();

				HibernateUtils.applyPredicates((CriteriaImpl)query, Arrays.asList(predicates));

				HibernateUtils.applySorting(returnType, (CriteriaImpl)query);

				return query;
			}
			
			@Override
			public List<T> executeList(Criteria query) {
				// Applying predicates may cause joins to occur if a predicate spans entity relationships.
				if (HibernateUtils.isIdentifiable(returnType)) {
					query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
				}
				
				return query.list();
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {
			return doAllWithEagerFetch(returnType, applyPredicatesSortAndList);
		}
		else {
			return doAllWithProjection(returnType, applyPredicatesSortAndList);
		}
	}

	@SuppressWarnings("unchecked")
	public <T> T findByPredicate(final Class<T> returnType, final Predicate... predicates) throws DataNotFoundException
	{
		QueryExecutor<T> applyPredicatesGetUnique = new BasicQueryExecutor<T>(returnType) {
			@Override
			public Criteria prepareQuery()
			{
				Criteria query = super.prepareQuery();
				HibernateUtils.applyPredicates((CriteriaImpl)query, Arrays.asList(predicates));
				return query;
			}
		};

		if (HibernateUtils.isIdentifiable(returnType)) {
			return applyPredicatesGetUnique.execute(applyPredicatesGetUnique.prepareQuery());
		}
		else {
			return doWithProjection(returnType, applyPredicatesGetUnique);
		}
	}

	/**
	 *
	 * @param returnType    todo
	 * @param <T>           todo
	 * @return
	 */
	protected <T> CriteriaImpl createEntityQuery(Class<T> returnType)
	{
		Class<? extends Identifiable> entityType = getEntityType(returnType);
		return (CriteriaImpl)getSession().createCriteria(entityType);
	}

	/**
	 *
	 * @param clazz     todo
	 * @return
	 */
	public Class<? extends Identifiable> getEntityType(Class<?> clazz)
	{
		Class<Identifiable> entityType;
		if (clazz.getAnnotation(Entity.class) != null) {
			if (Identifiable.class.isAssignableFrom(clazz)) {
				entityType = (Class<Identifiable>)clazz;
			}
			else {
				throw new RuntimeException(String.format("Cannot execute query with return type %s.  It does not implement the Identifiable interface", clazz.getName()));
			}
		}
		else {
			EntityMapping entityMapping = clazz.getAnnotation(EntityMapping.class);
			if (entityMapping == null) {
				throw new RuntimeException(
						String.format("Cannot execute query with return type %s.  It must have either an @Entity or @EntityMapping annotation", clazz.getName()));
			}

			try {
				return (Class<? extends Identifiable>)Class.forName(entityMapping.value());
			}
			catch (ClassNotFoundException e) {
				throw new RuntimeException("Mapped entity class not found: " + entityMapping.value());
			}
			catch (ClassCastException e) {
				throw new RuntimeException("Mapped entity class does not implement the IEntity interface");
			}
		}
		return entityType;
	}

	public <T> PagedSearchResult<T> superSearch(final Class<T> returnType, final SearchCriteria searchCriteria, final IPagedSearchOptions displayOptions, final Object... fetchOptions)
	{
		final PagedSearchResult<T> results = buildPagedSearchResult();

		final Class<? extends Identifiable> entityClass = getEntityType(returnType);

		final CriteriaImpl resultsQuery = createEntityQuery(entityClass);

		// TODO: for select count, project root entity's ID and sort fields, grouping by ID
		final CriteriaImpl countQuery = createEntityQuery(entityClass);

		// TODO: for deep nested filter query, project only ID and sort fields
		// This is the ideal way to run this query, which assumes a single value primary key.
		DetachedCriteria filterSubquery = DetachedCriteria.forClass(entityClass);

		DetachedCriteriaAdapter queryAdapter = new DetachedCriteriaAdapter(filterSubquery);
		HibernateUtils.applySearchCriteria(searchCriteria, (CriteriaImpl)queryAdapter.getExecutableCriteria(getSession()));

		ProjectionList projectionList = Projections.projectionList();
		projectionList.add(Projections.id());
		for (SortSpecification sortSpecification : displayOptions.getSort()) {
			projectionList.add(Projections.max(sortSpecification.getProperty()));
		}
		filterSubquery.setProjection(projectionList);

		// TODO: for nested paging query, project only ID
		
		// TODO: for the outer query, join paging query on ID and apply fetch options

		return null;
	}


	public <T> PagedSearchResult<T> search(final Class<T> returnType, final SearchCriteria searchCriteria, final IPagedSearchOptions displayOptions, final Object... fetchOptions)
	{
		final PagedSearchResult<T> results = buildPagedSearchResult();

		boolean hasEmbeddedId = null != FieldUtils.getFieldWithAnnotation(returnType, EmbeddedId.class);

		ListQueryExecutor<T> doSearch;
		if (displayOptions != null && displayOptions.getLimit() != Integer.MAX_VALUE) {

			final long totalItemCount = getCount(returnType, searchCriteria);

			results.setTotalItemCount(totalItemCount);

			if (hasEmbeddedId) {

				/**
				 * This search algorithm is necessary if paged data is desired, but the presence of an embedded ID
				 * prevents use of the preferred method of querying, which is to do filtering on a subquery and then
				 * do an outer query
				 */
				doSearch = new BasicListQueryExecutor<T>(returnType)
				{
					@Override
					public Criteria prepareQuery()
					{
						Criteria query = super.prepareQuery();

						HibernateUtils.applySearchCriteria(searchCriteria, (CriteriaImpl)query);

						HibernateUtils.applySorting(returnType, (CriteriaImpl)query, displayOptions.getSort());

						if (totalItemCount > displayOptions.getLimit()) {
							HibernateUtils.applyPaging(query, displayOptions);
						}

						return query;
					}
				};
			}
			else {
				/**
				 * This search algorithm uses a subquery to filter query results, and then joins those subquery results
				 * to the root query.  The reason for doing this is that when an entity has a multi-value field in it
				 * and the query results come back as a cartesian product, hibernate seems to be in
				 */
				doSearch = new BasicListQueryExecutor<T>(returnType)
				{
					@Override
					public Criteria prepareQuery()
					{
						Criteria query = super.prepareQuery();
						
						// This is the ideal way to run this query, which assumes a single value primary key.
						DetachedCriteria subQuery = DetachedCriteria.forClass(getEntityType(returnType));

						DetachedCriteriaAdapter queryAdapter = new DetachedCriteriaAdapter(subQuery);
						HibernateUtils.applySearchCriteria(searchCriteria, (CriteriaImpl)queryAdapter.getExecutableCriteria(getSession()));

						//TODO: include sort properties in this projection.
						subQuery.setProjection(
								Projections.distinct(Projections.property("id"))
						);

						//TODO: nest the subquery in another subquery that does the paging, and ha
						//DetachedCriteria pagingQuery = DetachedCriteria.forClass(returnType);

						query.add(Subqueries.propertyIn("id", subQuery));

						if (totalItemCount > displayOptions.getLimit()) {
							HibernateUtils.applyPaging(query, displayOptions);
						}

						HibernateUtils.applySorting(returnType, (CriteriaImpl)query, displayOptions.getSort());
						
						return query;
					}
					
					@Override
					public List<T> executeList(Criteria query)
					{
						if (HibernateUtils.isIdentifiable(returnType)) {
							query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
						}
						return query.list();
					}
				};
			}

		}
		else {
			doSearch = new BasicListQueryExecutor<T>(returnType)
			{
				@Override
				public Criteria prepareQuery()
				{
					Criteria query = super.prepareQuery();

					// The simple scenario--no paging.  Returning all results.  No count query necessary.  Apply the search
					// criteria and the sorting and get all results.
					HibernateUtils.applySearchCriteria(searchCriteria, (CriteriaImpl)query);

					if (displayOptions != null) {
						HibernateUtils.applySorting(returnType, (CriteriaImpl)query, displayOptions.getSort());
					}

					return query;
				}
				
				@Override
				public List<T> executeList(Criteria resultsQuery)
				{
					// If a 1-M relationship was eager fetched, make sure that hibernate collapses the results into a set
					// of unique entities.
					if (HibernateUtils.isIdentifiable(returnType)) {
						resultsQuery.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);
					}

					List<T> items = resultsQuery.list();

					results.setTotalItemCount(items.size());

					return items;
				}
			};
		}

		// Finally, let's run the search query using the appropriate closure and enclosing logic
		List<T> resultItems;
		if (HibernateUtils.isIdentifiable(returnType)) {

			if (hasEmbeddedId) {
				// Suppress eager fetching
				if (! ArrayUtils.isEmpty(fetchOptions)) {
					logger.warn("Suppressing eager fetch of fields {} because we're dealing with");
				}
				resultItems = doAllWithEagerFetch(returnType, doSearch);
			}
			else {
				resultItems = doAllWithEagerFetch(returnType, doSearch, fetchOptions);
			}
		}
		else {
			resultItems = doAllWithProjection(returnType, doSearch);
		}

		results.setItems(resultItems);

		return results;
	}

	public long getCount(Class<?> entityClass, SearchCriteria searchCriteria)
	{
		// handle paging of results
		CriteriaImpl countQuery = createEntityQuery(entityClass);

		HibernateUtils.applySearchCriteria(searchCriteria, countQuery);

		// Return a count, not the actual entity list.
		Projection countProjection;

		// Project a count of unique IDs, taking into account composite keys, which in hibernate are represented as
		// embedded IDs.
		Field embedded = FieldUtils.getFieldWithAnnotation(entityClass, EmbeddedId.class);
		if (embedded != null) {
			countProjection = HibernateUtils.createCompoundKeyCountProjection(embedded);
		}
		else {
			countProjection = Projections.countDistinct("id");
		}

		countQuery.setProjection(countProjection);

		// Save this value on the return object.
		Number count = (Number)countQuery.uniqueResult();

		return count.longValue();
	}

	@Deprecated
	public <T extends EntityBase, Y extends EntityBase> PagedSearchResult<T> search(Class<T> viewClass, Class<Y> criteriaClass, SearchCriteria searchCriteria, IPagedSearchOptions displayOptions, Object... fetchOptions)
	{
		final PagedSearchResult<T> results = buildPagedSearchResult();

		CriteriaImpl resultsQuery = null;

		// Eager fetch options that have to be applied after the search query.
		List deferred = null;

		// If display options indicate that the query results should be a page of partial results, run this query in
		// two phases.  First phase is to execute a count of distinct IDs using the same search criteria as the results
		// query.  Second phase is to perform the results query, with paging applied.  This allows the method to return
		// not only the page of results but the total number of matching records.
		if (displayOptions != null && displayOptions.getLimit() != Integer.MAX_VALUE) {

			CriteriaImpl countQuery = createEntityQuery(criteriaClass);

			HibernateUtils.applySearchCriteria(searchCriteria, countQuery);

			// Return a count, not the actual entity list.
			Projection countProjection;

			// If this entity has a field with an Embedded annotation on it, assume it is a composite key which requires
			// special handling to build the distinct count projection. TODO: is this a safe assumption?
			Field embedded = FieldUtils.getFieldWithAnnotation(criteriaClass, EmbeddedId.class);
			if (embedded != null) {
				countProjection = HibernateUtils.createCompoundKeyCountProjection(embedded);
			}
			else{
				countProjection = Projections.countDistinct("id");
			}

			countQuery.setProjection(countProjection);

			// Save this value on the return object.
			Number count = (Number)countQuery.uniqueResult();
			results.setTotalItemCount(count.intValue());

			// Now run the actual query, applying paging parameters.

			boolean entityHasEmbeddedId = FieldUtils.getFieldWithAnnotation(viewClass, EmbeddedId.class) == null ? false : true;

			//TODO: find a way to unify these two query approaches into one.  It shouldn't be this difficult to come up
			//TODO: with a distinct set of results for a hibernate query.  Perhaps add all entity fields to a projection
			//TODO: list with a distinct projection wrapper around it?
			if (entityHasEmbeddedId || (! FieldUtils.hasMultiValueField(viewClass) && viewClass.equals(criteriaClass))) {
				// This is the less than ideal way to run this query, necessary because the "ideal" way doesn't work
				// with a composite primary key.
				resultsQuery = createEntityQuery(viewClass);

				deferred = HibernateUtils.applyEagerFetchOptions(viewClass, resultsQuery, fetchOptions);

				HibernateUtils.applySearchCriteria(searchCriteria, resultsQuery);

				HibernateUtils.applySorting(viewClass, resultsQuery, displayOptions.getSort());

				HibernateUtils.applyPaging(resultsQuery, displayOptions);

				// DON'T USE THIS API WITH PAGING--THE LIST TRIMMING FOR DISTINCT AFTER DB CALL, THROWING OFF PAGING.
				//resultsQuery.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

				results.setItems(resultsQuery.list());
			}
			else {
				// This is the ideal way to run this query, which assumes a single value primary key.
				DetachedCriteria criteriaQuery = DetachedCriteria.forClass(criteriaClass);
				DetachedCriteriaAdapter queryAdapter = new DetachedCriteriaAdapter(criteriaQuery);
				HibernateUtils.applySearchCriteria(searchCriteria, (CriteriaImpl)queryAdapter.getExecutableCriteria(getSession()));

				resultsQuery.setProjection(
						Projections.distinct(Projections.property("id"))
				);

				resultsQuery = createEntityQuery(viewClass);
				deferred = HibernateUtils.applyEagerFetchOptions(viewClass, resultsQuery, fetchOptions);
				resultsQuery.add(Subqueries.propertyIn("id", criteriaQuery));
				HibernateUtils.applyPaging(resultsQuery, displayOptions);
				HibernateUtils.applySorting(viewClass, resultsQuery, displayOptions.getSort());

				resultsQuery.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

				results.setItems(resultsQuery.list());
			}

			logger.debug("Total items matching criteria: {}", results.getTotalItemCount());
		}
		else {
			// The simple scenario--no paging.  Returning all results.  No count query necessary.  Apply the search
			// criteria and the sorting and get all results.
			resultsQuery = createEntityQuery(viewClass);

			deferred = HibernateUtils.applyEagerFetchOptions(viewClass, resultsQuery, fetchOptions);

			HibernateUtils.applySearchCriteria(searchCriteria, resultsQuery);
			
			HibernateUtils.applySorting(viewClass, resultsQuery, displayOptions.getSort());

			resultsQuery.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

			results.setItems(resultsQuery.list());

			results.setTotalItemCount(results.getItems().size());
		}

		if (! CollectionUtils.isEmpty(deferred)) {
			for (T item : results.getItems()) {
				HibernateUtils.loadDeferredFetchOptions(resultsQuery, item, deferred.toArray());
			}
		}

		return results;
	}

	public <T extends Identifiable> Serializable insert(T entity) throws DuplicateDataException
	{
		try {
			getSession().save(entity);

			if (flushWrites) {
				getSession().flush();
			}
		}
		catch (ConstraintViolationException cve) {

			//TODO: not all constraints are unique constraints.  perhaps leverage the dialect to interpret the constraint
			throw new DuplicateDataException(cve.getMessage(), cve);
		}

		return entity.getId();
	}

	public <T extends Identifiable> void saveOrUpdate(T entity) throws DuplicateDataException
	{
		try {
			getSession().saveOrUpdate(entity);

			if (flushWrites) {
				getSession().flush();
			}
		}
		catch (ConstraintViolationException cve) {
			//TODO: not all constraints are unique constraints.  perhaps leverage the dialect to interpret the constraint
			throw new DuplicateDataException(cve.getMessage(), cve);
		}
	}

	public <T extends Identifiable> void update(T entity) throws DuplicateDataException
	{
		if (entity == null) {
			throw new IllegalArgumentException("entity cannot be null");
		}

		try {
			getSession().update(entity);

			if (flushWrites) {
				getSession().flush();
			}
		}
		catch (StaleStateException e) {
			throw new DataNotFoundException(entity.getClass(), entity.getId());
		}
		catch (ConstraintViolationException cve) {
			//TODO: not all constraints are unique constraints.  perhaps leverage the dialect to interpret the constraint
			throw new DuplicateDataException(cve.getMessage(), cve);
		}
	}

	public <T extends Identifiable> void delete(T entity) throws DataIntegrityException, IllegalArgumentException, DataNotFoundException
	{
		if (entity == null) {
			throw new IllegalArgumentException("entity is null");
		}

		try {
			getSession().delete(entity);

			if (flushWrites) {
				getSession().flush();
			}
		}
		catch (TransientObjectException e) {
			throw new DataNotFoundException(entity.getClass(), entity.getId());
		}
		catch (ConstraintViolationException e) {
			throw new DataIntegrityException("Delete of entity " + entity + " failed, dependencies exist", e);
		}
		catch (PropertyValueException e) {
			// For some odd reason, hibernate wants to ensure that non-nullable fields have values on delete
			throw new DataNotFoundException(entity.getClass(), entity.getId());
		}
		catch (StaleStateException e) {
			throw new DataNotFoundException(entity.getClass(), entity.getId());
		}
		catch (GenericJDBCException e) {
			// For some reason when calling flush with batch mode enabled, this generic exception gets thrown for a
			// constraint violation.  Not sure how you're supposed to know what the real problem was.
			throw new DataIntegrityException("Could not delete entity " + entity + ", dependencies exist");
		}
	}

	public <T extends Identifiable> void deleteById(Class<T> entityClass, Serializable id) throws DataNotFoundException
	{
		// Deleting by primary key using HQL does not clean up nested data for which cascade rules were configured.
		// Check for any JPA collection annotations on the fields of the class to be deleted, and if any are found,
		// and they have cascading deletes, then execute HQL to do those cascade deletes first.
		Field[] nestedFields = FieldUtils.getFields(entityClass);
		for (Field nestedField : nestedFields) {

			ManyToMany m2m = nestedField.getAnnotation(ManyToMany.class);
			if (m2m != null) {

				// NOTE: I'm not checking cascade options before deleting join table entries because I think it's silly
				// to do so.
				String joinTableName = null;
				JoinTable jt = nestedField.getAnnotation(JoinTable.class);
				if (jt != null) {
					joinTableName = jt.name();
				}
				else {
					//TODO: build the join table name the same way hibernate would if no JoinTable annotation is present
					throw new RuntimeException("JoinTable annotation is required for cascading delete by ID");
				}

				StringBuilder subDeleteSql = new StringBuilder();
				subDeleteSql
						.append("delete from ")
						.append(jt.name())
						.append(" where ")
						//TODO: handle composite foreign key reference
						.append(jt.joinColumns()[0].name())
						.append(" = :id");

				Query subDelete = getSession().createSQLQuery(subDeleteSql.toString());
				subDelete.setParameter("id", id);
				subDelete.executeUpdate();

				continue;
			}
			OneToMany o2m = nestedField.getAnnotation(OneToMany.class);
			if (o2m != null) {
				boolean shouldCascadeDeletes = (ArrayUtils.contains(o2m.cascade(), javax.persistence.CascadeType.ALL) ||
						ArrayUtils.contains(o2m.cascade(), javax.persistence.CascadeType.REMOVE));

				Cascade cascade = nestedField.getAnnotation(Cascade.class);
				shouldCascadeDeletes |= (cascade != null && (
						ArrayUtils.contains(cascade.value(), org.hibernate.annotations.CascadeType.ALL) ||
								ArrayUtils.contains(cascade.value(), org.hibernate.annotations.CascadeType.DELETE) ||
								ArrayUtils.contains(cascade.value(), org.hibernate.annotations.CascadeType.REMOVE)
						));

				if (shouldCascadeDeletes) {
					nestedField.setAccessible(true);
					ParameterizedType paramType = (ParameterizedType)nestedField.getGenericType();
					Class collType = (Class)paramType.getActualTypeArguments()[0];
					String mappedBy = o2m.mappedBy();
					try {
						Field mappedField = collType.getDeclaredField(mappedBy);
						JoinColumn jc = mappedField.getAnnotation(JoinColumn.class);

						StringBuilder subDeleteHql = new StringBuilder();
						subDeleteHql
								.append("delete from ")
								.append(collType.getSimpleName())
								.append(" where ")
								//TODO: handle composite foreign key reference
								.append(jc.name())
								.append(" = :id");
						Query subdDelete = getSession().createQuery(subDeleteHql.toString());
						subdDelete.setParameter("id", id);
						subdDelete.executeUpdate();
					}
					catch (NoSuchFieldException e) {
						e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
					}
				}
				continue;
			}
		}

		StringBuilder deleteHql = new StringBuilder();
		deleteHql.append("delete from ");
		deleteHql.append(entityClass.getName()).append(" x");
		deleteHql.append(" where x.id = :id");

		Query delete = getSession().createQuery(deleteHql.toString());
		delete.setParameter("id", id);

		int count = delete.executeUpdate();

		if (count == 0) {
			throw new DataNotFoundException(entityClass, id);
		}
		
		if (flushWrites) {
			getSession().flush();
		}
	}

	public <T extends Identifiable> void deleteAllById(Class<T> entityClass, Collection<? extends Serializable> ids) throws DataNotFoundException
	{
		deleteAllById(entityClass, ids.toArray(new Serializable[ids.size()]));
	}

	public <T extends Identifiable> void deleteAllById(Class<T> entityClass, Serializable... ids) throws DataNotFoundException
	{
		// Deleting by primary key using HQL does not clean up nested data for which cascade rules were configured.
		// Check for any JPA collection annotations on the fields of the class to be deleted, and if any are found,
		// and they have cascading deletes, then execute HQL to do those cascade deletes first.
		Field[] nestedFields = FieldUtils.getFields(entityClass);
		for (Field nestedField : nestedFields) {

			ManyToMany m2m = nestedField.getAnnotation(ManyToMany.class);
			if (m2m != null) {

				// NOTE: I'm not checking cascade options before deleting join table entries because I think it's silly
				// to do so.
				String joinTableName = null;
				JoinTable jt = nestedField.getAnnotation(JoinTable.class);
				if (jt != null) {
					joinTableName = jt.name();
				}
				else {
					//TODO: build the join table name the same way hibernate would if no JoinTable annotation is present
					throw new RuntimeException("JoinTable annotation is required for cascading delete by ID");
				}

				StringBuilder subDeleteSql = new StringBuilder();
				subDeleteSql
						.append("delete from ")
						.append(jt.name())
						.append(" where ")
								//TODO: handle composite foreign key reference
						.append(jt.joinColumns()[0].name())
						.append(" in ( :ids )");

				Query subDelete = getSession().createSQLQuery(subDeleteSql.toString());
				subDelete.setParameterList("ids", ids);
				subDelete.executeUpdate();

				continue;
			}
			OneToMany o2m = nestedField.getAnnotation(OneToMany.class);
			if (o2m != null) {
				boolean shouldCascadeDeletes = (ArrayUtils.contains(o2m.cascade(), javax.persistence.CascadeType.ALL) ||
						ArrayUtils.contains(o2m.cascade(), javax.persistence.CascadeType.REMOVE));

				Cascade cascade = nestedField.getAnnotation(Cascade.class);
				shouldCascadeDeletes |= (cascade != null && (
						ArrayUtils.contains(cascade.value(), org.hibernate.annotations.CascadeType.ALL) ||
								ArrayUtils.contains(cascade.value(), org.hibernate.annotations.CascadeType.DELETE) ||
								ArrayUtils.contains(cascade.value(), org.hibernate.annotations.CascadeType.REMOVE)
				));

				if (shouldCascadeDeletes) {
					nestedField.setAccessible(true);
					ParameterizedType paramType = (ParameterizedType)nestedField.getGenericType();
					Class collType = (Class)paramType.getActualTypeArguments()[0];
					String mappedBy = o2m.mappedBy();
					try {
						Field mappedField = collType.getDeclaredField(mappedBy);
						JoinColumn jc = mappedField.getAnnotation(JoinColumn.class);

						StringBuilder subDeleteHql = new StringBuilder();
						subDeleteHql
								.append("delete from ")
								.append(collType.getSimpleName())
								.append(" where ")
										//TODO: handle composite foreign key reference
								.append(jc.name())
								.append(" in ( :ids )");
						Query subdDelete = getSession().createQuery(subDeleteHql.toString());
						subdDelete.setParameterList("ids", ids);
						subdDelete.executeUpdate();
					}
					catch (NoSuchFieldException e) {
						e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
					}
				}
				continue;
			}
		}

		StringBuilder deleteHql = new StringBuilder();
		deleteHql.append("delete from ");
		deleteHql.append(entityClass.getName()).append(" x");
		deleteHql.append(" where x.id in ( :ids )");

		Query delete = getSession().createQuery(deleteHql.toString());
		delete.setParameterList("ids", ids);

		int count = delete.executeUpdate();

		if (count < ids.length) {
			throw new DataNotFoundException(entityClass, ids);
		}

		if (flushWrites) {
			getSession().flush();
		}
	}

	public <T extends Identifiable> void deleteAll(Collection<T> entities)
	{
		if (entities.size() == 0)
			return;

		Class entityClass = null;

		Serializable[] entityIds = new Serializable[entities.size()];
		int i = 0;
		for (T entity : entities) {
			if (entityClass == null) {
				entityClass = entity.getClass();
			}
			entityIds[i++] = entity.getId();
		}

		StringBuilder deleteHql = new StringBuilder();
		deleteHql.append("delete from ");
		deleteHql.append(entityClass.getName());
		deleteHql.append(" where id in ( :ids )");

		Query delete = getSession().createQuery(deleteHql.toString());
		delete.setParameterList("ids", entityIds);

		int count = delete.executeUpdate();

		if (count != entities.size()) {
			throw new DataNotFoundException(entityClass, (Serializable)null);
		}

		if (flushWrites) {
			getSession().flush();
		}
	}

	public void saveChanges()
	{
		getSession().flush();
	}

	public <T extends Serializable> void removeFromContext(Identifiable<T> identifiable)
	{
		getSession().evict(identifiable);
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> executeQuery(Class<T> returnType, String query, Object... params)
	{
		Query hql = getSession().createQuery(query);

		HibernateUtils.applyParamsToQuery(hql, params);

		HibernateUtils.applyTransformToReturnValue(returnType, hql);

		return hql.list();
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> executeNamedQuery(String name, Object... params)
	{
		Query hql = getSession().getNamedQuery(name);

		HibernateUtils.applyParamsToQuery(hql, params);

		return hql.list();
	}

	public int executeSQLUpdate(String sql, Object... params)
	{
		SQLQuery query = getSession().createSQLQuery(sql);

		HibernateUtils.applyParamsToQuery(query, params);

		return query.executeUpdate();
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> executeSQLQuery(Class<T> returnType, String sql, Object... params)
	{
		SQLQuery query = getSession().createSQLQuery(sql);

		HibernateUtils.applyTransformToSqlReturnValue(returnType, query);

		HibernateUtils.applyParamsToQuery(query, params);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public List<Map<String, Class<?>>> executeSQLQuery(Map<String, Class<?>> returnTypes, String sql, Object... params)
	{
		SQLQuery query = getSession().createSQLQuery(sql);

		for (String alias : returnTypes.keySet()) {
			Class<?> returnType = returnTypes.get(alias);
			if (FieldUtils.isValueType(returnType)) {
				query.addScalar(alias, HibernateUtils.getTypeForClass(returnType));
			}
//			else if (Collection.class.isAssignableFrom(returnType)) {
//				query.addJoin(alias, )
//			}
			else if (Identifiable.class.isAssignableFrom(returnType)) {
				query.addEntity(alias, returnType);
			}
		}

		HibernateUtils.applyTransformToSqlReturnValue(Map.class, query);

		HibernateUtils.applyParamsToQuery(query, params);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> executeNamedSQLQuery(Class<T> returnType, String name, Object... params)
	{
		Query namedQuery = getSession().getNamedQuery(name);

		if (! (namedQuery instanceof SQLQuery)) {
			throw new IllegalArgumentException("Cannot execute named HQL query as SQL: " + name);
		}

		SQLQuery query = (SQLQuery)namedQuery;

		HibernateUtils.applyTransformToSqlReturnValue(returnType, query);

		HibernateUtils.applyParamsToQuery(query, params);

		return query.list();
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> executeNamedEntitySQL(String name, Object... params)
	{
		Query query = getSession().getNamedQuery(name);

		HibernateUtils.applyParamsToQuery(query, params);

		query.setResultTransformer(Criteria.DISTINCT_ROOT_ENTITY);

		return query.list();
	}

	/**
	 *
	 * @param hql
	 * @param params
	 * @return
	 */
	public int executeUpdate(String hql, Object... params)
	{
		Query update = getSession().createQuery(hql);

		HibernateUtils.applyParamsToQuery(update, params);

		int count = update.executeUpdate();

		if (flushWrites) {
			getSession().flush();
		}
		
		return count;
	}

}
