package com.portlandwebworks.commons.hibernate;

import com.portlandwebworks.commons.dao.*;
import com.portlandwebworks.commons.data.FieldMapping;
import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.reflection.ClassUtils;
import com.portlandwebworks.commons.reflection.FieldUtils;
import com.portlandwebworks.commons.reflection.NodeType;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Factory;
import org.apache.commons.collections.map.LazyMap;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.hibernate.*;
import org.hibernate.criterion.*;
import org.hibernate.impl.CriteriaImpl;
import org.hibernate.transform.Transformers;
import org.hibernate.type.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.persistence.Column;
import javax.persistence.Transient;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class HibernateUtils
{
	private static Logger logger = LoggerFactory.getLogger(HibernateUtils.class);

	private static HashMap<Class<?>, ProjectionList> fieldProjections = new HashMap<Class<?>, ProjectionList>();

	private static Pattern regex = Pattern.compile("(?:\\sas\\s+\"?)([\\w\\d_]+)(?:\"?\\s*,|\"?\\s+from)",
			Pattern.CASE_INSENSITIVE | Pattern.MULTILINE );

	private static Map<Class<?>, Type> typeMappings = new HashMap<Class<?>, Type>();

	static {

		typeMappings.put(Integer.class, IntegerType.INSTANCE);
		typeMappings.put(Double.class, DoubleType.INSTANCE);
		typeMappings.put(Long.class, LongType.INSTANCE);
		typeMappings.put(String.class, StringType.INSTANCE);
		typeMappings.put(Float.class, FloatType.INSTANCE);
		typeMappings.put(Date.class, DateType.INSTANCE);
		typeMappings.put(Boolean.class, BooleanType.INSTANCE);
		typeMappings.put(Short.class, ShortType.INSTANCE);
		typeMappings.put(Object.class, ObjectType.INSTANCE);
		typeMappings.put(Calendar.class, CalendarType.INSTANCE);
		typeMappings.put(Character.class, CharacterType.INSTANCE);
		typeMappings.put(Byte.class, ByteType.INSTANCE);

		
		// Generate a lookup table of all Hibernate types keyed by java type,  Why doesn't the Hibernate API have this?
/*
		for (Field field : Hibernate.class.getFields()) {
			field.setAccessible(true);
			if (Modifier.isStatic(field.getModifiers())) {
				if (Type.class.isAssignableFrom(field.getType())) {
					try {
						Type type = (Type)field.get(null);
						typeMappings.put(type.getReturnedClass(), type);

						// Save an entry for the primitive type, if this type wraps a primitive
						if (PrimitiveType.class.isAssignableFrom(type.getClass())) {
							typeMappings.put(((PrimitiveType)type).getPrimitiveClass(), type);
						}
					}
					catch (IllegalAccessException e) {
					}
				}
			}
		}
*/
	}

	/**
	 * Determines which Hibernate type to use based on the java type provided.  There doesn't seem to be any mappings
	 * available in Hibernate to do such a fundamental task.
	 *
	 * @param clazz
	 * @return
	 */
	public static Type getTypeForClass(Class<?> clazz)
	{
		Type type = typeMappings.get(clazz);
		if (type != null) {
			return type;
		}
		return StringType.INSTANCE;
	}

	/**
	 * Determines if a
	 *
	 * @param clazz
	 * @return
	 */
	protected static boolean isIdentifiable(Class clazz)
	{
		return Identifiable.class.isAssignableFrom(clazz);
	}

	protected static <T> Class<T> getRootEntityForQuery(Criteria criteria)
	{
		String entityName = ((CriteriaImpl)criteria).getEntityOrClassName();
		Class<T> entityClass = ((CriteriaImpl)criteria).getSession().getFactory().getClassMetadata(entityName).getMappedClass(EntityMode.POJO);
		return entityClass;
	}	

	/**
	 * Builds a Hibernate Criterion instance from a field name, operator and test value.  Does not handle some of the
	 *
	 * @param operand   The name of the field to which the operator should be applied.
	 * @param operator  The operator to apply.  Note that there is not an Operator for all Hibenate Restrictions.  This
	 *                  is simply a limitation in the API.
	 * @param value     The value to test the operand against
	 * @return          A Hibernate Criterion instance.
	 */
	public static Criterion buildCriterion(String operand, Operator operator, Object value)
	{
		switch (operator)
		{
			// Convert the generic DAO operation into a hibernate restriction and add it to the
			// query.
			case EqualTo:
				return Restrictions.eq(operand, value);
			case GreaterThan:
				return Restrictions.gt(operand, value);
			case GreaterThanOrEqual:
				return Restrictions.ge(operand, value);
			case NotEqualTo:
				return Restrictions.ne(operand, value);
			case IsNotNull:
				return Restrictions.isNotNull(operand);
			case IsNull:
				return Restrictions.isNull(operand);
			case IsIn:
				Class operandType = value.getClass();
				if (Collection.class.isAssignableFrom(operandType)) {
					return Restrictions.in(operand, (Collection)value);
				}
				else if (operandType.isArray()) {
					Class arrayType = operandType.getComponentType();
					if (arrayType.isPrimitive()) {
						// primitive array needs to be converted into object array
						int length = Array.getLength(operand);
						Object[] arrayValues = new Object[length];
						for (int i = 0; i < Array.getLength(operand); i++) {
							arrayValues[i] = Array.get(operand, i);
						}
						return Restrictions.in(operand, arrayValues);
					}
					else {
						return Restrictions.in(operand, (Object[])value);
					}
				}
			case LessThan:
				return Restrictions.lt(operand, value);
			case LessThanOrEqual:
				return Restrictions.le(operand, value);
			case Matches:
			case MatchesAnywhere:
				return Restrictions.ilike(operand, value.toString(), MatchMode.ANYWHERE);
			case MatchesStart:
				return Restrictions.ilike(operand, value.toString(), MatchMode.START);
			case MatchesEnd:
				return Restrictions.ilike(operand, value.toString(), MatchMode.END);
			case IsLike:
				return Restrictions.ilike(operand, value.toString().replace('*', '%'));
			case CountEqualTo:
				return Restrictions.sizeEq(operand, ((Number)value).intValue());
			case CountGreaterThan:
				return Restrictions.sizeGt(operand, ((Number)value).intValue());
			case CountLessThan:
				return Restrictions.sizeLt(operand, ((Number)value).intValue());
			case CountNotEqualTo:
				return Restrictions.sizeNe(operand, ((Number)value).intValue());
			case CountGreaterThanOrEqual:
				return Restrictions.sizeGe(operand, ((Number)value).intValue());
			case CountLessThanOrEqual:
				return Restrictions.sizeLe(operand, ((Number)value).intValue());
			default:
				return null;
		}
	}

	/**
	 * Returns a query-safe representation of a property for inclusion in a projection, where clause, sort or group by.
	 * If the property is a field that is on the root object of the query, the property is returned as is.  If the
	 * property is on an associated entity of the root object, the returned value will be an alias for the associated
	 * entity followed by a dot, followed by the property name.
	 *
	 * @param query     The query to evaluate, and if necessary to apply the alias to.
	 * @param property  The field name to resolve.
	 *
	 * @return          Either property, or alias.property
	 *
	 * @throws IllegalArgumentException if he 
	 */
	public static String getPropertyAlias(CriteriaImpl query, String property) throws IllegalArgumentException
	{
		// If there are not dots in this property, it is not a path to a nested entity property.  Return the property.
		int lastIndexOfSeparator = StringUtils.lastIndexOf(property, '.');
		if (lastIndexOfSeparator == -1) {
			return property;
		}

		String associationPath = property.substring(0, lastIndexOfSeparator);
		property = property.substring(lastIndexOfSeparator + 1);

		String alias = HibernateUtils.getAssociationAlias(query, associationPath);
		return alias + "." + property;
	}

	/**
	 * Given an existing query, returns the alias for the associated entity represented by the path provided.  Will add
	 * a new alias to the query if one does not exist.
	 *
	 * @param query             The query to be evaluated
	 * @param associationPath   The path to the associated entity that must be aliased.
	 *
	 * @return                  The alias for the associated entity.
	 */
	public static String getAssociationAlias(CriteriaImpl query, String associationPath)
	{
		return _getAssociationAlias(query, associationPath, null);
	}

	/**
	 * Restricted access method for recursive calls, passing along the parent alias from the outer call.
	 *
	 * @param query
	 * @param associationPath
	 * @param parentAlias
	 * @return
	 */
	private static String _getAssociationAlias(CriteriaImpl query, String associationPath, String parentAlias)
	{
		String immediatePath;
		String fullyScopedImmediatePath;
		String restOfPath = null;

		// Check the association path for any dots, which will indicate that at least one more traversal is necessary
		// to completely resolve the association path.
		int nextSeparator = associationPath.indexOf('.');
		if (nextSeparator == -1) {
			// If no dot is in the association path, it is the only or final part of the path.
			immediatePath = associationPath;
		}
		else {
			// There is at least one traversal to make in this association path.  Process the immediate association,
			// and save the rest for a recursive call to this method.
			immediatePath = associationPath.substring(0, nextSeparator);
			restOfPath = associationPath.substring(nextSeparator + 1);
		}

		// The fully scoped immediate path should include a parent alias from an outer call to this method, if one was
		// made
		if (parentAlias != null) {
			fullyScopedImmediatePath = parentAlias + "." + immediatePath;
		}
		else {
			fullyScopedImmediatePath = immediatePath;
		}

		// Count the number of subcriteria associated with this query.  It will allow us to come up with a counter value
		// that can be used to ensure our alias has a unique name.
		String aliasForImmediate = null;
		int counter = 0;
		for (Iterator iter = query.iterateSubcriteria(); iter.hasNext(); ) {
			CriteriaImpl.Subcriteria sub = (CriteriaImpl.Subcriteria)iter.next();
			if (sub.getPath().equals(fullyScopedImmediatePath)) {
				aliasForImmediate = sub.getAlias();
				break;
			}
			counter++;
		}

		if (aliasForImmediate == null) {
			// Generate an alias for the immediate path that is guaranteed to be unique.
			aliasForImmediate = StringUtils.substring(immediatePath, 0, 3) + "_" + counter;

			// alias the immediate path on the query using an outer join.
			query.createAlias(fullyScopedImmediatePath, aliasForImmediate, Criteria.LEFT_JOIN);
		}

		if (restOfPath != null) {
			// Recurse on the child association
			return _getAssociationAlias(query, restOfPath, aliasForImmediate);
		}
		else {
			// Done.  return the generated alias.
			return aliasForImmediate;
		}
	}

	/**
	 * Converts a list of predicates into Hibernate criteria restrictions and applies them to a query.  Supports applying
	 * predicates as a set of subpredicates for a parent predicate AND or OR predicate, supporting nested AND/OR logic.
	 *
	 * @param query             The criteria to which the predicates will be applied.
	 * @param junction          The current junction that is being assembled, if any.  If a junction is specified, the
	 *                          predicates will be applied to it, not the criteria object.
	 * @param predicates        The set of Predicate instances that will be converted to query criteria.
	 */
	@SuppressWarnings("deprecation")
	public static void applyPredicates(CriteriaImpl query, Junction junction, List<Predicate> predicates)
	{
		for (Predicate predicate : predicates) {

			// Check the predicate for nested predicates. If any are found, ignore the basic properties for this predicate
			// and treat it as an AND or OR operation on a set of nested predicates.
			Predicate[] nestedPredicates = predicate.getPredicates();

			if (! ArrayUtils.isEmpty(nestedPredicates)) {

				// Determine whether to apply AND or OR to the nested predicates as specified by the annotation.
				Junction subJunction = predicate.getAnd() ? Restrictions.conjunction(): Restrictions.disjunction();

				// Recursively call this method, applying the generated predicates to the AND or OR that was created.
				applyPredicates(query, subJunction, Arrays.asList(nestedPredicates));

				if (junction != null) {
					// Apply the assembled criteria to the junction currently being worked on.
					junction.add(subJunction);
				}
				else {
					// Apply the assembled criteria to the root query.
					query.add(subJunction);
				}
			}
			else {
				// The query being acted upon has no nested predicates.  It should be applied to the root query.
				String property = predicate.getProperty();

				String operand = getPropertyAlias(query, property);

				// Convert the predicate into a Hibernate criterion.  if we are currently working on a junction, add it
				// to the junction, otherwise add it to the base criteria object.
				final Criterion criterion = buildCriterion(operand, predicate.getOperator(), predicate.getValue());
				if (junction != null) {
					junction.add(criterion);
				}
				else {
					query.add(criterion);
				}
			}
		}
	}

	/**
	 * Reflects over the "search criteria" object provided, looking for fields annotated with <code>@SearchCriterion</code>,
	 * <code>@AnySearchCriteria</code> and <code>@AllSearchCriteria</code> that have non-null values.  These field values
	 * are applied to a Hibernate query as restrictions to the query.
	 *
	 * @param searchCriteria    The object that contains fields annotated for applying Hibernate query restrictions.
	 * @param query             The query to which the criteria should be applied.
	 */
	public static void applySearchCriteria(Object searchCriteria, CriteriaImpl query)
	{
		if (searchCriteria == null) {
			throw new IllegalArgumentException("Search criteria is required");
		}

		// Short circuit this method if the search criteria ia already a list of predicates.
		if (searchCriteria instanceof PredicateSearchCriteria) {
			applyPredicates(query, ((PredicateSearchCriteria)searchCriteria).getFilter());
			return;
		}

		List<Predicate> predicates = null;
		Field currField = null;

		try {
			// Iterate through each field on the search criteria object and check it for any of the annotations that
			// support search criteria.
			for (Field field : FieldUtils.getFields(searchCriteria.getClass())) {

				// For error reporting
				currField = field;

				// Check to see if there's a single SearchCriteria annotation on this field.  If there is, build the
				// equivalent predicate for it and add it to the list of predicates.
				SearchCriterion criterion = field.getAnnotation(SearchCriterion.class);
				if (criterion != null) {

					assertFieldIsNonPrimitive(field);

					Object value = field.get(searchCriteria);

					Predicate predicate = buildPredicate(field, criterion, value);
					if (predicate != null) {
						// Lazy initialization of the predicates list to handle the situation where the majority of searches
						// performed have no search criteria.
						if (predicates == null) {
							predicates = new ArrayList<Predicate>();
						}
						predicates.add(predicate);
					}
				}
				else {
					SearchCriterion[] subCriteria = null;
					boolean and = true;

					// If this annotation appears on a field, all of its nested SearchCriterion annotations will be ANDed
					// together in one predicate.
					AllSearchCriteria andOp = field.getAnnotation(AllSearchCriteria.class);
					if (andOp != null) {
						subCriteria = andOp.value();
						and = true;
					}

					// If this annotation appears on a field, all of its nested SearchCriterion annotations will be ORed
					// together in one predicate.
					AnySearchCriteria orOp = field.getAnnotation(AnySearchCriteria.class);
					if (orOp != null) {
						subCriteria = orOp.value();
						and = false;
					}

					// If an "all" or "any" was found, assemble the nested predicates under one predicate.
					if (subCriteria != null) {
						List<Predicate> subPreds = null;

						for (SearchCriterion searchCriterion : subCriteria) {

							assertFieldIsNonPrimitive(field);

							Object value = field.get(searchCriteria);

							// Attempt to build a predicate for the field, and if one is returned add it to the sublist
							Predicate pred = buildPredicate(field, searchCriterion, value);
							if (pred != null) {
								// Lazy initialization of list for nested predicates.
								if (subPreds == null) {
									subPreds = new ArrayList<Predicate>();
								}
								subPreds.add(pred);
							}
						}

						if (! CollectionUtils.isEmpty(subPreds)) {
							// build the appropriate predicate for doing AND or OR operation.
							Predicate multiPredicate = and? Predicate.and(subPreds): Predicate.or(subPreds);

							if (predicates == null) {
								predicates = new ArrayList<Predicate>();
							}
							predicates.add(multiPredicate);
						}
					}
				}
			}
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException(String.format("Illegal access to %s.%s", searchCriteria.getClass().getName(), currField.getName()));
		}

		// If this list is non-null, at least one predicate was generated from search criteria.
		if (predicates != null) {
			applyPredicates(query, predicates);
		}
	}

	/**
	 * Throws an exception if a field represents a primitive value.
	 *
	 * @param field     The field to evaluate.
	 * @throws          RuntimeException if the field has a primitive value.
	 */
	static void assertFieldIsNonPrimitive(Field field)
	{
		if (field.getType().isPrimitive()) {
			throw new IllegalArgumentException("Primitive types cannot be used as search criteria.  Impossible to tell if a primitive type is excluded from a search.");
		}
	}

	/**
	 * Builds a Predicate instance from a SearchCriterion annotation.  The primary purpose of this method is to transform
	 * search criteria functionality into predicate functionality for plugging into the findByPredicate() APIs.
	 *
	 * @param field         The field from which the annotation will be built.
	 * @param criterion     The annotation for which the predicate will be built.
	 * @param value         The value assigned to the field provided.
	 *
	 * @return              A new instance of Predicate mirroring the properties of the annotation, or <code>null</code> if
	 *                      the field's value is null or is an empty string and the annotation was configured to ignore
	 *                      empty strings.
	 */
	public static Predicate buildPredicate(Field field, SearchCriterion criterion, Object value)
	{
		// If no property name specified on the annotation, use the name of the field to which it is applied.
		String operand = criterion.propertyName();
		if (operand.equals("")) {
			operand = field.getName();
		}

		// If a hard-coded value was specified on the criterion annotation, convert it from String to the field's type
		// and apply it.  Annotations do not allow these types for fields, so they have to be converted from string.
		if (StringUtils.isNotBlank(criterion.value())) {
			Class type = field.getType();
			value = ClassUtils.convertToType(value, type);
		}
		else {
			// Check to see if the field value is null.  If it is, don't return a Predicate.  This will indicate to the
			// caller that a predicate should not be included in the query for this search criterion.
			if (value == null) {
				return null;
			}

			// Don't apply an empty string if the annotation says not to do so.
			if (value instanceof String && StringUtils.isBlank((String)value)) {
				if (!criterion.useEmptyStringValue()) {
					return null;
				}
			}
		}

		return new Predicate(operand, criterion.operator(), value);
	}

	/**
	 *
	 * @param query
	 * @param predicates
	 */
	public static void applyPredicates(CriteriaImpl query, List<Predicate> predicates)
	{
		applyPredicates(query, null, predicates);
	}

	/**
	 * Convenience method for applying a list of query parameters to a Hibernate query.
	 *
	 * @param query     The query to which the parameters will be applied.
	 * @param params    The query parameters to apply to the query.  The following types of data are accepted as
	 *                  query parameters:
	 *                  <ul>
	 *                      <li>
	 *                          An object that exposes java bean properties that can be used to supply values for the
	 *                          named query parameters in the query provided.  The property names must map to the query
	 *                          parameters.  If the query contains a named parameter that does not have a bean property
	 *                          of the same name, an exception is thrown.
     *                      </li>
	 *                      <li>
	 *                          A Map implementation whose key/value pairs can be used to supply values for the named
	 *                          query parameters in the query provided.
     *                      </li>
	 *                      <li>
	 *                          A list of vararg values that correspond in order to the ? style query parameters i
	 *                          the query provided.  This parameter option is only available if the query uses ?
	 *                          placeholders instead of named placeholders.
	 *                      </li>
	 *                  </ul>
	 * @throws IllegalArgumentException
	 */
	public static void applyParamsToQuery(Query query, Object... params) throws IllegalArgumentException
	{
		if (params.length > 0) {

			// If the query contains no named parameters, assume that the query contains ? placeholders for varargs values
			// Apply each value in order of position in the array.
			if (ArrayUtils.isEmpty(query.getNamedParameters())) {
				int index = 0;
				for (Object param : params) {
					if (param.getClass().isArray() || Collection.class.isAssignableFrom(param.getClass())) {
						throw new IllegalArgumentException("Can't use positional ? placeholders for a query that has a WHERE X IN ( Y ) clause");					
					}
					query.setParameter(index++, param);
				}
			}
			else {
				for (Object param : params) {
					// Allow the caller to send a list of map entries to the varargs parameter.
					if (Map.Entry.class.isAssignableFrom(params[0].getClass())) {
						Map.Entry entry = (Map.Entry)param;
						String entryName = entry.getKey().toString();
						Object entryValue = entry.getValue();
						if (entryValue == null) {
							continue;
						}
						if (entryValue.getClass().isArray()) {
							query.setParameterList(entryName, (Object[])entryValue);
						} 
						else if (Collection.class.isAssignableFrom(entryValue.getClass())) {
							query.setParameterList(entryName, (Collection)entryValue);
						}
						query.setParameter(entryName, entryValue);
					}
					else if (Map.class.isAssignableFrom(param.getClass())) {
						// A Map was passed in.
						query.setProperties((Map)param);
					}
					else {
						// A bean was passed to the varargs.
						query.setProperties(param);
					}
				}
			}
		}
		else {
			logger.warn("No parameters provided for query");
		}
	}


	/**
	 * Given a bean type that maps fields to a type of persisted entity, this method will scan the bean class and build 
	 * a lookup hash of all projected fields on the root entity and its nested entities, and groups them by entity.
	 * 
	 * @param clazz
	 *
	 * @return
	 */
	public static Map<String, List<FieldProjection>> groupProjectionsByEntity(Class<?> clazz)
	{
		Map<String, List<FieldProjection>> map = new HashMap<String, List<FieldProjection>>();
		map = (Map<String, List<FieldProjection>>)LazyMap.decorate(map, new Factory() {
			public Object create()
			{
				return new ArrayList<FieldProjection>();
			}
		});
		groupProjectionsByEntity(clazz, null, null, map);

		return map;
	}

	private static void groupProjectionsByEntity(Class<?> clazz, String parentMapping, String parentField, Map<String, List<FieldProjection>> projections)
	{
/*
		TODO: this is an aborted attempt at using FieldUtils.traverseClassGraph() instead of re-writing the same algorithm
		final Map<String, List<ProjectedField>> projections = new HashMap<String, List<ProjectedField>>();
		
		Stack<String> entityPathStack = new Stack<String>();
		
		FieldUtils.traverseClassGraph(clazz, new FieldUtils.FieldTypeVisitor()
		{
			public boolean visit(String path, Class<?> elementType, NodeType nodeType)
			{
				switch (nodeType) {
					case ValueType:
						String projectionGroup = StringUtils.substringBefore(path, ".");
						projections.get(projectionGroup).add(new ProjectedField(parentMapping, parentField));
						break;
				}
			}
		});
*/

		if (parentMapping != null && FieldUtils.isValueType(clazz)) {
			String projectionGroup = StringUtils.substringBefore(parentField, ".");
			projections.get(projectionGroup).add(new FieldProjection(parentMapping, parentField));
		}

		// Get an array of all fields on the return type that have a mapping annotation.
		Field[] fieldsWithMapping = FieldUtils.getFieldsWithAnnotation(clazz, FieldMapping.class);
		for (Field field : fieldsWithMapping) {
			FieldMapping mapping = field.getAnnotation(FieldMapping.class);

			// The mapping annotation's value is the string path this field's value maps to on the associated
			// entity.  If no value is specified, assume it is the name of the field.
			String mappingProperty = mapping.value();
			if (StringUtils.isEmpty(mappingProperty)) {
				mappingProperty = field.getName();
			}

			logger.debug("Processing field mapping(value={}, unique={}) ", mappingProperty, mapping.unique());

			String fullEntityPath;
			if (parentMapping != null) {
				fullEntityPath = parentMapping + "." + mappingProperty;
			}
			else {
				fullEntityPath = mappingProperty;
			}

			String fullField;
			if (parentField != null) {
				fullField = parentField + "." + field.getName();
			}
			else {
				fullField = field.getName();
			}

			NodeType nodeType = FieldUtils.getNodeTypeForClass(field.getType());
			switch (nodeType) {
				case ValueType: {
					// Store the projection under a key that is the path to the first nested entity under the root element
					// of the query.
					String projectionGroup = StringUtils.substringBefore(parentField, ".");
					projections.get(projectionGroup).add(new FieldProjection(fullEntityPath, fullField));
					break;
				}
				case Collection: {
					Class<?> nestedType = (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[0];
					groupProjectionsByEntity(nestedType, fullEntityPath, fullField, projections);
					break;
				}
				case Map: {
					Class<?> nestedType = (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[1];
					groupProjectionsByEntity(nestedType, fullEntityPath, fullField, projections);
					break;
				}
				case Array: {
					Class<?> nestedType = field.getType().getComponentType();
					groupProjectionsByEntity(nestedType, fullEntityPath, fullField, projections);
					break;
				}
				case Object:
					//TODO: what to do here???
					break;
			}
		}
	}


	/**
	 * Utility method for analyzing a fetch option provided for a query, converting it into an association path.
	 *
	 * @param fetchOption
	 * @return
	 */
	private static String getAssociationPathForFetchOption(Object fetchOption)
	{
		String join;
		Class fetchClass = fetchOption.getClass();
		if (fetchClass.equals(String.class)) {
			// a string fetch option should represent the entity association path to the desired entity to be eager
			// fetched.
			join = fetchOption.toString();
		}
		else {
			// This must be an enum representation of the eager fetch.  Getting the field from the enum gives us a field
			// representation of the enum value passed to this method.
			Field fetchField = FieldUtils.getField(fetchClass, fetchOption.toString());

			// If the enum field has a FetchField annotation on it, then use the value of that enum as the association
			// path to the eagerly fetched field.
			FetchOption fetch = fetchField.getAnnotation(FetchOption.class);
			if (fetch != null) {
				join = fetch.value();
			}
			else {
				// No annotation present.  Use the name of the enum as the field to be eager fetched.
				join = fetchField.getName();
			}
		}

		return join;
	}

	/**
	 * Applied paging options to a Hibernate query.
	 *
	 * @param resultsQuery
	 * @param options
	 */
	public static void applyPaging(Criteria resultsQuery, IPagedSearchOptions options)
	{
		if (options != null) {
			logger.debug("Applying paging: start={}, limit={}", options.getStart(), options.getLimit());
		}

		// Skip ahead to page requested.
		resultsQuery.setFirstResult(options.getStart());

		resultsQuery.setMaxResults(options.getLimit());
	}

	/**
	 * Convenience method for applying sort specifications to a Hibernate query.
	 *
	 * @param entityClass
	 * @param criteria
	 * @param sortSpecs
	 */
	public static void applySorting(Class<?> entityClass, CriteriaImpl criteria, List<SortSpecification> sortSpecs)
	{
		SortSpecification[] sortSpecArray = sortSpecs.toArray(new SortSpecification[sortSpecs.size()]);
		
		applySorting(entityClass, criteria, sortSpecArray);
	}
	
	/**
	 * Convenience method for applying sort specifications to a Hibernate query.
	 *
	 * @param entityClass
	 * @param criteria
	 * @param sortSpecs
	 */
	public static void applySorting(Class<?> entityClass, CriteriaImpl criteria, SortSpecification... sortSpecs)
	{
		List<SortSpecification> sorts = getSortSpecifications(entityClass, sortSpecs);
		if (logger.isDebugEnabled() && sorts.size() == 0) {
			logger.debug("Entity {} has no default sorts specified.  No default sorting applied to query", entityClass.getName());
			return;
		}

		for (SortSpecification sortSpec : sorts) {

			// Sort property could refer to a field on a nested entity.  Handle that.
			String aliasedSortProperty = getPropertyAlias(criteria, sortSpec.getProperty());

			logger.debug("Applying sorting: property={}, direction={}", aliasedSortProperty, sortSpec.getDirection());

			// Apply sort direction based on spec provided.
			if (sortSpec.getDirection().equals(SortDirection.ASC)) {
				criteria.addOrder(Order.asc(aliasedSortProperty));
			} else {
				criteria.addOrder(Order.desc(aliasedSortProperty));
			}
		}
	}

	/**
	 * Merges the sort specifications provided with the default sorts (if any) defined at the class level with @DefaultOrderBy
	 * annotations, as well as applying any sort aliases defined at the class level with @SortProeperty annotations.
	 * 
	 * @param entityClass
	 * @param sortSpecs
	 * @return
	 * 
	 * @see com.portlandwebworks.commons.dao.DefaultOrderBy
	 * @see com.portlandwebworks.commons.dao.SortProperty
	 */
	protected static List<SortSpecification> getSortSpecifications(Class<?> entityClass, SortSpecification... sortSpecs)
	{
		List<SortSpecification> sorts = null;
		if (sortSpecs.length > 0) {
			sorts = Arrays.asList(sortSpecs);

			// Scan the class for any fields that declare a sort alias for them, and check for match.
			Field[] fieldsWithSortAlias = FieldUtils.getFieldsWithAnnotation(entityClass, SortProperty.class);
			if (fieldsWithSortAlias.length > 0) {
				//TODO: this could be optimized.  But really, how often will there be an aliased sort property?
				for (SortSpecification sortSpecification : sorts) {
					for (Field field : fieldsWithSortAlias) {
						if (field.getAnnotation(SortProperty.class).value().equals(sortSpecification.getProperty())) {
							logger.debug("Applying sort alias: alias={}, field={}", sortSpecification.getProperty(), field.getName());
							sortSpecification.setProperty(field.getName());
						}
					}
				}
			}
		}

		// If no sorts were passed to this method, check the class for any fields annotated with default sorts.  This
		// method already takes into account whether the fields have a SortProperty annotation.
		if (CollectionUtils.isEmpty(sorts)) {
			sorts = DaoUtils.getDefaultSortSpecifications(entityClass);
		}

		return sorts;
	}


	/**
	 * Thank you John Gordon for implementing proper record counting for legacy databases with composite primary keys.
	 *
	 * @param embedded  The composite primary key field
	 *
	 * @return          A Hibernate projection that can be applied to a Criteria instance.
	 */
	protected static Projection createCompoundKeyCountProjection(Field embedded) {

		// get a list of all declared fields on the composite key.
		Field[] compositeFields = FieldUtils.getFields(embedded.getType());

		if (compositeFields.length == 0) {
			throw new IllegalArgumentException("Embedded id class contains no fields.");
		}

		String countSql = "COUNT(DISTINCT %s) AS %s";

		// This will hold the comma-delimited set of composite key names to plug into the query.
		StringBuilder fieldSql = new StringBuilder();

		for (Field compositeField : compositeFields) {

			Column columnAnnotation = compositeField.getAnnotation(Column.class);

			if (columnAnnotation != null) {

				if (fieldSql.length() > 0) {
					fieldSql.append(", ");
				}
				// This field is part of the primary key.  Add it to the list.
				String columnName = columnAnnotation.name();
				if (StringUtils.isNotEmpty(columnName)) {
					fieldSql.append(columnName);
				}
				else {
					fieldSql.append(compositeField.getName());
				}
			}
			else {
				// R.N. 1/29/2011 - I disabled this line of JG code becasue I don't think you should include a field in
				// the list of composite key names if the field has no @Column annotation on it.
				//fieldSql.append(compositeField.getName());
				logger.warn("Skipping over field {} in composite key class {}--it has no @Column annotation ",
						compositeField.getName(), embedded.getName());
			}
		}

		String[] rowAliases = new String[] { "resultCount" };

		countSql = String.format(countSql, fieldSql.toString(), rowAliases[0]);

		Type[] types = new Type[] { getTypeForClass(Integer.class) };

		return  Projections.sqlProjection(countSql, rowAliases, types);

	}

	/**
	 * Convert one or more "fetch options" into query joins that will be returned with the fields of the query's root
	 * entity.  This allows retrieving a root entity's associated entity fields in one query instead of by lazy loading
	 * after the initial query has returned.
	 *
	 * Fetch options can be either a String value that is the property name or association path pointing to the
	 * associated entity, any class type whose toString() value is a property or path, or any class that has a
	 * <code>@FetchOption</code> annotation indicating the property or path to fetch.
	 *
	 * If multiple one-to-many associations are eager fetched, only the first one in the list of arguments will actually
	 * be eagerly fetched.  The remaining one-to-many associations will be fetched in separate queries after the initial
	 * root entity query is completed.  This is a simplistic way to address Hibernate's restriction on eager fetching
	 * one-to-many fields.
	 *
	 * @param criteria      The query to which the eager fetches will be applied
	 * @param fetchOptions  The fetch options to apply.
	 *
	 * @return
	 */
	protected static List applyEagerFetchOptions(Class<?> entityClass, final CriteriaImpl criteria, Object... fetchOptions)
	{
		// This will be the singular 1-M entity field that is pre-fetched.  Hibernate only allows
		// pre-fetching one aggregate field when retrieving an entity.
		//TODO: this does not take into consideration whether the entity has eager fetch on any 1-M associations
		Object prefetch = null;

		// This is a list tracking all 1-M fetches that must be deferred until after the query.
		List deferred = null;

		// Iterate through the fetch options, eager joining all 1-1 relationships, an skipping over any
		// 1-M relationships after the first one has been eager joined.
		for (Object fetchOption : fetchOptions) {

			String join = getAssociationPathForFetchOption(fetchOption);
			if (FieldUtils.isMultiValueField(entityClass, join)) {
				if (prefetch != null) {
					logger.debug("Skipping fetch option {} as eager load field in query: can only eager load one 1-M field ({})", fetchOption, prefetch);
					if (deferred == null) {
						deferred = new ArrayList();
						deferred.add(fetchOption);
					}
				}
				else {
					logger.debug("Eager joining collection: {}", join);

					prefetch = fetchOption;

					FieldUtils.traversePathInClassGraph(entityClass, join, new FieldUtils.FieldTypeVisitor()
					{
						int multipleCount = 0;
						StringBuilder assembledPath = new StringBuilder();

						public boolean visit(String path, Class<?> elementClass, NodeType nodeType)
						{
							if (nodeType == NodeType.Array || nodeType == NodeType.Collection || nodeType == NodeType.Map) {
								multipleCount++;
							}

							// Assemble the intermediate path processed so far.
							if (assembledPath.length() > 0) {
								assembledPath.append('.');
							}
							assembledPath.append(path);

							// Only eager join this field if does not push the number of eager joined 1-M associations
							// beyond one, which messes with hibernate's ability to handle the query results.
							//TODO: this doesn't take into account mechanisms that allow Hibernate to eager fetch more than
							//TODO: one 1-M field on a root entity.
							if (multipleCount <= 1) {
								criteria.setFetchMode(assembledPath.toString(), FetchMode.JOIN);
							}

							// Don't continue evaluating this path if more than one multi-value field has been eager joined.
							return multipleCount <= 1;
						}
					});
				}
			}
			else {
				logger.debug("Eager joining association: {}", join);

				criteria.setFetchMode(join, FetchMode.JOIN);
			}
		}

		return deferred;
	}

	/**
	 * Triggers a load of all associated entities represented by the (presumably) one-to-many fetch options provided.
	 *
	 * @param entity        The entity on which the loads are to be applied.
	 * @param fetchOptions  The remaining one-to-many fetch options that could not be included in the root entity query.
	 */
	public static void loadDeferredFetchOptions(final Criteria query, final Object entity, Object... fetchOptions)
	{
		// This is needed to interrogate the query for entity fetches
		final CriteriaImpl impl = (CriteriaImpl)query;

		// Post-fetch all requested aggregate fields except for the one that was pre-fetched.  The name
		// of the enum is the same as the bean property.
		for (Object fetchOption : fetchOptions) {

			final String associationPath = getAssociationPathForFetchOption(fetchOption);

			FieldUtils.traversePathInObjectGraph(entity, associationPath, new FieldUtils.NodeVisitor()
			{
				public boolean visit(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
				{
					if (nodeValue != null) {
						if (impl.getFetchMode(nodePath) == null) {
							if (nodeType == NodeType.Collection) {
								logger.warn("Applying eager post fetch to collection {}.{}.  This can be an expensive operation for large query results", entity.getClass().getName(), nodePath);
								((Collection)nodeValue).size();
							}
							else if (nodeType == NodeType.Map) {
								logger.warn("Applying eager post fetch to map {}.{}.  This can be an expensive operation for large query results", entity.getClass().getName(), nodePath);
								((Map)nodeValue).size();
							}
							else if (nodeType == NodeType.Object) {
								// The traversal algorithm will take care of this object as it attempts to travers over it.
							}
						}
					}

					return true;
				}

				public void leaving(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
				{
					//To change body of implemented methods use File | Settings | File Templates.
				}   
			});
		}
	}

	/**
	 *
	 * @param returnType
	 * @param hql
	 * @param <T>
	 */
	public static <T> void applyTransformToReturnValue(Class<T> returnType, Query hql)
	{
		if (hql.getQueryString().contains("select new")) {
			// Don't try to apply a transformer to this query.  the "select new" tells Hibernate what should be returned.
			logger.debug("HQL query specifies its own return value type with new operator");
		}
		else {
			NodeType nodeType = FieldUtils.getNodeTypeForClass(returnType);
			switch (nodeType) {
				case Map:
					hql.setResultTransformer(Transformers.ALIAS_TO_ENTITY_MAP);
					break;
				case Collection:
					hql.setResultTransformer(Transformers.TO_LIST);
					break;
				case Array:
					// Result will be object[]
					break;
				case ValueType:
					// Result will be Object, not Object[]
					break;
				case Object:
					if (isIdentifiable(returnType)) {
						// Let the query impl handle the return type, but verify the return type matches after query since there is
						// a loose coupling between the HQL and the method argument.
					}
					else {
						// The return type is not an entity, a collection, or a scalar type.  Tell hibernate to map the return values
						// to the result bean by mapping sql aliases to bean properties.
						hql.setResultTransformer(Transformers.aliasToBean(returnType));
					}
					break;
			}
		}
	}


	/**
	 * Prepares the appropriate transform to a SQL query based on the return type provided.
	 *
	 * @param returnType
	 * @param query
	 * @param <T>
	 */
	public static <T> void applyTransformToSqlReturnValue(Class<T> returnType, SQLQuery query)
	{
		NodeType nodeType = FieldUtils.getNodeTypeForClass(returnType);
		switch (nodeType) {
			case Map:
				String queryString = query.getQueryString();
				Matcher matcher = regex.matcher(queryString);
				boolean foundAny = false;
				while (matcher.find()) {
					String alias = matcher.group(1);
					query.addScalar(alias);
					foundAny = true;
				}
				query.setResultTransformer(Transformers.ALIAS_TO_ENTITY_MAP);
				if (! foundAny) {
					throw new RuntimeException("Can't apply ALIAS_TO_ENTITY_MAP transformation to query--no SQL aliases found");
				}
				break;
			case Collection:
				query.setResultTransformer(Transformers.TO_LIST);
				break;
			case Array:
				// Default return type provided by Hibernate.
				break;
			case ValueType:
				// Result will be Object, not Object[]
				break;
			case Object:
				if (Identifiable.class.isAssignableFrom(returnType)) {
					query.addEntity(returnType);
				}
				else {
					// The return value is neither an entity, a collection, or a simple scalar type.  Attempt to map the non-transient
					// fields of the return type to aliased sql values.
					Field[] returnFields = FieldUtils.getFields(returnType);
					for (Field returnField : returnFields) {
						if (returnField.getAnnotation(Transient.class) != null) {
							logger.debug("Skipping transient field in SQL query: {}.{}", returnType, returnField.getName());
							continue;
						}

						if (FieldUtils.isMultiValueField(returnType, returnField.getName())) {
							logger.debug("Skipping multi-value field in SQL query: {}.{}", returnType, returnField.getName());
							continue;
						}

						query.addScalar(returnField.getName(), getTypeForClass(returnField.getType()));
					}

					// Tell hibernate to map the column values to the return type based, mapping the column aliases to properties
					// on the return type.
					query.setResultTransformer(Transformers.aliasToBean(returnType));
				}
				break;
		}
	}
}
