package com.portlandwebworks.commons.jsf.lifeCycle;

import javax.faces.context.FacesContext;

/**
 * This software is the property of Portland Webworks.
 *
 * Use this class to implement custom logic before of after JSF lifecycle phases.  If you prefer to handle only a small
 * subset of phases, consider extending the <code>LifeCycleAwareBackingBean</code> class.
 *
 * In addition to implementing this interface, or extending the previously mentioned class, it is critical that the view
 * ID of your backing bean match the name of the page that it supports, with the extension removed.
 *
 * Below is a description of each of the JSF lifecycle phases.
 *
 *
 * RESTORE_VIEW
 *
 * Create/Restore view phase is responsible for creating or restoring the server side component tree representing the
 * page on the client side. The view is stored in a parent container object called FacesContext . The servlet programming
 * model guarantees thread safety. So if you hit the url for /faces/jsp/actor.jsp first time, the request is a first
 * request for the actor.jsp page. The /faces/* url pattern brings FacesServlet into picture. FacesServlet traps the
 * request, strips the /faces part and looks for a jsp page at /jsp/actor.jsp. Though it is a JSP page but notice that
 * no JSP compiler is involved. Rather the page is read by JSF and a component tree is built. If the request comes for
 * second time the same component tree is restored. If the request is a first request the lifecycle directly goes to
 * Render Response phase otherwise it goes to next phase which is Apply Request Values phase.
 *
 *
 * APPLY_REQUEST_VALUES
 *
 * In apply request value phase, the incoming request is processed. The parameter map is fetched and parsed for name
 * value pair. Only UI component who have an input value participate in this phase. So if we look into actor.jsp, in
 * this phase the value contained in the input box will be set into the server side UIComponent representing the input
 * box. Also the click of the submit button is noted in this phase.
 *
 *
 * PROCESS_VALIDATION
 *
 * In this phase both conversion and validations happen. Conversion is done first followed by validation. For example if
 * we have Date property in our actor bean, than the conversion of string value received from front end is converted to
 * a Java object at this stage. If we put validations of say length on the input box than those validations are checked
 * here. This is a server side validation and has nothing to do with if you put a javascript validation on the front end.
 *
 *
 * UPDATE_MODEL_VALUES
 *
 * This is the place where backing beans are populated. So the actor bean will get the name value set by JSF at this
 * stage by the value entered at front end.
 *
 *
 * INVOKE_APPLICATION
 *
 * Here the actions or actions listeners are handled. In this case we have clicked a submit button which generated an
 * action "save". The save in this case results in navigation from actor.jsp page to confirm.jsp page as can be seen in
 * the navigation block of faces-config.xml.
 *
 *
 * RENDER_RESPONSE
 *
 * In the render response phase the response is created. In the example, the view of confirm.jsp will be created and
 * based on the component tree, the html markup is created and the response is send back to the client.
 *
 *
 * @author Ronnie Northup
 */
public interface LifeCycleAware
{

	/**
	 * This method is invoked before the RESTORE_VIEW phase is initiated, for initial page loads only.  It can be used
	 * to control how JSF builds the initial server-side object representation of the view.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeInitializeView(FacesContext facesContext);


	/**
	 * This method is invoked after the RESTORE_VIEW phase is completed, for initial page loads only.
	 *
	 * This might be a good time to make custom changes to the JSF component tree for the loaded view.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterInitializeView(FacesContext facesContext);


	/**
	 * This method is invoked before the RESTORE_VIEW phase is initiated, for page post backs only.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeRestoreView(FacesContext facesContext);

	/**
	 * This method is invoked after the RESTORE_VIEW phase is completed, for page post backs only.
	 *
	 * This might be a good time to make custom changes to the JSF component tree for the loaded view.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterRestoreView(FacesContext facesContext);

	/**
	 * This method is invoked before the APPLY_REQUEST_VALUES phase is initiated.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeApplyRequestValues(FacesContext facesContext);

	/**
	 * This method is invoked after the APPLY_REQUEST_VALUES phase is completed.
	 *
	 * This might be a good time to 
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterApplyRequestValues(FacesContext facesContext);

	/**
	 * This method is invoked before the PROCESS_VALIDATIONS phase is initiated.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeProcessValidations(FacesContext facesContext);

	/**
	 * This method is invoked after the PROCESS_VALIDATIONS phase is completed.
	 *
	 * This might be a good time to do custom things related to validation that can't be achieved through component
	 * attributes and custom validators.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterProcessValidations(FacesContext facesContext);

	/**
	 * This method is invoked before the UPDATE_MODEL_VALUES phase is initiated.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeUpdateModelValues(FacesContext facesContext);

	/**
	 * This method is invoked after the UPDATE_MODEL_VALUES phase is completed.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterUpdateModelValues(FacesContext facesContext);

	/**
	 * This method is invoked before the INVOKE_APPLICATION phase is initiated.
	 *
	 * This might be a good time to do any work on a backing bean that needs to be done prior to the execution of event
	 * handlers.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeInvokeApplication(FacesContext facesContext);

	/**
	 * This method is invoked after the INVOKE_APPLICATION phase is completed.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterInvokeApplication(FacesContext facesContext);

	/**
	 * This method is invoked before the RENDER_RESPONSE phase is initiated.
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void beforeRenderResponse(FacesContext facesContext);

	/**
	 * This method is invoked after the RENDER_RESPONSE phase is completed.  I don't think there is much 
	 *
	 * @param facesContext  The JSF context that is in use for the life cycle phase.
	 */
	void afterRenderResponse(FacesContext facesContext);
}
