package com.portlandwebworks.commons.jsf.lifeCycle;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;

import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.faces.event.PhaseEvent;
import javax.faces.event.PhaseId;
import javax.faces.event.PhaseListener;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import java.util.HashMap;

/**
 * This software is the property of Portland Webworks.
 *
 * A phase listener that invokes callbacks on backing beans, when available, based on where the current JSF request
 * life cycle.  The naming convention that is used for locating the backing bean in either a JSF or Spring scope is not
 * included in this abstract class and must be implemented by either a custom implementation or by an existing subclass
 * such as <code>RegexPhaseCallbackSupport</code>
 *
 * @author Ronnie Northup
 */
public abstract class PhaseCallbackSupport implements PhaseListener
{
	private Log logger = LogFactory.getLog(PhaseCallbackSupport.class);

	/**
	 * A cache of derived bean names for each view ID
	 */
	private HashMap<String, String> viewBeanNames = new HashMap<String, String>();

	/**
	 * Implementing subclass should apply application-specific logic to derive the name of the backing bean based on the
	 * path (ID) of the view being loaded.
	 *
	 * @param facesContext  The context to be evaluated.
	 * @param viewId         The path of the view that is being loaded during this JSF request.
	 * @return              The name of the backing bean in whatever context it is stored.
	 */
	public abstract String getBeanNameForViewPath(FacesContext facesContext, String viewId);


	@Override
	public void afterPhase(PhaseEvent phaseEvent)
	{
		FacesContext facesContext = phaseEvent.getFacesContext();

		LifeCycleAware lifeCycleAwareBean = this.getLifeCycleAwareBean(facesContext);
		if (lifeCycleAwareBean != null) {

			PhaseId phaseId = phaseEvent.getPhaseId();
			if (phaseId.equals(PhaseId.RESTORE_VIEW)) {

				// Distinguish between initial view setup and existing view restore.
				if (isInitialPageRequest(facesContext)) {
					lifeCycleAwareBean.afterInitializeView(facesContext);
				}
				else {
					lifeCycleAwareBean.afterRestoreView(facesContext);
				}
			}
			else if (phaseId.equals(PhaseId.APPLY_REQUEST_VALUES)) {
				lifeCycleAwareBean.afterApplyRequestValues(facesContext);
			}
			else if (phaseId.equals(PhaseId.PROCESS_VALIDATIONS)) {
				lifeCycleAwareBean.afterProcessValidations(facesContext);
			}
			else if (phaseId.equals(PhaseId.UPDATE_MODEL_VALUES)) {
				lifeCycleAwareBean.afterUpdateModelValues(facesContext);
			}
			else if (phaseId.equals(PhaseId.INVOKE_APPLICATION)) {
				lifeCycleAwareBean.afterInvokeApplication(facesContext);
			}
			else if (phaseId.equals(PhaseId.RENDER_RESPONSE)) {
				lifeCycleAwareBean.afterRenderResponse(facesContext);
			}
		}
	}

	@Override
	public void beforePhase(PhaseEvent phaseEvent)
	{
		FacesContext facesContext = phaseEvent.getFacesContext();

		LifeCycleAware lifeCycleAwareBean = this.getLifeCycleAwareBean(facesContext);
		if (lifeCycleAwareBean != null) {

			PhaseId phaseId = phaseEvent.getPhaseId();
			if (phaseId.equals(PhaseId.RESTORE_VIEW)) {

				// Distinguish between initial view setup and existing view restore.
				if (isInitialPageRequest(facesContext)) {
					lifeCycleAwareBean.beforeInitializeView(facesContext);
				}
				else {
					lifeCycleAwareBean.beforeRestoreView(facesContext);
				}
			}
			else if (phaseId.equals(PhaseId.APPLY_REQUEST_VALUES)) {
				lifeCycleAwareBean.beforeApplyRequestValues(facesContext);
			}
			else if (phaseId.equals(PhaseId.PROCESS_VALIDATIONS)) {
				lifeCycleAwareBean.beforeProcessValidations(facesContext);
			}
			else if (phaseId.equals(PhaseId.UPDATE_MODEL_VALUES)) {
				lifeCycleAwareBean.beforeUpdateModelValues(facesContext);
			}
			else if (phaseId.equals(PhaseId.INVOKE_APPLICATION)) {
				lifeCycleAwareBean.beforeInvokeApplication(facesContext);
			}
			else if (phaseId.equals(PhaseId.RENDER_RESPONSE)) {
				lifeCycleAwareBean.beforeRenderResponse(facesContext);
			}
		}
	}

	@Override
	public PhaseId getPhaseId()
	{
		// This listener needs to be invoked on any phase.
		return PhaseId.ANY_PHASE;
	}

	/**
	 * Attempts to locate a backing bean with the specified name that implements <code>LifeCycleAware</code>.  It will
	 * search both the JSF EL context and a Spring context, if one is present.
	 *
	 * @param facesContext      The faces context containing the View ID and whose context will be searched for a bean.
	 * @return                  A reference to the backing bean, or null if no bean by the specified name is found in
	 *                          either context.
	 */
	private LifeCycleAware getLifeCycleAwareBean(FacesContext facesContext)
	{
		String viewPath;

		UIViewRoot viewRoot = facesContext.getViewRoot();
		if (viewRoot == null) {
			// On some JSF life cycle phases UI elements such as the root view are not available.
			HttpServletRequest request = (HttpServletRequest)facesContext.getExternalContext().getRequest();
			viewPath = request.getRequestURI();

			String context = request.getContextPath();
			// If there is no context, context is just a slash.  Don't strip the slash off the front of the path.
			if (! context.equals("/")) {
				viewPath = StringUtils.substringAfter(viewPath, context);
			}

			logger.debug("Extracted request URI from servlet request: " + viewPath);
		}
		else {
			// View ID is the same thing as the path to the xhtml file.
			viewPath = viewRoot.getViewId();

			logger.debug("Extracted view ID from faces context: " + viewPath);
		}

		// Calculating a bean name from a view ID might be a costly operation.  Check the local cache first.
		String beanName = viewBeanNames.get(viewPath);
		if (beanName == null) {

			// Ask the subclass to determine the bean name based on the view path.
			beanName = getBeanNameForViewPath(facesContext, viewPath);

			viewBeanNames.put(viewPath, beanName);
		}


		// Check the Faces EL Context for a backing bean hooked up using the standard JSF mechanism.
		Object object = facesContext.getELContext().getELResolver().getValue(facesContext.getELContext(), null, "#{" + beanName + "}");
		if (object != null) {
			logger.debug("Found backing bean in Faces context: " + beanName);
		}

		// Check the Spring Context for the backing bean
		ServletContext servletContext = (ServletContext)facesContext.getExternalContext().getContext();
		WebApplicationContext context = WebApplicationContextUtils.getWebApplicationContext(servletContext);

		// Need to check first before calling getBean() to prevent an exception from being thrown.  If there is no Spring
		// context available, getWebApplication will have returned null.
		if (context != null && context.containsBean(beanName)) {
			object = context.getBean(beanName);
			logger.debug("Found backing bean in Spring context: " + beanName);
		}

		// Only return the object if it implements the required interface.
		if (object != null && LifeCycleAware.class.isAssignableFrom(object.getClass())) {
			return (LifeCycleAware)object;
		}
		else {
			logger.debug("Backing bean " + beanName + " does not implement the LifeCycleAware interface");
			return null;
		}
	}

	/**
	 * Returns true if there is no view state in the faces request, indicating it is the view being loaded is an initial
	 * load, not the result of a page postback.
	 *
	 * @param facesContext  The faces context to evaluate.
	 * @return              True if initial view load.
	 */
	private boolean isInitialPageRequest(FacesContext facesContext) {
		ExternalContext external = facesContext.getExternalContext();
		if (external != null) {
			return !external.getRequestParameterMap().containsKey("javax.faces.ViewState");
		}
		return true;
	}
}
