/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.reflection;

import com.portlandwebworks.commons.data.FieldMapping;
import org.apache.commons.collections.Factory;
import org.apache.commons.collections.map.LazyMap;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.mock.web.MockPageContext;

import java.beans.PropertyDescriptor;
import java.lang.annotation.Annotation;
import java.lang.reflect.*;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class FieldUtils
{
	private static Logger logger = LoggerFactory.getLogger(FieldUtils.class);
	
	/**
	 * Cached class fields, for those times when you want to process the whole list.
	 */
	private final static HashMap<Class<?>, Field[]> arrayCache = new HashMap<Class<?>, Field[]>();

	/**
	 * Cached class fields, mapped by name for quick access to a field by its declared name.
	 */
	private final static HashMap<Class<?>, HashMap<String, Field>> mapCache = new HashMap<Class<?>, HashMap<String, Field>>();

	/**
	 * Cached TODO finish description
	 */
	private final static HashMap<Class<?>, List<String>> fieldMultiplicities = new HashMap<Class<?>, List<String>>();

	/**
	 * Cached class fields where the field represents any type of one-to-many relationship.
	 */
	private final static HashMap<Class<?>, Field[]> uniqueFields = new HashMap<Class<?>, Field[]>();

	/**
	 * Cached class fields, categorized by the type of annotation assigned to the field.  Lazily 
	 */
	private final static Map<Class<? extends Annotation>, Map<Class<?>, Field[]>> annotatedFields =
			(Map<Class<? extends Annotation>, Map<Class<?>, Field[]>>)LazyMap.decorate(
					new HashMap<Class<? extends Annotation>, Map<Class<?>, Field[]>>(), new Factory()
			{
				public Object create()
				{
					return new HashMap<Class<?>, Field[]>();
				}
			});

	/**
	 * Retrieves all declared fields from a class and all of its super classes.  This has a few advantages over using
	 * the java reflection API to get fields:
	 * <ul>
	 *     <li>Uses Field.setAccessible(boolean) to gain access to class fields that are declared private or protected.</li>
	 *     <li>Throws <code>RuntimeException</code> for any problems encountered.</li>
	 *     <li>Caches results from reflective calls to improve performance.</li>
	 *     <li>Ignores static fields.</li>
	 *     <li>Ignores the fields named class and TYPE.</li>
	 * </ul>
	 * 
	 * @param clazz
	 * @return
	 */
	public static Field[] getFields(Class<?> clazz)
	{
		Field[] fieldArray = arrayCache.get(clazz);
		if (fieldArray == null) {
			synchronized (arrayCache) {
				if (fieldArray == null) {
					List<Field> fields = new ArrayList<Field>();
					HashMap<String, Field> fieldMap = new HashMap<String, Field>();

					Class<?> currClass = clazz;

					while (currClass != null) {
						for (Field field : currClass.getDeclaredFields()) {
							field.setAccessible(true);
							// Skip static fields, unless we're working with an enumeration, whose enumerated values are
							// static fields.
							if (Modifier.isStatic(field.getModifiers()) && ! currClass.isEnum()) {
								continue;
							}
							if (field.getName().equals("class")) {
								continue;
							}
							if (field.getName().equals("TYPE")) {
								continue;
							}
							fields.add(field);
							fieldMap.put(field.getName(), field);
						}

						currClass = currClass.getSuperclass();
					}

					fieldArray = fields.toArray(new Field[fields.size()]);
					arrayCache.put(clazz, fieldArray);
					mapCache.put(clazz, fieldMap);
				}
			}
		}
		return fieldArray;
	}

	/**
	 * Retrieves all fields of a class, including those inherited from superclasses or interfaces, that have the
	 * specified annotation.
	 *
	 * @param clazz             The class to inspect for annotated fields.  All of its ancestors will also be inspected.
	 * @param annotationTypes   The types of annotation to look for.
	 *
	 * @return                  An array of all fields with the specified annotation.  If no fields are found with the
	 *                          annotation, returns an empty array.
	 */
	public static Field[] getFieldsWithAnnotation(Class<?> clazz, Class<? extends Annotation>... annotationTypes)
	{
		List<Field> temp = null;

		for (Class<? extends Annotation> annotationType : annotationTypes) {

			Map<Class<?>, Field[]> cachedFields = annotatedFields.get(annotationType);

			Field[] fields = cachedFields.get(clazz);
			if (fields == null) {
				List<Field> annotatedFields = new ArrayList<Field>();

				fields = FieldUtils.getFields(clazz);
				for (Field field : fields) {
					Annotation annot = field.getAnnotation(annotationType);
					if (annot != null) {
						annotatedFields.add(field);
					}
				}
				fields = annotatedFields.toArray(new Field[annotatedFields.size()]);
				cachedFields.put(clazz, fields);
			}

			if (annotationTypes.length == 1) {
				return fields;
			}
			else {
				if (temp == null) {
					temp = new ArrayList<Field>();
					temp.addAll(Arrays.asList(fields));
				}
			}
		}

		if (temp == null) {
			return new Field[0];
		}

		return temp.toArray(new Field[temp.size()]);
	}

	/**
	 * Retrieves a singular fields of a class, including those inherited from superclasses or interfaces, that has the
	 * specified annotation.  It is assumed that there will be only one field with this annotation.
	 *
	 * @param clazz             The class to inspect for annotated fields.  All of its ancestors will also be inspected.
	 * @param annotationType    The type of annotation to look for.
	 *
	 * @return                  The field with the specified annotation, or <code>null</code> if none is found.
	 *
	 * @throws IllegalArgumentException If more than one field with the specified annotation is found.
	 */
	public static <T extends Annotation> Field getFieldWithAnnotation(Class clazz, Class<T> annotationType)
			throws IllegalArgumentException
	{
		Field[] annotatedFields = getFieldsWithAnnotation(clazz, annotationType);
		switch (annotatedFields.length) {
			case 0:
				return null;
			case 1:
				return annotatedFields[0];
			default:
				throw new IllegalArgumentException("Field has " + annotatedFields.length + " annotations of type " + annotationType.getName());
		}
	}

	/**
	 * Convenience method for getting a field defined on a class or any of its superclasses.  Using this method has
	 * several benefits over using Class.getDeclaredField(String)
	 * <ul>
	 *      <li>This method iteratively scans all superclasses, whereas the Class API does not.</li>
	 *      <li>This method generates an in memory cache of all results, reducing number of reflection API calls.</li>
	 *      <li>Calling this method does not require setting up an exception handling block.</li>
	 *      <li>When this call fails, the exception thrown is derived from RuntimeException so it doesn't have to be caught.</li>
	 * </ul>
	 *
	 * @param clazz     The class on which the desired field is defined, or any subclass of it.
	 * @param name      The name of the field.
	 * @return
	 * @throws IllegalArgumentException     If the class does not have a field by the name specified, nor does any of its
	 *                                      superclasses.
	 */
	public static Field getField(Class<?> clazz, String name) throws IllegalArgumentException
	{
		getFields(clazz);
		Field field = mapCache.get(clazz).get(name);
		if (field == null) {
			throw new IllegalArgumentException("Field does not exist: " + clazz.getName() + "." + name);
		}
		return field;
	}
	
	public static boolean containsField(Class<?> clazz, String name)
	{
		getFields(clazz);
		return mapCache.get(clazz).get(name) != null;
	}

	/**
	 *
	 * @param obj
	 * @param name
	 * @param value
	 */
	public static void setFieldValue(Object obj, String name, Object value)
	{
		Field field = FieldUtils.getField(obj.getClass(), name);
		FieldUtils.setFieldValue(obj, field, value);
	}

	/**
	 *
	 * @param obj
	 * @param name
	 * @param value
	 */
	public static void setConvertedFieldValue(Object obj, String name, Object value)
	{
		Field field = FieldUtils.getField(obj.getClass(), name);

		Object typedValue = com.portlandwebworks.commons.reflection.ClassUtils.convertToType(value, field.getType());

		try {
			field.set(obj, typedValue);
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("Field not accessible: " + obj.getClass().getName() + "." + field.getName());
		}
	}

	/**
	 * 
	 * @param obj
	 * @param field
	 * @param value
	 */
	public static void setFieldValue(Object obj, Field field, Object value)
	{
		try {
			field.set(obj, value);
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("Field not accessible: " + obj.getClass().getName() + "." + field.getName());
		}
	}

	/**
	 * Returns the empty constructor for a class, and throws an exception if none is found or is inaccessible.
	 *
	 * @param clazz         The class whose constructor is desired.
	 * @return              The empty constructor definition.
	 * @throws              RuntimeException if an empty constructor is not found or is inaccessible.
	 */
	public static <T> Constructor<T> getDefaultConstructor(Class<T> clazz)
	{
		if (clazz == null) {
			throw new NullPointerException("Class cannot be null");
		}
		
		try {
			Constructor<T> defaultConstructor = clazz.getConstructor();
			if (! Modifier.isPublic(defaultConstructor.getModifiers())) {
				throw new RuntimeException(
						String.format("Empty constructor for class of type %s is inaccessible", clazz.getName())
				);
			}

			return defaultConstructor;
		}
		catch (NoSuchMethodException e) {
			throw new RuntimeException(String.format("No empty constructor found for class %s", clazz.getName()));
		}
	}

	/**
	 *
	 * @param entityClass
	 * @return
	 */
	public static boolean hasMultiValueField(Class entityClass)
	{
		if (! fieldMultiplicities.containsKey(entityClass)) {
			//process this class
			List<String> fieldNames = FieldUtils.getMultiValueFieldNames(entityClass);
			fieldMultiplicities.put(entityClass, fieldNames);
		}
		return fieldMultiplicities.get(entityClass).size() > 0;
	}


	/**
	 * Generates a list of all fields that are considered multi-value, as defined in #isMultiValueField.
	 *
	 * @param entityClass
	 *
	 * @return  A <code>List</code> of all field names that are multi-value.
	 */
	public static List<String> getMultiValueFieldNames(Class entityClass) {

		List<String> fieldNames = new ArrayList<String>();

		PropertyDescriptor[] descriptors = BeanUtils.getPropertyDescriptors(entityClass);
		for (PropertyDescriptor descriptor : descriptors) {
			Class<?> propertyType = descriptor.getPropertyType();
			if (propertyType.isArray() || Collection.class.isAssignableFrom(propertyType) || Map.class.isAssignableFrom(propertyType)) {
				fieldNames.add(descriptor.getName());
			}
		}

		return fieldNames;
	}

	/**
	 * Determines whether the class provided can be represented in SQL query results by a single columnar value.  This
	 * would be satisfied by the following conditions:
	 * <ul>
	 *      <li>The class represents a primitive type.</li>
	 *      <li>The class is a wrapper for a primitive type.</li>
	 *      <li>The class is of type java.lang.String</li>
	 *      <li>The class is of type java.util.Date, java.sql.Time or java.sql.Timestamp</li>
	 *      <li>The class is an enum</li>
	 * </ul>
	 * @param type
	 * @return
	 */
	public static boolean isValueType(Class<?> type)
	{
		return type.isPrimitive() || type.equals(String.class) || type.isAssignableFrom(Date.class) ||
				Enum.class.isAssignableFrom(type) || ClassUtils.wrapperToPrimitive(type) != null;
	}

	/**
	 * Creates a new instance of the class provided without having to declare an exception handling block.
	 *
	 * @param clazz
	 * @param <T>
	 * @return
	 * @throws IllegalArgumentException If an instance of the class cannot be instantiated due to no default constructor
	 *                                  or a private empty constructor.
	 */
	public static <T> T newInstance(Class<T> clazz) throws IllegalArgumentException
	{
		try {
			return clazz.newInstance();
		}
		catch (InstantiationException e) {
			throw new IllegalArgumentException("Unable to instantiate class " + clazz.getName());
		}
		catch (IllegalAccessException e) {
			throw new IllegalArgumentException("No public default constructor available");
		}
	}

	/**
	 * TODO: move this out of field utils.  It is specific to entity mapping
	 * @param type
	 * @return
	 */
	public static Field[] getIdentifierFieldsForClass(Class<?> type)
	{
		Field[] mappedFields = uniqueFields.get(type);
		if (mappedFields == null) {
			List<Field> fieldList = new ArrayList<Field>();
			mappedFields = FieldUtils.getFieldsWithAnnotation(type, FieldMapping.class);
			for (Field field : mappedFields) {
				FieldMapping mapping = field.getAnnotation(FieldMapping.class);
				if (mapping.identifier()) {
					// If a field mapping has a dot in it, it refers to the identifier of a nested entity, not the
					// one passed to this method.
					if (mapping.value().length() == 0 || mapping.value().indexOf(".")== -1) {
						fieldList.add(field);
					}
				}
			}
			mappedFields = fieldList.toArray(new Field[fieldList.size()]);
			uniqueFields.put(type, mappedFields);
		}
		return mappedFields;
	}

	/**
	 * Determines if any nested entity referenced by the field path provided represents an array, Map or Collection.
	 *
	 * @param entityClass
	 * @param fieldPath
	 * @return
	 */
	public static boolean isMultiValueField(Class<?> entityClass, String fieldPath)
	{
		if (! fieldMultiplicities.containsKey(entityClass)) {
			// Lazily evaluate all of the fields available on the class and its superclasses.
			List<String> fieldNames = FieldUtils.getMultiValueFieldNames(entityClass);
			fieldMultiplicities.put(entityClass, fieldNames);
		}

		// If the fieldName provided refers to a nested element instead of a field directly on the class, traverse the
		// relationship to the target property to determine whether that property's multiplicity.
		boolean isMultiple = false;
		Class<?> iterClass = entityClass;
		String[] parts = fieldPath.split("\\.");
		if (parts.length > 1) {
			for (String part: parts) {

				Field field = FieldUtils.getField(iterClass, part);
				Class<?> fieldType = field.getType();
				if (fieldType.isArray() || Collection.class.isAssignableFrom(fieldType) || Map.class.isAssignableFrom(fieldType)) {
					isMultiple = true;
					break;
				}
				else {
					iterClass = fieldType;
				}
			}

			if (isMultiple) {
				fieldMultiplicities.get(entityClass).add(fieldPath);
				return true;
			}
		}

		return fieldMultiplicities.get(entityClass).contains(fieldPath);
	}

	/**
	 * 
	 * @param entityClass
	 * @param visitor
	 */
	public static void traverseClassGraph(Class<?> entityClass, FieldTypeVisitor visitor, Class<? extends Annotation>... filters)
	{
		Stack<String> pathStack = new Stack<String>();
		Stack<Class> entityStack = new Stack<Class>();
		entityStack.push(entityClass);
		
		while (! entityStack.isEmpty()) {

			Class<?> iterClass = entityStack.peek();

			Field[] fields = FieldUtils.getFieldsWithAnnotation(entityClass, filters);
			for (Field field : fields) {
				Class<?> fieldType = field.getType();
				NodeType nodeType = FieldUtils.getNodeTypeForClass(fieldType);
				Class nestedClass = null;
				switch (nodeType) {
					case Array:
						nestedClass = fieldType.getComponentType();
						break;
					case Collection:
						nestedClass = (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[0];
						break;
					case Map:
						nestedClass = (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[1];
						break;
					case Object:
						nestedClass = fieldType;
						break;
					case ValueType:
						break;
				}
	
				String part = StringUtils.join(pathStack, ".");
				if (! visitor.visit(part, iterClass, nodeType)) {
					break;
				}

				if (nestedClass != null) {
					entityStack.push(nestedClass);
					pathStack.push(field.getName());
				}
			}
			
			entityStack.pop();
			pathStack.pop();
		}
	}


	/**
	 * Given a root Class, recursively traverses all relationships in the class definition and reports the type of each
	 * node, as well as the path at which it was found and the generalized "type" of the node.  This is useful for
	 * traversing over an object graph defined by a set of classes, as opposed to traversing over an actual graph of
	 * existing objects.  It does not build the object structure as it traverses, nor does it expect the visitor to do
	 * so.  Rather, it interprets class definitions to navigate the graph using the dot-notated field path provided.
	 *
	 * @param entityClass
	 * @param fieldPath
	 * @param visitor
	 */
	public static void traversePathInClassGraph(Class<?> entityClass, String fieldPath, FieldTypeVisitor visitor)
	{
		Class<?> iterClass = entityClass;
		String[] parts = fieldPath.split("\\.");

		for (String part: parts) {
			Field field = FieldUtils.getField(iterClass, part);
			Class<?> fieldType = field.getType();
			NodeType nodeType = FieldUtils.getNodeTypeForClass(fieldType);
			switch (nodeType) {
				case Array:
					iterClass = fieldType.getComponentType();
					break;
				case Collection:
					Type type = ((ParameterizedType)field.getGenericType()).getActualTypeArguments()[0];
					if (TypeVariable.class.isAssignableFrom(type.getClass())){
						iterClass = (Class)((TypeVariable)type).getBounds()[0];
					}
					else {
						iterClass = (Class)type;
					}
					break;
				case Map:
					iterClass = (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[1];
					break;
				case ValueType:
				case Object:
					iterClass = fieldType;
					break;
			}

			if (! visitor.visit(part, iterClass, nodeType)) {
				break;
			}
		}
	}

	/**
	 * Determines the "node type" for the given class.
	 *
	 * @param clazz The class to evaluate.
	 *
	 * @return  The NodeType for the class.
	 */
	public static NodeType getNodeTypeForClass(Class<?> clazz)
	{
		if (clazz.isArray()) {
			return NodeType.Array;
		}
		else if (Collection.class.isAssignableFrom(clazz)) {
			return NodeType.Collection;
		}
		else if (Map.class.isAssignableFrom(clazz)) {
			return NodeType.Map;
		}
		else if (FieldUtils.isValueType(clazz)) {
			return NodeType.ValueType;
		}
		else {
			return NodeType.Object;
		}
	}

	public static Class<?> getComponentTypeForField(Field field)
	{
		Class<?> type = field.getType();

		if (type.isArray()) {
			return type.getComponentType();
		}
		else if (Collection.class.isAssignableFrom(type)) {
			return (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[0];
		}
		else if (Map.class.isAssignableFrom(type)) {
			return (Class)((ParameterizedType)field.getGenericType()).getActualTypeArguments()[1];
		}
		else {
			return type;
		}
	}

	public static Object[] getFieldValues(Object obj, Field[] fields)
	{
		List list = new ArrayList(fields.length);

		try {
			for (Field field : fields) {
				list.add(field.get(obj));
			}
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("Unhandled exception accessing field");
		}

		return list.toArray();
	}

	public static Object getFieldValue(Object obj, Field field)
	{
		try {
			return field.get(obj);
		}
		catch (IllegalAccessException e) {
			throw new RuntimeException("Unhandled exception accessing field");
		}
	}

	public static Object getFieldValue(Object obj, String fieldName)
	{
		Field field = FieldUtils.getField(obj.getClass(), fieldName);
		return FieldUtils.getFieldValue(obj, field);
	}

	/**
	 *
	 * @return
	public static Field getFieldForPath(Object bean, String fieldPath)
	{
		final Field[] field = new Field[1];

		FieldUtils.traversePathInObjectGraph(bean, fieldPath, new NodeVisitor()
		{
			public boolean visit(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
			{
				if (nodeType == NodeType.ValueType) {
					field[0] = nodeField;
				}
			}

			public void leaving(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
			{
			}
		});

		return field[0];
	}
	 */

	public static interface FieldTypeVisitor
	{
		/**
		 *  Callback for evaluating the types of nodes in a dot-notated field path.
		 *
		 * @param path          The name of the visited field.
		 * @param elementType   The type of the visited field.  If the field is an array, it is the component type of the
		 *                      array.  If the field is a map or collection, it is the generic type that is stored in
		 *                      the map or collection.
		 * @param nodeType      The type of node that is being visited.
		 *
		 * @return              Return true to continue traversing, or false to terminate.
		 */
		boolean visit(String path, Class<?> elementType, NodeType nodeType);
	}


	/**
	 * Iterates over each entity node in a path of entity associations, and applies a visitor implementation to each of
	 * the nodes.  It will restrict its traversal based on the dot-notated path provided.  If any node in the path
	 * represents an array, collection or map, the iteration will branch out over each element in the data structure.
	 * It will also short circuit when it reaches a null value in the object graph.
	 *
	 * @param rootObject        The starting point of the object graph to traverse.
	 * @param pathToProcess     The dot-notated field path that needs to be traversed from the root object.
	 * @param visitor           A callback that should be executed at each node visited.
	 */
	public static void traversePathInObjectGraph(Object rootObject, String pathToProcess, NodeVisitor visitor)
	{
		if (StringUtils.isBlank(pathToProcess)) {
			throw new IllegalArgumentException("associationPath must not be null or empty");
		}

		Object iterObject = rootObject;
		StringBuilder assembledPath = new StringBuilder();

		// Iterate over each part of the dot-separated association path.
		String[] pathParts = pathToProcess.split("\\.");
		for (int i = 0; i < pathParts.length; i++) {

			String iterProperty = pathParts[i];

			// The assembled path will be the amount of the path processed so far, including the node we are currently
			// processing
			assembledPath.append(i == 0? "": ".").append(iterProperty);

			// Get the field that the path element represents by name.
			Field iterField = FieldUtils.getField(rootObject.getClass(), iterProperty);

			// Check the value of the field being iterated over.  If it is null, can't continue the traversal.
			Object iterValue = FieldUtils.getFieldValue(iterObject, iterField);

			NodeType nodeType = FieldUtils.getNodeTypeForClass(iterField.getType());

			// Attempt to recurse on the next object in the associationPath.  If the visitor returns null, done
			// traversing the association path, this recursion will return false.
			if (! visitor.visit(assembledPath.toString(), iterObject, iterField, iterValue, nodeType)) {
				break;
			}

			// Re-fetch the node value in case the visitor changed it.
			// Check the value of the field being iterated over.  If it is null, can't continue the traversal.
			iterValue = FieldUtils.getFieldValue(iterObject, iterField);
			if (iterValue == null) {
				break;
			}

			// Build the remaining association path that needs to be traversed.
			if (i < pathParts.length - 1) {
				String[] remainingParts = Arrays.copyOfRange(pathParts, i + 1, pathParts.length);
				String remainingPath = StringUtils.join(remainingParts, '.');

				// Determine the type of the value being iterated over to determine whether it needs to be iterated
				// over to visit every item value.
				if (nodeType == NodeType.ValueType) {
					// Doesn't make sense to traverse over a scalar type.
					break;
				}
				else if (nodeType == NodeType.Array) {
					List list = Arrays.asList((Object[])iterValue);
					for (Object obj : list) {
						// Branch the traversal for each element in the array
						traversePathInObjectGraph(obj, remainingPath, visitor);
					}
					break;
				}
				else if (nodeType == NodeType.Collection) {
					for (Object o : ((Collection)iterValue)) {
						// Branch the traversal for each element in the collection
						traversePathInObjectGraph(o, remainingPath, visitor);
					}
					break;
				}
				else if (nodeType == NodeType.Map) {
					for (Object o : ((Map)iterValue).values()) {
						// Branch the traversal for each value in the map
						traversePathInObjectGraph(o, remainingPath, visitor);
					}
					break;
				}
				else if (nodeType == NodeType.Object) {
					// The current object is a single value.  Allow this for loop to continue the traversal.  Update the
					// object being iterated over to be the new value first.
					iterObject = iterValue;
				}
			}
		}
	}

	/**
	 * Traverses all nodes in a graph of associated objects, allowing custome operations to be performed at each node.
	 * It will traverse into the fields of nested objects, and will spawn traversal "branches" for each nested object in
	 * a one-to-many relationship.
	 * If the NodeVisitor implementation returns false at any specific node, the traversal of the
	 *
	 * @param rootObject
	 * @param visitor
	 */
	public static void traverseObjectGraph(Object rootObject, NodeVisitor visitor)
	{
		if (rootObject == null) {
			throw new IllegalArgumentException("Root object for traversal is null");
		}

		Map<String, Object> traversalContext = new HashMap<String, Object>();
		traversalContext.put("visitedObjects", new HashSet());
		traverseObjectGraph(rootObject, "", visitor, traversalContext);
	}

	private static void traverseObjectGraph(Object rootObject, String pathTraversed, NodeVisitor visitor, Map<String, Object> traversalContext)
	{
		Object iterObject = rootObject;

		// Short-circuit traversal of the root object is a value type that has no fields to traverse.
		if (FieldUtils.isValueType(iterObject.getClass())) {
			return;
		}

		Field[] fields = FieldUtils.getFields(iterObject.getClass());

		for (Field iterField : fields) {

			String iterProperty = iterField.getName();

			String pathSoFar = pathTraversed.equals("")? iterProperty: pathTraversed + "." + iterProperty;

			NodeType nodeType = FieldUtils.getNodeTypeForClass(iterField.getType());

			// Re-fetch the node value, in case the visitor changed its value.
			Object iterValue = FieldUtils.getFieldValue(iterObject, iterField);
			if (iterValue == null) {
				// Nothing left to traverse here.  Move on to the next field on the object
				continue;
			}

			Set visitedObjects = (Set)traversalContext.get("visitedObjects");
			if (visitedObjects.contains(iterValue)) {

				logger.debug("Avoiding circular reference by skipping object {}", iterValue);

				// This object has already been visited.  Avoid circular chain on mutual references by bailing now.
				continue;
			}

			// Determine the type of the value being iterated over to determine whether it needs to be iterated
			// over to visit every item value.
			if (nodeType == NodeType.ValueType) {

				if (logger.isDebugEnabled()) {
					logger.debug(String.format("Visit %s, type=%s, value=%s", pathSoFar, nodeType, iterValue));
				}

				// Ignore the return value on this visit because it is not possible to traverse beyond a value type.
				visitor.visit(pathSoFar, iterObject, iterField, iterValue, nodeType);

				if (logger.isDebugEnabled()) {
					logger.debug(String.format("Leave %s, type=%s, value=%s", pathSoFar, nodeType, iterValue));
				}

				visitor.leaving(pathSoFar, iterObject, iterField, iterValue, nodeType);
			}
			else if (nodeType == NodeType.Object) {

				if (logger.isDebugEnabled()) {
					logger.debug(String.format("Visit %s, type=%s, value=%s", pathSoFar, nodeType, iterValue));
				}

				// Field has a single object value.  "Visit" this object and then try to traverse it.
				boolean shouldContinue = visitor.visit(pathSoFar, iterObject, iterField, iterValue, nodeType);

				visitedObjects.add(iterValue);

				if (shouldContinue) {
					traverseObjectGraph(iterValue, pathSoFar, visitor, traversalContext);
				}

				if (logger.isDebugEnabled()) {
					logger.debug(String.format("Leave %s, type=%s, value=%s", pathSoFar, nodeType, iterValue));
				}

				// Inform the visitor that we are leaving this object node.
				visitor.leaving(pathSoFar, iterObject, iterField, iterValue, nodeType);
			}
			else {
				// Convert one-to-many fields into a uniform Collection type.
				Collection children = null;
				if (nodeType == NodeType.Array) {
					children = Arrays.asList((Object[])iterValue);
				}
				else if (nodeType == NodeType.Collection) {
					children = (Collection)iterValue;
				}
				else if (nodeType == NodeType.Map) {
					children = ((Map)iterValue).values();
				}

				if (logger.isDebugEnabled()) {
					logger.debug(String.format("Visit %s, type=%s, value=%s", pathSoFar, nodeType, iterValue));
				}

				// Visit the array/map/collection itself before visiting each member of the collection
				if (visitor.visit(pathSoFar, iterObject, iterField, iterValue, nodeType)) {

					// Find out what the type is of the collection members.
					Class componentClass = FieldUtils.getComponentTypeForField(iterField);
					NodeType childNodeType = FieldUtils.getNodeTypeForClass(componentClass);

					// Visit each child in the array/map/collection
					int index = 0;
					for (Object child : children) {

						if (logger.isDebugEnabled()) {
							logger.debug(String.format("Visit %s[%d], type=%s, value=%s", pathSoFar, index, childNodeType, child));
						}

						boolean shouldContinue = visitor.visit(pathSoFar, iterObject, null, child, childNodeType);

						// Remember we traversed this object.
						visitedObjects.add(child);

						if (shouldContinue && childNodeType != NodeType.ValueType) {
							// If the collection object is not a value type, traverse into it.
							traverseObjectGraph(child, pathSoFar, visitor, traversalContext);
						}

						if (logger.isDebugEnabled()) {
							logger.debug(String.format("Leave %s[%d], type=%s, value=%s", pathSoFar, index, childNodeType, child));
						}

						// Leaving the collection member.
						visitor.leaving(pathSoFar, iterObject, null, child, childNodeType);

						index++;
					}
				}

				if (logger.isDebugEnabled()) {
					logger.debug(String.format("Leave %s, type=%s, value=%s", pathSoFar, nodeType, iterValue));
				}

				// Leaving the collection.
				visitor.leaving(pathSoFar, iterObject, iterField, iterValue, nodeType);
			}
		}
	}

	/**
	 * Visitor pattern used to perform custom operations on each "node" in an object graph visited using the algorithm
	 * selected.
	 */
	public static interface NodeVisitor
	{
		/**
		 * Callback that is executed at each "node" visited.  A node is the object stored in a field or collection as
		 * part of the dot-notated path being traversed.
		 *
		 *
		 * @param nodePath      The path leading to this node, starting from the root object.
		 * @param nodeObject    The object that owns the field that is being traversed.
		 * @param nodeField     The field that represents the node being traversed
		 * @param nodeType      Determines whether the node represents an object, array, collection, map, or value type.
		 * @return              Return true to continue traversing beyond this field into its aggregate fields, or false
		 *                      to end this traversal.
		 */
		boolean visit(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType);

		void leaving(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType);
	}

	public static Map mapBeanData(Object bean)
	{
		final Map<String, Object> rootMappings = new HashMap<String, Object>();
		rootMappings.put("class", bean.getClass());

		final Stack containerStack = new Stack();

		// Put the root object onto the stack.
		containerStack.push(rootMappings);

		// Convert the object graph into a graph of maps.  Much easier to process as maps.
		FieldUtils.traverseObjectGraph(bean, new FieldUtils.NodeVisitor() {
			public boolean visit(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
			{
				switch (nodeType) {
					case ValueType: {
						Object target = containerStack.peek();
						//ValueToApply valueToApply = new ValueToApply(nodeField, nodeValue);
						if (target instanceof Map) {
							Map map = (Map)target;
							//map.put(nodeField.getName(), valueToApply);
							map.put(nodeField.getName(), nodeValue);
						}
						else if (target instanceof ArrayList) {
							ArrayList list = (ArrayList)target;
							//list.add(valueToApply);
							list.add(nodeValue);
						}
						break;
					}
					case Object: {
						Object target = containerStack.peek();

						HashMap<String, Object> newMap = new HashMap<String, Object>();
						newMap.put("class", nodeValue.getClass());

						if (target instanceof Map) {
							Map map = (Map)target;
							map.put(nodeField.getName(), newMap);
							containerStack.push(newMap);
						}
						else if (target instanceof ArrayList) {
							ArrayList list = (ArrayList)target;
							list.add(newMap);
							containerStack.push(newMap);
						}
						break;
					}
					case Array:
					case Collection:
					case Map: {
						// One-to-many relationship.  Add a collection to the stack for storing multiple values.
						Map target = (Map)containerStack.peek();
						ArrayList beans = new ArrayList();
						target.put(nodeField.getName(), beans);
						containerStack.push(beans);
						break;
					}
				}
				return true;
			}

			public void leaving(String nodePath, Object nodeObject, Field nodeField, Object nodeValue, NodeType nodeType)
			{
				// Since we're not pushing a new container onto the stack when traversing into a value field, it makes no
				// sense to pop a container off the stack when traversing out of of a value type field.
				if (nodeType != NodeType.ValueType) {
					containerStack.pop();
				}
			}
		});

		return rootMappings;
	}



}
