/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.security;

import org.springframework.security.access.AccessDeniedException;

import java.util.List;


/**
 * This software is the property of Portland Webworks.
 *
 * Abstracts out the manner by which code leveraging spring security gets the security context.
 *
 * @Author rnorthup
 */
public interface ISecurityProvider
{
	/**
	 * Determines whether the current session user has fully authenticated.
	 *
	 * @return  <code>true</code> if the user is authenticated
	 */
	boolean isAuthenticated();

	/**
	 * Retrieves the object that represents the logged in user.  This could be the string value of the user's login or
	 * the User object built by the underlying security implementation.
	 *
	 * @return  Implementation dependent.
	 *
	 * @throws SecurityException
	 */
	Object getPrincipal() throws SecurityException;

	/**
	 * Returns a list of strings representing
	 *
	 * @return
	 */
	List<String> getRoles();

	/**
	 * Authenticates a user with the credentials provided.  Once authentication is complete, <code>isAuthenticated()</code> will
	 * return true and <code>getPrincipal()</code> will return either the logged-in principal.
	 * @param login
	 * @param password
	 * @return
	 */
	AuthenticationResult authenticate(String login, String password);

	/**
	 *
	 * @param principal
	 * @param password
	 * @param roles
	 */
	void changeAuthentication(Object principal, String password, String... roles);

	/**
	 *
	 * @param role
	 * @return
	 */
	boolean hasRole(Object role);

	/**
	 *
	 * @param role
	 * @throws AccessDeniedException
	 */
	void assertHasRole(Object role) throws AccessDeniedException;

	/**
	 *
	 * @param roles
	 * @return
	 */
	boolean hasAnyRole(Object... roles);

	/**
	 * Throws an exception if the context principal does not have any of the specified roles.
	 *
	 * @param roles
	 * @throws AccessDeniedException
	 */
	void assertHasAnyRole(Object... roles) throws AccessDeniedException;
}
