package com.portlandwebworks.commons.security.user;

import com.portlandwebworks.commons.data.LabelValuePair;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.ldap.core.AttributesMapper;
import org.springframework.ldap.core.DirContextAdapter;
import org.springframework.ldap.core.DirContextOperations;
import org.springframework.ldap.core.LdapTemplate;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.ldap.userdetails.UserDetailsContextMapper;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class LdapEnabledUserServiceImpl extends UserServiceImpl implements ILdapEnabledUserService, ApplicationContextAware, InitializingBean, UserDetailsContextMapper
{
	private static Logger logger = LoggerFactory.getLogger(LdapEnabledUserServiceImpl.class);

	@Autowired(required = false)
	private LdapTemplate ldapTemplate;

	private String ldapUserSearchBaseDn;

	private String ldapGroupSearchBaseDn;

	private String ldapGroupSearchFilter;

	private String ldapGroupMembersAttribute;

	private String ldapGroupMemberIdPattern;

	protected AttributesMapper ldapAttributeMapper = new DefaultLdapAttributeMapper();

	private ApplicationContext applicationContext;


	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException
	{
		this.applicationContext = applicationContext;
	}

	public String getLdapUserSearchBaseDn()
	{
		return ldapUserSearchBaseDn;
	}

	/**
	 *
	 * @param ldapUserSearchBaseDn
	 */
	public void setLdapUserSearchBaseDn(String ldapUserSearchBaseDn)
	{
		this.ldapUserSearchBaseDn = ldapUserSearchBaseDn;
	}

	/**
	 * Indicates what the base DN should be for locating the group which contains all members who can be entitled to this
	 * application.  This DN will be prepended to the base DN defined for the LdapTemplate utilized by this service for
	 * directory access.
	 *
	 * @param ldapGroupSearchBaseDn  The DN, for example ou=Groups
	 */
	@Required
	public void setLdapGroupSearchBaseDn(String ldapGroupSearchBaseDn)
	{
		this.ldapGroupSearchBaseDn = ldapGroupSearchBaseDn;
	}

	/**
	 * Indicates what LDAP query filter to apply to the group base DN to find the group containing all members who can be
	 * entitled to this application.
	 *
	 * @param ldapGroupSearchFilter  The filter, for example (cn=JmesUsers)
	 */
	@Required
	public void setLdapGroupSearchFilter(String ldapGroupSearchFilter)
	{
		this.ldapGroupSearchFilter = ldapGroupSearchFilter;
	}

	/**
	 * Indicates the name of the multi-value attribute that contains the DNs of all members of the group entitled to this
	 * application.
	 *
	 * @param ldapGroupEntryAttribute   The name of the attribute.
	 */
	@Required
	public void setLdapGroupMembersAttribute(String ldapGroupEntryAttribute)
	{
		this.ldapGroupMembersAttribute = ldapGroupEntryAttribute;
	}

	/**
	 * Indicates the regular expression that should be used to strip out the user ID from the list of attribute values
	 * retrieved from the LDAP group entitled to this application.
	 *
	 * @param ldapGroupMemberIdPattern  A regular expression
	 */
	@Required
	public void setLdapGroupMemberIdPattern(String ldapGroupMemberIdPattern)
	{
		this.ldapGroupMemberIdPattern = ldapGroupMemberIdPattern;
	}

	/**
	 * Specifies which class to use for mapping LDAP attributes to a user retrieved from a directory.  This property is
	 * exposed because different environments' LDAPs may have some variance in the way the store values in a user entry.
	 *
	 * @param ldapAttributeMapperClass
	 * @see    DefaultLdapAttributeMapper
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 */
	@Required
	public void setLdapAttributeMapperClass(String ldapAttributeMapperClass) throws ClassNotFoundException, IllegalAccessException, InstantiationException
	{
		this.ldapAttributeMapper = (AttributesMapper)Class.forName(ldapAttributeMapperClass).newInstance();
	}

	public void afterPropertiesSet() throws Exception
	{
		if (this.ldapTemplate != null) {
			if (StringUtils.isBlank(this.ldapGroupMemberIdPattern)) {
				throw new Exception("");
			}
			if (StringUtils.isBlank(this.ldapGroupMembersAttribute)) {
				throw new Exception("");
			}
			if (StringUtils.isBlank(this.ldapUserSearchBaseDn)) {
				throw new Exception("");
			}
			//TODO: what else?
		}
	}

	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public List<LabelValuePair<String>> searchDirectoryUsers(String lastName, boolean excludeLocalUsers)
	{
		// Query the configured LDAP location for all users who can be entitled to this application.
		List results = ldapTemplate.search(ldapGroupSearchBaseDn, ldapGroupSearchFilter, new AttributesMapper()
		{
			public Object mapFromAttributes(Attributes attributes) throws NamingException
			{
				// The list of users in the group consist of a multi-value attribute.  Iterate through each of the values
				// in the attribute and return a list of those attribute values.
				List<String> dns = new ArrayList<String>();
				Attribute attr = attributes.get(ldapGroupMembersAttribute);
				NamingEnumeration<?> enumer = attr.getAll();
				while (enumer.hasMoreElements()) {
					String userDn = (String)enumer.nextElement();
					dns.add(userDn);
				}
				return dns;
			}
		});
		// Can't find the specified group?  Complain.
		if (results.size() == 0) {
			throw new RuntimeException(String.format("No LDAP entry found: base=%s, filter=(%s)" + ldapGroupSearchBaseDn, ldapGroupSearchFilter));
		}

		// There will be only one result from this search.
		List<String> dns = (List<String>)results.get(0);

		Set<String> usedLogins = new HashSet<String>();

		if (excludeLocalUsers) {
			// Generate an in-memory set of all login names in the database.  These should not be made available to the client
			// in the list of external users.
			List<String> allUserNames = entityManager.executeQuery(String.class, "select u.username from User u");
			usedLogins.addAll(allUserNames);
		}

		// Parse the user IDs from the multi-valued ldap attributes in the group using the configured regexp
		List<LabelValuePair<String>> usernames = new ArrayList<LabelValuePair<String>>(dns.size());
		Pattern regex = Pattern.compile(ldapGroupMemberIdPattern);
		for (String dn : dns) {
			Matcher matcher = regex.matcher(dn);
			if (matcher.find()) {
				final String userId = matcher.group(0);
				logger.debug("Parsed user ID {} from ldap attribute {}", userId, dn);

				if (excludeLocalUsers && usedLogins.contains(userId)) {
					logger.debug("Skipping external user ID {}, already in local database", userId);
				}
				else {
					usernames.add(new LabelValuePair<String>(userId, userId));
				}
			}
			else {
				logger.warn("Failed to parse user ID from ldap attribute {}", dn);
			}
		}

		Collections.sort(usernames, new Comparator<LabelValuePair<String>>()
		{
			public int compare(LabelValuePair<String> o1, LabelValuePair<String> o2)
			{
				return o1.getValue().compareTo(o2.getValue());
			}
		});

		return usernames;
	}

	public UserInput getUserInputFromDirectory(String username)
	{
		if (ldapTemplate == null) {
			throw new RuntimeException("Cannot query for external users--no LDAP configuration specified.");
		}

		List result = ldapTemplate.search(this.ldapUserSearchBaseDn, "uid=" + username, ldapAttributeMapper);

		if (result.size() > 0) {
			return (UserInput)result.get(0);
		}
		else {
			return null;
		}

	}


	/**
	 * Implementation of the UserDetailsContextMapper method that populates a custom UserDetails object after LDAP
	 * authentication.
	 *
	 * @param dirContextOperations
	 * @param username
	 * @param grantedAuthorities
	 * @return
	 */
	@Transactional(propagation = Propagation.MANDATORY, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public UserDetails mapUserFromContext(DirContextOperations dirContextOperations, String username, Collection<GrantedAuthority> grantedAuthorities)
	{
		return this.loadUserByUsername(username);
	}

	public void mapUserToContext(UserDetails userDetails, DirContextAdapter dirContextAdapter)
	{
		throw new UnsupportedOperationException("Not implemented");
	}
}
