/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.security.user;

import com.portlandwebworks.commons.dao.*;
import com.portlandwebworks.commons.data.IEntityBuilder;
import com.portlandwebworks.commons.data.handlers.DefaultEntityHandler;
import com.portlandwebworks.commons.domain.Identifiable;
import com.portlandwebworks.commons.domain.Permission;
import com.portlandwebworks.commons.domain.Role;
import com.portlandwebworks.commons.domain.User;
import com.portlandwebworks.commons.email.EmailResult;
import com.portlandwebworks.commons.email.EmailSender;
import com.portlandwebworks.commons.reflection.ClassUtils;
import com.portlandwebworks.commons.security.ISecurityProvider;
import com.portlandwebworks.commons.service.ServiceBase;
import com.portlandwebworks.commons.service.ServiceOperationResult;
import com.portlandwebworks.commons.validation.CustomValidationException;
import com.portlandwebworks.commons.validation.SimpleConstraintViolation;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.time.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.authority.GrantedAuthorityImpl;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.stereotype.Controller;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;

import javax.validation.ConstraintViolation;
import javax.validation.Valid;
import java.util.*;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
@Controller("userService")
@RequestMapping("/user")
public class UserServiceImpl extends ServiceBase implements IUserService, UserDetailsService, InitializingBean
{
	private static Logger logger = LoggerFactory.getLogger(UserServiceImpl.class);
	
	@Autowired
	private ISecurityProvider securityProvider;

	@Autowired(required = false)
	private IEntityBuilder entityBuilder;
	
	@Autowired(required = false)
	private EmailSender emailSender;

	@Autowired(required = false)
	private IPasswordGenerator passwordGenerator;
	
	private boolean permissionsEnabled = false;
	
	private String[] userEditRoles = null;

	private String[] userAccessRoles = null;

	/**
	 * This delegate object will handle any custom implementations of User and Role
	 */
	@Autowired(required = false)
	private IUserDelegate userDelegate = new DefaultUserDelegate();

	public static final Integer DEFAULT_PASSWORD_EXPIRATION_DAYS = 30;
	public static final Integer DEFAULT_PASSWORD_RESET_EXPIRATION_DAYS = 10;
	public static final String DEFAULT_PASSWORD_RESET_EMAIL_TEMPLATE = "com/portlandwebworks/email/templates/passwordReset";
	public static final String DEFAULT_PASSWORD_CHANGE_EMAIL_TEMPLATE = "com/portlandwebworks/email/templates/passwordChanged";
	public static final String DEFAULT_USER_ADDED_EMAIL_TEMPLATE = "com/portlandwebworks/email/templates/userAdded";

	private Integer passwordExpirationDays = DEFAULT_PASSWORD_EXPIRATION_DAYS;

	private Integer passwordResetExpirationDays = DEFAULT_PASSWORD_RESET_EXPIRATION_DAYS;
	
	private String userAddedEmailTemplate = DEFAULT_USER_ADDED_EMAIL_TEMPLATE;

	private String passwordChangeEmailTemplate = DEFAULT_PASSWORD_CHANGE_EMAIL_TEMPLATE;

	private String passwordResetEmailTemplate = DEFAULT_PASSWORD_RESET_EMAIL_TEMPLATE;


	/**
	 * Injects application-specific functionality to supplement this user service.
	 * 
	 * @param userDelegate
	 */
	public void setUserDelegate(IUserDelegate userDelegate)
	{
		this.userDelegate = userDelegate;
	}

	/**
	 * Allows injecting an entity builder for persisting and retrieving user related data.
	 * 
	 * @param entityBuilder
	 */
	public void setEntityBuilder(IEntityBuilder entityBuilder)
	{
		this.entityBuilder = entityBuilder;
	}

	/**
	 * Allows injecting an email sender service for sending user-related events such as password expiration warnings and
	 * password reset instructions.
	 * 
	 * @param emailSender
	 */
	public void setEmailSender(EmailSender emailSender)
	{
		this.emailSender = emailSender;
	}

	public void setPasswordGenerator(IPasswordGenerator passwordGenerator)
	{
		this.passwordGenerator = passwordGenerator;
	}

	/**
	 * Configures the number of days a password reset token is valid.  Once this window of days has passed, the token can
	 * no longer be used to reset the user's password.
	 * 
	 * @param passwordResetExpirationDays
	 */
	public void setPasswordResetExpirationDays(Integer passwordResetExpirationDays)
	{
		this.passwordResetExpirationDays = passwordResetExpirationDays;
	}

	/**
	 * Configures the number of days that will pass before a user's password must be changed.
	 * 
	 * @param passwordExpirationDays
	 */
	public void setPasswordExpirationDays(Integer passwordExpirationDays)
	{
		this.passwordExpirationDays = passwordExpirationDays;
	}

	/**
	 * Whe3n set to true, this user service will utilize a permissions-based authorization mechanism that involves a set
	 * of permissions mapped to roles, instead of the simplisitc role-only approach.  This means that when a user is
	 * authenticated, their principal object will be populated with fine grained permissions instead of the higher level
	 * roles.
	 * 
	 * @param permissionsEnabled
	 */
	public void setPermissionsEnabled(boolean permissionsEnabled)
	{
		this.permissionsEnabled = permissionsEnabled;
	}

	public void setUserAddedEmailTemplate(String userAddedEmailTemplate)
	{
		this.userAddedEmailTemplate = userAddedEmailTemplate;
	}

	public void setPasswordChangeEmailTemplate(String passwordChangeEmailTemplate)
	{
		this.passwordChangeEmailTemplate = passwordChangeEmailTemplate;
	}

	public void setPasswordResetEmailTemplate(String passwordResetEmailTemplate)
	{
		this.passwordResetEmailTemplate = passwordResetEmailTemplate;
	}

	/**
	 * Indicates which user roles are allowed to make edits to users (other than their own personal information).
	 *
	 * @param userEditRoles     A comma-delimited set of one or more role names.
	 */
	public void setUserEditRoles(String userEditRoles)
	{
		this.userEditRoles = userEditRoles.split(",");
	}

	/**
	 * Indicates which user roles are allowed to view information about users other than themselves
	 *
	 * @param userAccessRoles  A comma-delimited set of one or more role names.
	 */
	public void setUserAccessRoles(String userAccessRoles)
	{
		this.userAccessRoles = userAccessRoles.split(",");
	}

	public void setSecurityProvider(ISecurityProvider securityProvider)
	{
		this.securityProvider = securityProvider;
	}


	public void afterPropertiesSet() throws Exception
	{
		if (passwordGenerator == null) {
			passwordGenerator = new DefaultPasswordGenerator();
		}
	}

	/**
	 * Custom validation for UserInput object that requres access to an IEntityManager instance.
	 *
	 *
	 *
	 * @param input
	 *
	 * @throws CustomValidationException
	 */
	protected Set<ConstraintViolation<UserInput>> validateUserInput(UserInput input)
	{
		Set<ConstraintViolation<UserInput>> validationErrors = super.validateInput(input);
		
		Predicate duplicateUserCheck;
		if (input.getUserId() != null) {
			duplicateUserCheck = Predicate.and(
					Predicate.notEqualTo("id", input.getUserId()),
					Predicate.equalTo("username", input.getUsername())
			);
		}
		else {
			duplicateUserCheck = Predicate.equalTo("username", input.getUsername());
		}

		// Do check for duplicate username before attempting to save
		User duplicateUser = entityManager.findByPredicate(userDelegate.getUserType(), duplicateUserCheck);
		if (duplicateUser != null) {
			ConstraintViolation duplicateUsername = new SimpleConstraintViolation<UserInput>(
					input, "username", input.getUsername(),
					getDefaultMessage("userService.edit.duplicateUsername", "Username %s is already in use", input.getUsername())
			);
			
			validationErrors.add(duplicateUsername);
		}
		
		return validationErrors;
	}

	@RequestMapping(value = "", method = RequestMethod.POST)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED)
	public @ResponseBody ServiceOperationResult<Long> addUser(@RequestBody UserInput input) throws DuplicateLoginException
	{
		if (this.userEditRoles != null) {
			securityProvider.assertHasAnyRole(this.userEditRoles);
		}

		if (! userDelegate.getUserInputType().isAssignableFrom(input.getClass())) {
			UserInput customInput = ClassUtils.newInstanceOfClass(userDelegate.getUserInputType());
			customInput.putAll(input);
			input = customInput;
		}

		Set<ConstraintViolation<UserInput>> validationErrors = this.validateUserInput(input);
		if (! validationErrors.isEmpty()) {
			return new ServiceOperationResult<Long>(null, false,
					getDefaultMessage("userService.add.inputValidationFailed", "Invalid form fields"),
					super.mapConstraintViolations(validationErrors));
		}

		User user = (User)entityBuilder.createEntity(input, "admin");

		userDelegate.prepareUserForAdd(input, user);

		Map<String, Object> params = new HashMap<String, Object>();
		if (StringUtils.isNotBlank(input.getPassword())) {
			params.put("temporaryPassword", input.getPassword());
		}
		else {
			String generated = passwordGenerator.generateTemporaryPassword();
			user.setPassword(passwordGenerator.getPasswordHash(generated));
			params.put("temporaryPassword", generated);
		}

		entityManager.insert(user);

		userDelegate.userAdded(input, user);
		
		emailSender.sendUserEmail(userAddedEmailTemplate, params, user);

		return new ServiceOperationResult<Long>(user.getId(), true);
	}


	@RequestMapping(value = "/{userId}", method = RequestMethod.PUT)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED)
	public @ResponseBody ServiceOperationResult<Void> updateUser(@PathVariable("userId") long userId, @RequestBody UserInput input)
	{
		if (this.userEditRoles != null) {
			securityProvider.assertHasAnyRole(this.userEditRoles);
		}

		if (! userDelegate.getUserInputType().isAssignableFrom(input.getClass())) {
			UserInput customInput = ClassUtils.newInstanceOfClass(userDelegate.getUserInputType());
			customInput.putAll(input);
			input = customInput;
		}

		final User user = entityManager.findById(userDelegate.getUserType(), userId);

		userDelegate.prepareUserForUpdate(input, user);

		Set<ConstraintViolation<UserInput>> validationErrors = this.validateUserInput(input);
		if (! validationErrors.isEmpty()) {
			return new ServiceOperationResult<Void>(null, false,
					getDefaultMessage("userService.update.inputValidationFailed", "Invalid form fields"),
					super.mapConstraintViolations(validationErrors));
		}

		final boolean[] passwordChanged = new boolean[] { false };

		entityBuilder.updateEntity(user, input, "admin", new DefaultEntityHandler() {
			@Override
			public void entityFieldValueApplied(Identifiable entity, String path, Object oldValue, Object newValue)
			{
				if (path.equals("password")) {
					// Could use a transformer to convert this value to a hash, but it is dependent upon the password generator
					// configured for this service.
					user.setPassword(passwordGenerator.getPasswordHash((String)newValue));
					passwordChanged[0] = true;
				}
			}
		});

		entityManager.update(user);

		userDelegate.userUpdated(user, input);
		
		if (passwordChanged[0]) {
			emailSender.sendUserEmail(passwordChangeEmailTemplate, null, user);
		}
		
		return ServiceOperationResult.simpleSuccessfulResult;
	}

	@RequestMapping(value = "/options", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody UserInputOptions getUserInputOptions()
	{
		return getUserInputOptions(null);
	}

	@RequestMapping(value = "/{userId}/options", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody UserInputOptions getUserInputOptions(@PathVariable("userId") Long userId)
	{
		UserInputOptions uio = ClassUtils.newInstanceOfClass(userDelegate.getUserInputOptionsType());

		uio.setRoles(entityManager.findAll(userDelegate.getRoleType()));

		userDelegate.prepareUserInputOptions(userId, uio);

		return uio;
	}

	@RequestMapping(value = "/options/{optionName}", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody List<?> getUserInputOptions(
			@PathVariable("optionName")String optionName,
			@RequestParam(value = "qualifier", required = false) String qualifier)
	{
		return userDelegate.getUserInputOptions(optionName, qualifier);
	}

	@RequestMapping(value = "/search/options", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody UserSearchOptions getUserSearchOptions()
	{
		UserSearchOptions uso = ClassUtils.newInstanceOfClass(userDelegate.getUserSearchOptionsType());

		uso.setRoles(entityManager.findAll(Role.class));

		userDelegate.prepareUserSearchOptions(uso);

		return uso;
	}

	@RequestMapping(value = "/search/options/{optionName}", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody List getUserSearchOptions(
			@PathVariable("optionName") String optionName,
			@RequestParam(value = "qualifier", required = false) String qualifier)
	{
		return userDelegate.getUserSearchOptions(optionName, qualifier);
	}

	@RequestMapping(value = "", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody PagedSearchResult<? extends UserSearchResult> searchUsers(@ModelAttribute PredicateSearchCriteria criteria)
	{
		if (this.userAccessRoles != null) {
			securityProvider.assertHasAnyRole(this.userAccessRoles);
		}

		userDelegate.prepareUserSearchCriteria(criteria);

		PagedSearchResult<? extends UserSearchResult> results =
				entityManager.search(userDelegate.getUserSearchResultType(), criteria, criteria);

		userDelegate.prepareUserSearchResults(results.getItems());

		return results;
	}
	
	@RequestMapping(value = "/self")
	@Override
	public @ResponseBody User getSessionUser()
	{
		User user = super.getSessionUser();
		user.setPassword(null);
		user.setPasswordHistory(null);
		return user;
	}


	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public User getUserByLogin(String login)
	{
		User user = entityManager.findByPredicate(userDelegate.getUserType(), Predicate.equalTo("username", login));

		// Eager fetch roles to avoid lazy load errors.
		for (Role role : (List<Role>)user.getRoles()) {
			role.getCode();
		}
		
		userDelegate.prepareUserForView(user);
		
		return user;
	}


	@RequestMapping(value = "/{userId}/input", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody UserInput getUserInput(@PathVariable("userId") long userId)
	{
		if (this.userEditRoles != null) {
			securityProvider.assertHasAnyRole(this.userEditRoles);
		}

		UserInput input = entityManager.findById(userDelegate.getUserInputType(), userId);
		
		input.setPassword(null);

		userDelegate.prepareUserInputForEdit(input);

		return input;
	}

	@RequestMapping(value = "/{userId}", method = RequestMethod.GET)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public @ResponseBody User getUser(@PathVariable("userId") long userId)
	{
		if (this.userEditRoles != null) {
			securityProvider.assertHasAnyRole(this.userEditRoles);
		}

		User user = entityManager.findById(userDelegate.getUserType(), userId);

		userDelegate.prepareUserForView(user);

		return user;
	}


	@RequestMapping(value = "/password/reset", method = RequestMethod.POST)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED)
	public @ResponseBody ServiceOperationResult<Void> resetUserPassword(@RequestParam("username") String username)
	{
		User user = entityManager.findByPredicate(userDelegate.getUserType(), Predicate.equalTo("username", username));
		
		if (user == null) {
			logger.warn("Attempt to reset password for non-existent user " + username);
			return new ServiceOperationResult<Void>(false,
					getDefaultMessage("userService.passwordReset.invalidUsername", "No account found with username %s", username));
		}

		// Assign a random, non-guessable "token" to the user.  This token will be used to resolve this user for a
		// password change in the future.
		user.setPasswordResetToken(UUID.randomUUID().toString());
		user.setPasswordResetDate(new Date());
		
		entityManager.update(user);
		
		EmailResult result = emailSender.sendUserEmail(passwordResetEmailTemplate, null, user);
		if (result.getFailures().size() > 0) {
			return new ServiceOperationResult<Void>(false,
					getDefaultMessage("userService.passwordReset.emailError", "Unable to send reset password to %s", user.getEmail()));
		}
		
		return ServiceOperationResult.simpleSuccessfulResult;
	}


	@RequestMapping(value = "/password/change", method = RequestMethod.POST)
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED)
	public @ResponseBody ServiceOperationResult<Void> changeUserPassword(@RequestBody @Valid ChangePasswordInput input)
	{
		User user;

		if (StringUtils.isNotBlank(input.getPasswordResetToken())) {
			// If a password change token is provided, the user is changing their password without an active login
			// session.  Look up the user by this token.
			user = entityManager.findByPredicate(userDelegate.getUserType(), Predicate.equalTo("passwordResetToken", input.getPasswordResetToken()));
			if (user == null) {
				logger.warn("Attempt to change password for non-existent password reset token " + input.getPasswordResetToken());
				return new ServiceOperationResult<Void>(false, getDefaultMessage("userService.passwordChange.invalidResetToken", "Invalid password reset token"));
			}

			if (user.getPasswordResetDate() == null) {
				return new ServiceOperationResult<Void>(false, getDefaultMessage("userService.passwordChange.noResetDate", "Bad user data. User has password reset token but no date when it was generated"));
			}

			Date passwordResetLimit = DateUtils.addDays(user.getPasswordResetDate(), 10);
			if (passwordResetLimit.before(new Date())) {
				return new ServiceOperationResult<Void>(false, getDefaultMessage("userService.passwordChange.invalidResetDate", "Password reset token expired"));
			}
		}
		else {
			// User is changing their password with an active login session.
			try {
				user = super.getSessionUser();
				user.getUsername(); // force user != null
			}
			catch (Exception e) {
				logger.warn("Attempting to change password for logged-in user, unable to retrieve session user");
				return new ServiceOperationResult<Void>(false, getDefaultMessage("userService.passwordChange.loginSessionRequred", "Invalid login session for password reset"));
			}
		}

		if (! user.isEnabled()) {
			return new ServiceOperationResult<Void>(false, getDefaultMessage("userService.passwordChange.userDisabled", "Can't change password for a disabled user"));
		}
		
		//TODO: password requirements?

		//TODO: make enforcement of password reuse a settable service property.
		String passwordHash = passwordGenerator.getPasswordHash(input.getNewPassword());
		if (StringUtils.defaultString(user.getPasswordHistory()).contains(passwordHash)) {
			return new ServiceOperationResult<Void>(false, getDefaultMessage("userService.passwordChange.repeatedPassword", "You have re-used a recent password."));
		}

		user.setPassword(passwordHash);
		user.setPasswordResetToken(null);
		user.setPasswordResetDate(null);
		user.setLastPasswordChangeDate(new Date());
		
		// Update the user's password change history, keeping a maximum of 10 most recent passwords.
		String passwordHashHistory = user.getPasswordHistory();
		if (StringUtils.isBlank(passwordHashHistory)) {
			user.setPasswordHistory(user.getPassword());
		}
		else {
			String[] passwords = passwordHashHistory.split(",");
			while (passwords.length > 9) {
				passwords = (String[])ArrayUtils.remove(passwords, 0);
			}
			passwords = (String[])ArrayUtils.add(passwords, user.getPassword());
			user.setPasswordHistory(StringUtils.join(passwords, ","));
		}

		entityManager.update(user);

		// If the client wants to establish a login session with the new credentials, do so now.
		if (BooleanUtils.isTrue(input.isLoginOnSuccess())) {
			
			List<Role> roles = user.getRoles();
			List<String> roleNames = new ArrayList<String>(user.getRoles().size());
			
			for (Role role : roles) {
				roleNames.add(role.getCode().toString());
			}
			securityProvider.changeAuthentication(user, input.getNewPassword(), roleNames.toArray(new String[roleNames.size()]));
		}

		emailSender.sendUserEmail(passwordChangeEmailTemplate, null, user);
		
		return ServiceOperationResult.simpleSuccessfulResult;
	}


	/**
	 * This method was created for spring security integration.  It provides the spring security framework with a custom
	 * implementation of UserDetails that provides some additional information about the user.  It is not exposed as a
	 * service.
	 *
	 * @param username
	 * @return
	 * @throws org.springframework.security.core.userdetails.UsernameNotFoundException
	 * @throws org.springframework.dao.DataAccessException          If more than one user found matching the provided username
	 */
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED, readOnly = true)
	public UserDetails loadUserByUsername(String username) throws AuthenticationException
	{
		try {
			User user = entityManager.findByPredicate(userDelegate.getUserType(), Predicate.equalTo("username", username));

			if (user == null) {
				throw new UsernameNotFoundException(username);
			}

			// I think this user object is getting reused by hibernate.
			user.getAuthorities().clear();

			Date pwdChangeDate = user.getLastPasswordChangeDate();
			if (pwdChangeDate == null) {
				// User has a temporary password.
				logger.info("User lastPasswordChangeDate is null.  Adding ROLE_CHANGE_PASSWORD");

				user.getAuthorities().add(new GrantedAuthorityImpl("ROLE_CHANGE_PASSWORD"));
			}
			else {
				final Date pwdExpireDate = DateUtils.addDays(pwdChangeDate, passwordExpirationDays);
				if (pwdExpireDate.compareTo(new Date()) < 0) {
					user.setPasswordExpired(true);
				}
			}

			// Convert internal user role structure into a list of strings to satisfy spring's granted authority equivalent.
			List<Role> roles = user.getRoles();

			if (this.permissionsEnabled) {
				List<String> roleCodes = new ArrayList<String>(roles.size());
				for (Role role : (List<Role>)user.getRoles()) {
					roleCodes.add(role.getCode().toString());
				}
				List<Permission> permissions = entityManager.findAllByPredicate(
						Permission.class, Predicate.isIn("role", roleCodes)
				);

				// The more complex (and more powerful/flexible) permissions-based authorization strategy.
				for (Permission perm : permissions) {
					user.getAuthorities().add(new GrantedAuthorityImpl(perm.getName()));
				}
			}
			else {
				for (Role role : roles) {
					// The simpler (and less flexible) role-based authorization strategy.
					user.getAuthorities().add(new GrantedAuthorityImpl(role.getCode().toString()));
				}
			}

			return user;
		}
		catch (DataIntegrityException die) {
			throw new DataIntegrityViolationException(getDefaultMessage("userService.load.duplicateAccounts", "More than one match found for username %s", username));
		}
	}

	/**
	 * Updates a user's last login date.  This method must be called within a transaction.
	 *
	 * @param user
	 */
	@Transactional(propagation = Propagation.REQUIRED, isolation = Isolation.READ_COMMITTED)
	void logUserAuthentication(User user)
	{
		user.setLastLoginDate(new Date());
		
		entityManager.executeUpdate(
				"update " + userDelegate.getUserType().getName() + " set lastLoginDate = ? where id = ?",
				user.getLastLoginDate(), user.getId());
		
//		entityManager.update(user);
	}
}
