/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.service;

import com.portlandwebworks.commons.dao.IEntityManager;
import com.portlandwebworks.commons.dao.Predicate;
import com.portlandwebworks.commons.domain.User;
import com.portlandwebworks.commons.security.ISecurityProvider;
import com.portlandwebworks.commons.validation.CustomValidationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;

import javax.annotation.PostConstruct;
import javax.validation.ConstraintViolation;
import javax.validation.Validation;
import javax.validation.Validator;
import javax.validation.ValidatorFactory;
import java.util.HashMap;
import java.util.Set;

/**
 * This software is the property of Portland Webworks.
 *
 * Common functionality for service type classes.
 *
 * @author Ronnie Northup
 */
public class ServiceBase
{
	protected Logger logger = LoggerFactory.getLogger(this.getClass());
	
	@Autowired(required = false)
	protected ISecurityProvider securityProvider;

	@Autowired(required = true)
	protected IEntityManager entityManager;

	private static Validator validator;

	@Autowired(required = false)
	private MessageSource messages;

	/**
	 * Initializes the validator used by this service instance.
	 */
	@PostConstruct
 	public void init()
	{
		if (validator == null) {
			ValidatorFactory factory = Validation.buildDefaultValidatorFactory();

			//TODO: should an exception be thrown if a validator cannot be resolved?

			validator = factory.getValidator();
		}
	}

	/**
	 * Returns the entity representation of the currently logged in user.
	 *
	 * @return
	 */
	protected User getSessionUser()
	{
		if (securityProvider == null) {
			throw new RuntimeException("No security provider is configured for this service");
		}

		Object principal = securityProvider.getPrincipal();
		if (principal instanceof User) {
			return (User)principal;
		}
		else {
			if (principal instanceof String) {
				return entityManager.findByPredicate(User.class, Predicate.equalTo("username", principal));
			}
		}
		return null;
	}

	/**
	 * Validates an input object using the configured JSR303 validation provider.  If validation fails
	 *
	 * @param input     Any POJO will suffice
	 * @param <T>       The generic type used to perform validation.
	 * @throws          CustomValidationException if input object fails validation.
	 */
	protected <T> void assertInputIsValid(T input) throws CustomValidationException
	{
		assertInputIsValid(input, null);
	}

	/**
	 * Validates an input object using the configured JSR303 validation provider.  If validation fails
	 *
	 * @param input     Any POJO will suffice
	 * @param message
	 * @param <T>       The generic type used to perform validation.
	 * @throws          CustomValidationException if input object fails validation.
	 */
	protected <T> void assertInputIsValid(T input, String message) throws CustomValidationException
	{
		Set<ConstraintViolation<T>> violations = validator.validate(input);
		if (violations.size() > 0) {
			throw new CustomValidationException(message == null? "Input validation failed": message, violations.toArray());
		}
	}

	/**
	 * Runs the configured validator against the input object provided, and returns a set of any constraint violations
	 * generated by the validation process.
	 *
	 * @param input
	 * @param <T>
	 * @return
	 */
	protected <T> Set<ConstraintViolation<T>> validateInput(T input)
	{
		return validator.validate(input);
	}

	/**
	 * Loads a set of constraint violations into a Map, keyed by the name of the violating field.
	 * @param violations
	 * @param <T>
	 * @return
	 */
	protected <T> HashMap<String, String> mapConstraintViolations(Set<ConstraintViolation<T>> violations)
	{
		HashMap<String, String> errors = new HashMap<String, String>();
		for (ConstraintViolation<T> violation : violations) {
			errors.put(violation.getPropertyPath().toString(), violation.getMessage());
		}
		return errors;
	}

	/**
	 * Validates an input object using the configured JSR303 provider, and collects all errors into a NewServiceOperationResult
	 * if any validations fail.
	 *
	 * @param input     Any POJO will suffice
	 * @param <T>       The type of input, which will be used during validation
	 * @param <R>       The generic type for the result object.
	 * @return          A ServiceOperationResult if validation fails, or null if input passes validation.
	 */
	protected <T, R> ServiceOperationResult<R> getValidationErrors(T input)
	{
		Set<ConstraintViolation<T>> violations = validator.validate(input);
		if (violations.size() > 0) {
			final ServiceOperationResult<R> result = new ServiceOperationResult<R>(null, false, "Validation failed");
			for (ConstraintViolation<T> violation : violations) {
				result.getErrors().put(violation.getPropertyPath().toString(), violation.getMessage());
			}
			return result;
		}

		return null;
	}

	/**
	 * Returns a string defined in a resource bundle configured in Spring.
	 *
	 * @param key
	 * @param params
	 * @return
	 */
	protected String getMessage(String key, Object... params)
	{
		return messages.getMessage(key, params, null);
	}

	/**
	 * Generates a string message by first checking for a name/value property defined in a spring-configured message
	 * bundle.  If no message is found with the specified key, uses the default message provided and applies the parameters
	 * to it using String.format() notation.
	 *
	 *
	 * @param key
	 * @param defaultMessage
	 *@param params  @return
	 */
	protected String getDefaultMessage(String key, String defaultMessage, Object... params)
	{
		try {
			return getMessage(key, params);
		}
		catch (Exception e) {
			return String.format(defaultMessage, params);
		}
	}

	@ExceptionHandler
//	@ResponseStatus(value = HttpStatus.INTERNAL_SERVER_ERROR)
	public @ResponseBody	ServiceOperationResult<Void> handleException(Exception e)
	{
		logger.error("Unhandled service exception", e);
		if (e.getCause() != null) {
			return new ServiceOperationResult<Void>(false, e.getCause().getMessage());
		}
		else {
			return new ServiceOperationResult<Void>(false, e.getMessage());
		}
	}
}
