package com.portlandwebworks.commons.servlet;

import com.portlandwebworks.commons.spring.security.PasswordChangeEnforcingFilter;
import org.apache.commons.lang.ArrayUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public abstract class PatternMatchingFilter extends FilterBean
{
	private String[] includeUrlPatterns = ArrayUtils.EMPTY_STRING_ARRAY;

	private String[] excludeUrlPatterns = ArrayUtils.EMPTY_STRING_ARRAY;

	private Logger logger = LoggerFactory.getLogger(PasswordChangeEnforcingFilter.class);

	/**
	 * The filter implementation should apply its logic to this request.
	 *
	 * @param request
	 * @param response
	 * @throws IOException
	 */
	public abstract void urlIncluded(HttpServletRequest request, HttpServletResponse response) throws IOException;

	/**
	 * The filter implementation should not apply its logic to this request.
	 * @param request
	 * @param response
	 * @throws IOException
	 */
	public abstract void urlExcluded(HttpServletRequest request, HttpServletResponse response) throws IOException;

	/**
	 * An array of URL fragments that indicate what URLs should be enforced by this filter.  If none are specified, all
	 * URLs will be validated for access.  The filter accepts a * wildcard.
	 *
	 * @param includeUrlPatterns
	 */
	public void setIncludeUrlPatterns(String[] includeUrlPatterns)
	{
		this.includeUrlPatterns = includeUrlPatterns;
	}

	public String[] getIncludeUrlPatterns()
	{
		return includeUrlPatterns;
	}

	/**
	 * An array of URL fragments that indicate what URLs should not be enforced by this filter.  These excludes will 
	 * override all includes applied a request URI.  It is not necessary to put the passwordChangeUrl in this list.
	 *
	 * @param excludeUrlPatterns
	 */
	public void setExcludeUrlPatterns(String[] excludeUrlPatterns)
	{
		this.excludeUrlPatterns = excludeUrlPatterns;
	}

	public String[] getExcludeUrlPatterns()
	{
		return excludeUrlPatterns;
	}

	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException
	{
		HttpServletRequest httpRequest = (HttpServletRequest)request;
		HttpServletResponse httpResponse = (HttpServletResponse)response;

		String requestUri = httpRequest.getRequestURI();
		String requestSansContext = requestUri.substring(httpRequest.getContextPath().length());

		boolean isExcluded = testUrlPatternMatch(this.excludeUrlPatterns, requestSansContext);
		if (isExcluded) {
			this.urlExcluded(httpRequest, httpResponse);
		}
		else {
			boolean isIncluded = testUrlPatternMatch(this.includeUrlPatterns, requestSansContext);
			if (isIncluded) {
				this.urlIncluded(httpRequest, httpResponse);
			}
		}

		if (! response.isCommitted()) {
			chain.doFilter(request, response);
		}
	}

	private boolean testUrlPatternMatch(String[] patterns, String requestSansContext)
	{
		for (String pattern : patterns) {
			if (pattern.endsWith("**")) {
				String patternSansWildcard = pattern.substring(0, pattern.length() - 2);
				if (requestSansContext.startsWith(patternSansWildcard)) {
					return true;
				}
			}
			else if (pattern.endsWith("*")) {
				String patternSansWildcard = pattern.substring(0, pattern.length() - 1);
				if (requestSansContext.startsWith(patternSansWildcard)) {

					String remainder = requestSansContext.substring(patternSansWildcard.length());
					if (! remainder.contains("/")) {
						logger.debug("Pattern {} matches request URI {}", pattern, requestSansContext);
						return true;
					}
				}
			}
			else {
				if (requestSansContext.equals(pattern)) {
					logger.debug("Pattern {} matches request URI {}", pattern, requestSansContext);
					return true;
				}
			}
		}
		return false;
	}

	public void destroy()
	{
	}
}
