/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.template;

import org.apache.velocity.Template;
import org.apache.velocity.app.VelocityEngine;

import javax.annotation.PostConstruct;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class VelocityTemplateProvider implements ITemplateProvider
{
	protected VelocityEngine velocityEngine;

	protected String templateDirectory;
	
	protected Map<String, Object> globalVariables = new HashMap<String, Object>();
	
	protected Map<String, String> cachedResourcePaths = new HashMap<String, String>();

	public String getTemplateDirectory()
	{
		return templateDirectory;
	}

	public void setTemplateDirectory(String templateDirectory)
	{
		this.templateDirectory = templateDirectory;
	}

	public void setGlobalVariables(Map<String, Object> globalVariables)
	{
		this.globalVariables.putAll(globalVariables);
	}

	@PostConstruct
	public void initialize() throws Exception
	{
		velocityEngine = new VelocityEngine();
		velocityEngine.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM_CLASS, this.getClass());

		Properties props = new Properties();
		props.setProperty("resource.loader", "class");
		props.setProperty("class.resource.loader.class", "org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader");
		//props.setProperty("class.resource.loader.path", templateDirectory);
		props.setProperty("class.resource.cache", "false");
		props.setProperty("class.resource.loader.modificationCheckInterval", "120");
/*
		props.setProperty("file.resource.loader.class", "org.apache.velocity.runtime.resource.loader.FileResourceLoader");
		props.setProperty("file.resource.loader.path", templateDirectory);
		props.setProperty("file.resource.cache", "false");
		props.setProperty("file.resource.loader.modificationCheckInterval", "120");
*/

		try {
			velocityEngine.init(props);
		}
		catch (Exception e) {
			throw new RuntimeException("Unhandled exception initializing velocity engine", e);
		}

	}

	public ITemplate getTemplate(String templateName)
	{
		return getTemplate(templateName, Locale.getDefault());
	}

	public ITemplate getTemplate(String templateName, Locale locale)
	{
		String resourcePath = cachedResourcePaths.get(templateName);
		if (resourcePath == null) {
			// Attempt to resolve the template by first trying the most locale-specific resource name possible, and if the
			// resource is not found keep reducing the specificity of the template until it is found.
			resourcePath = templateDirectory + "/" + locale.getLanguage() + "_" +
					locale.getCountry() + "/" + templateName + ".vm";

			InputStream in = this.getClass().getResourceAsStream(resourcePath);
			if (in == null) {

				resourcePath = templateDirectory + "/" + locale.getLanguage() + "/" + templateName + ".vm";
				in = this.getClass().getResourceAsStream(resourcePath);
				if (in == null) {

					resourcePath = templateDirectory + "/" + templateName + ".vm";
					in = this.getClass().getResourceAsStream(resourcePath);
					if (in == null) {
						throw new RuntimeException("Could not find template " + templateName + " for any locale");
					}
				}
			}
			cachedResourcePaths.put(templateName, resourcePath);
		}

		try {
			Template vm = velocityEngine.getTemplate(resourcePath);

			VelocityTemplate vt = new VelocityTemplate(vm);

			// Every template that is created should contain the configured globals.
			vt.setParamValues(globalVariables);
			
			return vt;
		}
		catch (Exception e) {
			throw new RuntimeException("Unhandled exception loading velocity template " + resourcePath, e);
		}
	}
}
