/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.test.dbunit;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.dbunit.DatabaseUnitException;
import org.dbunit.database.DatabaseSequenceFilter;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.FilteredDataSet;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.filter.ITableFilter;
import org.dbunit.dataset.xml.XmlDataSet;
import org.dbunit.operation.DatabaseOperation;
import org.springframework.test.context.TestContext;

import java.io.*;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * @author Ronnie Northup
 */
public class DbUnitUtils
{
	private static Logger logger = Logger.getLogger(DbUnitTestExecutionListener.class);

	/**
	 * 
	 * @param dbConnProvider
	 * @param filePath
	 * @param operation
	 * @throws SQLException
	 * @throws IOException
	 */
	public static void doDatabaseOperation(IDatabaseConnectionProvider dbConnProvider, String filePath, String operation)
			throws SQLException, IOException
	{
		DatabaseOperation dbop = null;

		if (operation.equals("CLEAN_INSERT")) {
			dbop = DatabaseOperation.CLEAN_INSERT;
		}
		else if (operation.equals("REFRESH")) {
			dbop = DatabaseOperation.REFRESH;
		}
		else if (operation.equals("UPDATE")) {
			dbop = DatabaseOperation.UPDATE;
		}
		else if (operation.equals("INSERT")) {
			dbop = DatabaseOperation.INSERT;
		}
		else if (operation.equals("DELETE")) {
			dbop = DatabaseOperation.DELETE;
		}
		else if (operation.equals("DELETE_ALL")) {
			dbop = DatabaseOperation.DELETE_ALL;
		}
		else if (operation.equals("TRUNCATE_TABLE")) {
			dbop = DatabaseOperation.TRUNCATE_TABLE;
		}
		else {
			throw new IllegalArgumentException("Unknown database operation " + dbop);
		}

		InputStream fileStream = null;

		try {
			if (filePath.equals("")) {
				throw new IllegalArgumentException("No dataset name provided");
			}

			// Load the dbunit XML dataset from either classpath or file system.
			if (filePath.startsWith("classpath:")) {

				String resourceName = filePath.substring(10).trim();
				logger.info("Loading DbUnit dataset from classpath: " + resourceName);

				fileStream = DbUnitUtils.class.getResourceAsStream(resourceName);
				if (fileStream == null) {
					throw new RuntimeException("Dataset resource not found in classpath: " + resourceName);
				}
			}
			else {
				try {
					logger.info("Loading DbUnit dataset from file: " + filePath);

					fileStream = new FileInputStream(filePath);
				}
				catch (FileNotFoundException e) {
					throw new RuntimeException("Dataset file not found: " + filePath);
				}
			}

			logger.info("Loading DbUnit dataset");

			// Load the dataset file.
			IDataSet dataSet = new XmlDataSet(fileStream);

			IDatabaseConnection connection = dbConnProvider.setupDatabaseConnection();

			dbop.execute(connection, dataSet);

/*
			Statement statement = connection.getConnection().createStatement();
			ResultSet resultSet = statement.executeQuery("select table_name from user_tables inner join user_sequences on 'SEQ_' || table_name = sequence_name;");
			List<String> tableNames = new ArrayList<String>();
			while (resultSet.next()) {
				tableNames.add(resultSet.getString(0));
			}
			resultSet.close();

			for (String tableName : tableNames) {

				ResultSet maxIdResult = statement.executeQuery("select max(id) from " + tableName + ";");
				if (maxIdResult.next()) {
					long maxId = maxIdResult.getLong(0) + 1;
					statement.execute("alter sequence " + tableName + " minvalue " + maxId + ";");
				}

			}
*/

		}
		catch (DatabaseUnitException e) {
			throw new RuntimeException("Unable to initialize database using dataset " + filePath, e);
		}
		finally {
			dbConnProvider.tearDownDatabaseConnection();

			if (fileStream != null) {
				fileStream.close();
			}
		}
	}

	/**
	 * Utility method for exporting the current state of a data source to an XML file that can be utilized by DBUnit
	 * for imports.
	 *
	 * @param provider  An instance of IDatabaseConnectionProvider
	 * @param filePath
	 */
	public static void doOrderedExport(IDatabaseConnectionProvider provider, String filePath)
	{
		FileWriter writer = null;
		try {
			IDatabaseConnection conn = provider.setupDatabaseConnection();
			ITableFilter filter = new DatabaseSequenceFilter(conn);
			IDataSet dataset = new FilteredDataSet(filter, conn.createDataSet());

			File exportFile = new File(filePath);
			File dir = exportFile.getParentFile();
			if (! dir.exists() && ! dir.mkdirs()) {
				throw new RuntimeException("Can't build directory structure " + dir.getAbsolutePath());
			}
			writer = new FileWriter(exportFile, false);
			XmlDataSet.write(dataset, writer);
		}
		catch (Exception e) {
			throw new RuntimeException("Unhandled exception running ordered export of database", e);
		}
		finally {
			provider.tearDownDatabaseConnection();
			IOUtils.closeQuietly(writer);
		}
	}

}
