/*
 * Copyright 2010 Portland Webworks, Inc.
 *
 * This file is part of PWW Commons.
 *
 * PWW Commons is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * PWW Commons is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with PWW Commons.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

package com.portlandwebworks.commons.text;

import org.apache.log4j.Logger;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * This software is the property of Portland Webworks.
 *
 * @Author rnorthup
 */
public class Dictionary
{
	// Pre-size this
	private static List<String> dictionary;

	private static HashMap<Integer, List<String>> wordsByLength;

	private static int wordCount;

	private static Logger logger = Logger.getLogger(Dictionary.class);


	private static void loadDictionary()
	{
		File words = null;
		try {
			long startTime = System.currentTimeMillis();

			// Pre-size the list to accommodate the size of words on my computer.
			dictionary = new ArrayList<String>(250000);
			wordsByLength = new HashMap<Integer, List<String>>();

			// This file could be in two possible locations.  First, try the unix standard, then Mac OS X.
			words = new File("/usr/dict/words");
			if (! words.exists()) {
				words = new File("/usr/share/dict/words");
			}

			if (words.exists()) {
				BufferedReader reader = new BufferedReader(new FileReader(words));
				String line = null;
				while (null != (line = reader.readLine())) {

					if (line.length() == 0)
						continue;

					dictionary.add(line);
				}
				wordCount = dictionary.size();

				long duration = System.currentTimeMillis() - startTime;
				logger.debug(MessageFormat.format("Finished loading dictionary file {0} in {1} millis", words.getAbsolutePath(), duration));
			}
			else {
				logger.warn("Could not locate a readable dictionary file");
				dictionary.add("One");
				dictionary.add("Two");
				dictionary.add("Three");
			}
		}
		catch (IOException e) {
			logger.error(MessageFormat.format("Unhandled exception loading dictionary file {0}", words.getAbsolutePath()), e);
		}
	}

	private static void loadWordsByLength()
	{
		if (dictionary == null) {
			loadDictionary();
		}

		for (String word : dictionary) {
			List<String> wordList = wordsByLength.get(word.length());
			if (wordList == null) {
				wordList = new ArrayList<String>();
				wordsByLength.put(word.length(), wordList);
			}
			wordList.add(word);

		}
	}

	/**
	 * Returns a randomly selected word of the specified length.
	 *
	 * @param length
	 * @return
	 */
	public static String getRandomWord(int length)
	{
		if (wordsByLength == null) {
			loadWordsByLength();
		}

		List<String> wordList = wordsByLength.get(length);
		if (wordList == null) {
			return "";
		}
		int randomIndex = (int)(Math.random() * 1000000) % length;

		return wordList.get(randomIndex);
	}

	public static String getRandomPhrase(int maxLength)
	{
		StringBuilder phrase = new StringBuilder();

		while (true) {
			
			int wordLen = ((int)(Math.random() * 100) % 6) + 1;
			
			if (phrase.length() + wordLen + 1 > maxLength)
				break;

			if (phrase.length() > 0)
				phrase.append(' ');

			phrase.append(getRandomWord(wordLen));
		}

		return phrase.toString();
	}

	/**
	 * Returns a randomly selected word.
	 *
	 * @return
	 */
	public static String getRandomWord()
	{
		if (dictionary == null) {
			synchronized (Dictionary.class) {
				if (dictionary == null) {
					loadDictionary();
				}
			}
		}
		int randomIndex = (int)(Math.random() * 1000000) % wordCount;

		return dictionary.get(randomIndex);
	}

	public static String getRandomEmailAddress()
	{
		return "";
	}
}